IFDEF TEST THEN
open Dumper
ENDIF
open Pervasives2

(* B: en attendant un peu mieux.... *)
open Transitions
open Term
open Definitions

type statelist= Statelist.t

type automaton =
    { trans: Transitions.t;
      fstates: Stateset.t
    }

let transtable (a: automaton) : Transitions.transtable =
  a.trans.transitions


(* Mapping a state to the corresponding term.
   Necessary to produce relevant outputs when detecting patterns (associated to a state
   in automaton representing lhs and patterns) *)

type s2t = (state, term) Hashtbl.t

(* mapping a state to the corresponding variable for lhs automaton *)
type statevar = (state, variable) Hashtbl.t

let state_list_to_string l =
  String.concat " " (List.map State.to_string l)

(* Tree automaton copy *)
let copy (a: automaton) : automaton =
  { trans = Transitions.copy a.trans;
    fstates = a.fstates
  }


(* Build a tree automaton from an external_trans list and a set of final states *)

let external_to_TA (fs: Stateset.t)(ls: Transitions.external_trans list): automaton=
  let max = (List2.fmax (function (Trans (_, _, q)) -> q | (Epsilon (q1,q2,_)) -> max q1 q2) 0 ls) + 1 in
  let transT = Transitions.make () in
    (* We build states and tables accordingly to the maximum state label *)
    for i=0 to max do ignore(Transitions.make_state transT) done;
    (try 
      List.iter (function (Trans t) -> Transitions.add transT t 
                   | (Epsilon (q,q',phi)) -> if not(phi=Formula.mk_false) then Transitions.addR transT (q,q',phi) else ()) ls
    with Invalid_argument("index out of bounds") -> failwith "Bad state max number!");
    { fstates = fs; trans = transT }


let new_statevar () = Hashtbl.create Common.max_var

let new_s2t () = Hashtbl.create Common.max_equation

let statevar_add (s: statevar) (q1: state) (v: variable) : unit =
  try
    if Hashtbl.find s q1 <> v then
      failwith
        "Automaton:statevar_add: different variable association for a state"
  with
    | Not_found -> Hashtbl.add s q1 v

let statevar_find = Hashtbl.find

let statevar_to_string t =
  Hashtbl.fold
    (fun k e str ->
       (State.to_string k) ^ " -> " ^ (Variable.to_string e) ^ "\n" ^ str)
    t ""

let s2t_to_string t =
  Hashtbl.fold
    (fun k e str ->
       (State.to_string k) ^ " -> " ^ (Term.to_string e) ^ "\n" ^ str)
    t ""

(* Run function guided by destination state.
   If t=f(q1) dest=q2 and delta contains
   f(q1) -> q3 and q3 -> q2 it is OK.  (The Transitions.get_trans is modulo epsilon)

   If t=f(q1) and dest=q2, then it is covered by a delta where
   f(q3) -> q2 and q1 ->E q3.

   If t=f(q1) and dest=q2, then it is NOT covered by a delta where
   f(q3) -> q2 and q1 ->R q3.
*)

(* answers true if it is possible to rewrite t to dest with transitions and
   epsilon transitions. Does not rewrite states of t *)

let rec oriented_run (d: Transitions.t) (t: term) (dest: state) : bool =
  Verbose.print 3
    (fun () -> "Oriented Run for t=" ^ (Term.to_string t) ^
       "\nwith delta=" ^ (Transitions.to_string d) ^ "\n");

  let oriented_subrun (tl: term list) (destl: state list) : bool =
    try
      List.for_all2 (oriented_run d) tl destl
    with
      | Invalid_argument _ ->
          failwith "Automaton:oriented_run_list: size of lists is not the same!"
  in
    match t with
      | Fsym (f, tl) ->
	  List.exists
	    (fun (_, sl, _) -> oriented_subrun tl sl)
	    (Transitions.get_trans dest f d)
      | State q ->
          q = dest || Transitions.is_succE d q dest (* q=dest or q =E dest *)
      | Var _ ->
          failwith "Automaton:oriented_run: non ground term!"


(* Deterministic run of t, with no epsilon resolving *)
let rec det_run (d: Transitions.t) : term -> term = function
  | Fsym (f, []) ->
      (match Transitions.get_trans_sym_constant f d with
 	 | [] -> Fsym (f, []) (* B: le cas ou il n'y a aucune transition *)
	 | (_, _, q)::_ -> State q (* B: le pattern pourrait etre [(_, _, q)],
                                      1 seule transition de la forme f() -> q *)
      )
  | Fsym (f, tl) ->
      (* B: reduction des sous-termes par delta *)
      let res_tl = List.map (det_run d) tl in
        (try
	   let res_states = List.map Term.get_state res_tl in
             (* B: calcul assez obscure des transitions de la forme
                f(q_1, \dots, q_n) -> q, en fait je crois qu'on recupere directement
                la liste de la forme [(q_1^1, \dots, q_n^1); \dots; (q_1^m, \dots, q_n^m)]
                ou chaque tuple j pour 1 <= j <= m, est de la forme [q_1^j; \dots; q_n^j]
             *)
	   let relevant_trans =
	     List.flatten
               (List.map
                  (fun q -> Transitions.get_trans_sym_parameter f q d)
		  res_states
               )
           in
	     try
	       let _, _, q =
                 List.find (fun  (_, sl ,_) -> sl = res_states) relevant_trans
               in
                 State q
	     with (* No transition matches *)
	       | Not_found -> Fsym (f, res_tl)
	 with (* Transition is not fully normalizable *)
	   | Term.Not_a_state -> Fsym (f, res_tl)
               (* B: qui balance cette exception et pourquoi ???
                  ca doit etre Term.get_state mais a quoi sert cette derniere
                  on n'a pas deja des etats ?!?!?
                  Reponse: get_state (State q) -> q sinon Not_a_state
               *)
        )
  | State q -> State q
  | Var _ -> failwith "Automaton:oriented_run: non ground term!"



(* Non deterministic run: use any transition to normalize a term (with epsilonE resolving only).
   Used for application of equations: given a \sigma and an equation l=r produces, for l\sigma all
   possible states reachable, modulo epsilonE. *)
(* gives every possible state in which a term can be rewritten or the empty list *)
(* Principle is to find all transitions rooted by the same top symbol (all
   f(q1...qn)->q for f(t1...tn) and then removing parameter states q...qn of transitions
   one after one, and keep only transitions such that they were matching t1...tn successively *)

(* if l\sigma -> q and q=q' then, by default, we don't give q' as a "reachable" state, since it will be added
   as a successor by E of all new predecessors of q by transitivity closure of epsilon E. More precisely,
   if r\sigma q'', then we only give the result q for l\sigma. Since all_run is only used for equation application
   we have l\sigma -> q and r\sigma -> q'' and q <->q'. By adding q <->q'' we add (by transitivity) q'' <-> q'.
   Hence, no need to give q' as a result.
 *)
let rec all_run (d: Transitions.t) (t: term) : state list =
  (*  let _= print_string ("Term= "^(Term.to_string t)^"\n") in *)
  match t with
    | Fsym (f, []) ->
        (match Transitions.get_trans_sym_constant f d with
	   | [] -> []
	   | l -> List.map (fun  (_, _, q) -> q) l
        )
    | Fsym (f, tl) ->
        all_run_list d tl (Transitions.simple_get_trans_f f d)
    | State q -> [q]
    | Var "_" ->
        failwith "Automaton:all_run: anonymous variable cannot be used on top!"
    | Var _ ->
        failwith "Automaton:all_run: free variable uninstantiated!"
          (* all_run_list recursively run through all transitions of l
             and either remove the 1st parameter of each transition if
             it is accepted, or reject the complete transition *)
          (* B: Reprendre ici !!! *)
          (* La fonction all_run_list execute tous les run possible
             sur les sous-termes en parallele *)

(* print_string ("Term list= "^(String.concat ", " (List.map Term.to_string tl))^
   "\non l=" ^ (Transitions.symextlist_to_string l)^"\n"); *)
and all_run_list (d: Transitions.t) (tl: term list) (l: external_symtrans list) : state list =
  let remain_trans =
    List.fold_left
      (fun trans subterm ->
         match subterm with
           | Var "_" -> List.map (fun (f, sl, dest) -> f, List.tl sl, dest) trans
               (* anonymous variables can match any state i.e. keep all
                  trans, only remove the 1st parameter of every trans *)
           | Var x -> failwith ("Automaton:all_run: non ground term: variable " ^ x)
               (* | State q -> Common.list_omap (fun (f, sl, dest) ->
                  let q' = List.hd sl in if q' = q ||
                  (Transitions.is_predE q q' d) then Some (f, List.tl
                  sl, dest) else None ) trans *)
               (* B: Ce cas-la n'est pas vraiment utile, il est deja
                  pris en compte dans le suivant...  Raison
                  d'optimalite ??? *)
           | t ->
               let states = all_run d t in
                 List2.optmap
	           (fun (f, sl, dest) ->
                      let q' = List.hd sl in
                        (* if q' is an Esucc of one state q of sl then it
                           is ok, i.e.  t -> Delta+E q ->E q' *)
		        if
                          List.exists
                            (fun q -> q = q' || (Transitions.is_succE d q' q))
                            states
                        then
                          Some (f, List.tl sl, dest)
		        else
                          None
                   ) trans
      ) l tl
  in
    List.map (fun (_, _, dest) -> dest) remain_trans


(*   Normalization adds transitions to delta *)

let rec normalize_term (d: Transitions.t) (t: term) (wvar: bool): (state * external_trans list)=
  match t with
    | Fsym (f, []) ->
	let relevant_trans = Transitions.get_trans_sym_constant f d in
	  (match relevant_trans with
	     | [] ->
		 let s = Transitions.make_state d in
		   Transitions.add d (f, [], s);
		   (s, [Trans(f, [], s)])
	     | (_,_,q)::_ -> (q, [])
          )
    | Fsym (f, tl) ->
	let (res_state_list, translist) = normalize_term_list d tl wvar in
	  (try
	     let relevant_trans =
               List.fold_right
                 (fun q acc -> Transitions.get_trans_sym_parameter f q d @ acc)
		 res_state_list
                 []
             in
	       try
		 let _, _, q =
                   List.find (fun (_, sl, _) -> sl = res_state_list) relevant_trans
		 in
                   (q, translist)
                     (* A transition matches, no new state is necessary *)
               with (* No transition matches, a new state and a new transition are necessary *)
		 | Not_found ->
		     let s = Transitions.make_state d in
		       Transitions.add d (f, res_state_list, s);
		       (s, (Trans (f, res_state_list, s)) :: translist)
	   with (* Transition is not fully normalizable *)
	     | Term.Not_a_state ->
                 failwith "Automaton:normalize_term: Subterms are not fully normalized!"
          )
    | State q ->
	if wvar then
          failwith "Automaton:normalize_term: cannot be a state"
	else (q, [])
    | Var x ->
	if wvar then
	  (* For variables, in the case where we normalize a lhs
	     we simply add a new constant (and state) if it is
	     not already present *)
	  let relevant_trans = Transitions.get_trans_sym_constant x d in
	    match relevant_trans with
	      | [] ->
		  let s = Transitions.make_state d in
		    Transitions.add d (x,[],s);
		    (s, [Trans(x, [], s)])
	      | (_, _, q)::_ -> (q, [])
	else
          failwith "Automaton:normalize_term: cannot be a variable!"

and normalize_term_list (d: Transitions.t) (terms: term list) (wvar: bool): state list * external_trans list =
  (* let res = List.fold_right
     (fun t acc ->
     let q, trans = normalize_term d t wvar in
     let sl, trans' = acc in
     (q::sl, trans @ trans')
     )
     terms ([], [])
     in*)
  Verbose.print 0 (fun _ -> "Warning: B: normalize_term_list (let res) vs normalize_term_list2 ne produisent pas la meme chose\n");
  normalize_term_list2 d terms wvar
    (* DBG: pour le warning ci-dessus
       if res = normalize_term_list2 d terms wvar then
       res
       else
       failwith "le fold_right et l'autre fonction ne calcule pas la meme chose..."
    *)
and normalize_term_list2 d terms wvar =
  (* B: Version originale de la fonction, c'est a dire sans le bug !!!! *)
  match terms with
    | [t] -> let (s, tl) = normalize_term d t wvar in
	([s], tl)
    | t::rem1 ->
	let (s, tl) = normalize_term d t wvar in
	let (sl2, tl2) = normalize_term_list d rem1 wvar in
	  (s::sl2,  tl @tl2)
    | [] -> failwith "Automaton:normalize_term_list: empty list!"



(* Returns a list such that its first element is the transition f(sl)->q, for
   any q, if such a transition exists in Delta. Otherwise, it searches for
   transitions f(sl')->q', for any q', such that for all state qi of sl there
   exists a state qi' in sl' (at the same rank) such that qi ->epsiE* qi'.  For
   simplicity and efficiency, if f(sl) -> q does not occur, but f(sl')->q'
   exists, we give back the first transition found and not all of them.

   YB : Attention : il me semble qu'il y a une erreur ici.
   j'ai l'impression que l'on considère comme atome (q,q') à partir du moment où nous avons q ->*E q'.
   ce qui n'est plus le cas.
*)

let closest_transitions (f: symbol) (sl: state list) (d: Transitions.t) : (external_symtrans * Formula.t) list =
  (* We check if there exists a q such that f(sl) -> q *)
  let _ = IFDEF TEST THEN
    Transitions.print_fclosure d.epsilonR
    ENDIF
  in
  let relevant_trans =
    List.flatten (List.map (fun q -> Transitions.get_trans_sym_parameter f q d) sl)
  in
    try
      (* If such a transition exists, we return it *)
      [ List.find (fun (_, sl2, _) -> sl = sl2) relevant_trans, Formula.mk_true ]
    with
      | Not_found ->
          (* Otherwise we need to check modulo epsilonE on the SAME set of transition *)
          try
            let (_, sl2, q) =
              List.find
                (fun (_, sl2, _) ->
                   try
                     List.for_all2 (fun s1 s2 -> s1 = s2 || is_succE d s1 s2) sl sl2
                   with Invalid_argument _ -> assert false
                ) relevant_trans
            in
            let phi =
              List.fold_right2
                (fun s1 s2 f -> if s1 = s2 then f else
	           let interesting = d.epsilonE#fclosure.(s2).(s1) in
	             Formula.mk_and interesting f
                ) sl sl2 Formula.mk_true
            in
              [(f, sl2, q), phi]
          with
            | Not_found ->
                (* Otherwise we check on transitions that are related
                   to states of sl by epsilonE *)
                (* we first gather all successors of states of sl *)
                let succ = List.fold_right (fun s sl -> (get_succE d s) @ sl) sl [] in
                let relevant_trans =
                  List.flatten
                    (List.map (fun q -> Transitions.get_trans_sym_parameter f q d) succ)
                in
                  try
                    let (_, sl2, q) =
                      List.find
                        (fun (_, sl2, _) ->
                           try
                             List.for_all2 (fun s1 s2 -> s1 = s2 || is_succE d s1 s2) sl sl2
                           with Invalid_argument _ -> assert false
                        ) relevant_trans
                    in
                    let phi =
                      List.fold_right2
                        (fun s1 s2 f -> if s1 = s2 then f else
                           let interesting = d.epsilonE#fclosure.(s2).(s1) in
	 	             Formula.mk_and interesting f
                        ) sl sl2 Formula.mk_true
                    in
                      [ (f, sl2, q), phi ]
	          with
                    | Not_found -> []




(* Normalize_termE adds transitions to a transition set.

   If normalization is impossible using Delta alone, we try using
   Delta ∪ epsilonE ...  and return the formula corresponding to the
   epsilonE transitions used. Such normalization cannot be applied to
   terms with variables.

   The result is a top state, a formula and the corresponding transitions.

   For simplicity, in this implementation, we only keep track of the
   first solution (except for the case where the term is a state
   (i.e. we need all solutions).  *)

let rec normalize_termE (t: term) (d: Transitions.t) :
    (state * Formula.t * external_trans list) =
  match t with
    | Fsym (f, []) ->
	let relevant_trans = Transitions.get_trans_sym_constant f d in
	  (match relevant_trans with
	     | [] ->
		 let s = Transitions.make_state d in
		   Transitions.add d (f, [], s);
		   (s, Formula.mk_true, [Trans(f, [], s)])
	     | (_, _, q)::_ -> (q, Formula.mk_true, [])
          )
    | Fsym (f, tl) ->
        (* We normalize subterms of f: tl *)
	let (res_state_list,phi,translist) = normalize_termE_list tl d in
	   (* relevant_trans is the list of the transitions closest (in term of
              number of applications of epsiE necessary) to res_state_list *)
	 let relevant_trans = closest_transitions f res_state_list d in
(*         let _ = print_string ("Closest trans for "^(Symbol.to_string f)^"("^(state_list_to_string res_state_list)^") is: "^
                               (String.concat "\n" 
                                  (List.map 
                                     (function (t,phi) -> ("("^(Transitions.symext_to_string t)^","^(Formula.to_string phi)^")")) relevant_trans))^"\n") in
*)
           (match relevant_trans with
              | [] -> (* No transition matches, a new state and a new transition are necessary *)
		  let s = Transitions.make_state d in
		    Transitions.add d (f, res_state_list, s);
		    (s, phi, (Trans (f, res_state_list, s)) :: translist)
              | ((_, sl, s), phi2) :: _ ->
                  (s, Formula.mk_and phi phi2, translist)
	   )
    | State q -> (q, Formula.mk_true, [])
    | Var x ->
        failwith "No variable in terms for normalize_termE"

and normalize_termE_list (tl: term list) (d: Transitions.t) :
    (state list * Formula.t * external_trans list) =
  match tl with
    | [t] -> let (s, phi, tl) = normalize_termE t d in
	([s], phi, tl)
    | t::rem1 ->
	let (s, phi1, tl) = normalize_termE t d in
	let (sl2, phi2, tl2) = normalize_termE_list rem1 d in
	  (s::sl2, Formula.mk_and phi1 phi2, List.append tl tl2)
    | [] -> failwith "Automaton:normalize_term_list: empty list!"



(* Normalisation of a term for adding in delta, during completion (flag variable false) *)
(* Also: normalisation of a term to produce a term_automaton: for left-hand
   side of rules, or initial term sets (flag variable true) *)

(* Add a term to a delta and returns the top state of this term
   as well as the list of new external transitions resulting in
   this adding. Uses d to normalize t and adds transitions for t to d. Also
   creates new states in d when necessary.

   The boolean wvar is a flag which is true when normalizing terms with
   variables (left-hand side of rules) *)


(* Normalisation of a term for adding in delta, using only new states! *)

let rec normalize_new_states (t: term) (d: Transitions.t) : state * external_trans list =
  match t with
    | Fsym (f, []) ->
	let s = Transitions.make_state d in
	  Transitions.add d (f,[],s);
	  (s, [Trans(f,[],s)])
    | Fsym (f, tl) ->
	let (res_state_list, translist) = normalize_new_states_list tl d in
	let s = Transitions.make_state d in
	  Transitions.add d (f, res_state_list,s);
	  (s, (Trans (f, res_state_list, s)) :: translist)
    | State q -> (q, [])
    | Var x -> failwith "Automaton:normalize_term: cannot be a variable!"

and normalize_new_states_list (tl: term list) (d: Transitions.t) :
state list * external_trans list =
    match tl with
      | [t] -> let (s, tl)= normalize_new_states t d in
	  ([s], tl)
      | t::rem1 ->
	  let (s, tl)= normalize_new_states t d in
	  let (sl2, tl2)= normalize_new_states_list rem1 d in
	    (s::sl2, List.append tl tl2)
      | [] -> failwith "Automaton:normalize_term_list: empty list!"



(* Normalisation of a term for adding in delta, during completion (flag variable
   false. This corresponds to oldstyle normalization, i.e. normalize f(t1...tn)
   ->q such that the top transition is of the form f(q1...qn) -> q.

   Normalization adds transitions to delta.
 *)

let oldnormalize_term (newstates: bool)(t: term)(dest: state)(d: Transitions.t) : external_trans list=
  match t with
    | Fsym (f, []) ->
	if Transitions.add' d (f, [], dest) then
	  [ Trans (f, [], dest) ]
	else
          []

    | Fsym (f, tl) ->
	let (res_state_list, translist) =
	  if newstates then
            normalize_new_states_list tl d
	  else
            normalize_term_list d tl false
	in
	  if Transitions.add' d (f,res_state_list,dest) then
	    Trans(f,res_state_list,dest)::translist
	  else
            translist
    | State q ->
        [ Epsilon (q, dest, Formula.mk_true) ]
    | Var x ->
        failwith "Automaton:oldnormalize_term: cannot be a variable!"


let term_to_automaton (tl: term list) (vl: variable list) (stt: s2t) : automaton * statevar * state list =
  let rec aux (tl: term list) (a: automaton) (var_assoc: statevar) (fs: state list) :
      automaton * statevar * state list =
    match tl with
      | [] -> (a, var_assoc, List.rev fs)
      | t::rem ->
	  let (final, trans) = normalize_term a.trans t true in
	    Hashtbl.add stt final t;
	    List.iter
              (function
		 | Trans (f,_,dest) ->
                     if List.mem f vl then statevar_add var_assoc dest f
		 | _ ->
                     failwith "Automaton.term_to_automaton: transitions cannot be epsilon here!"
              ) trans;
	    aux rem { trans = a.trans;
                      fstates = Stateset.add final a.fstates
                    } var_assoc (final::fs)
  in
    aux tl { trans = Transitions.make ();
             fstates = Stateset.empty
           } (new_statevar ()) []



let to_string (a: automaton) : string =
  ("States "^(state_list_to_string (Transitions.get_all_states a.trans))^"\nFinal States "^
     (state_list_to_string (Stateset.elements a.fstates))^"\nTransitions\n"^(Transitions.to_string a.trans))

let to_timbuk (a: automaton) : string =
  ("States "^(state_list_to_string (Transitions.get_all_states a.trans))^"\nFinal States "^
     (state_list_to_string (Stateset.elements a.fstates))^"\nTransitions\n"^(Transitions.to_timbuk a.trans))

IFDEF TEST THEN
let to_string_check (a: automaton) : string=
  ("States "^(state_list_to_string (Transitions.get_all_states a.trans))^"\nFinal States "^
     (state_list_to_string (Stateset.elements a.fstates))^"\nTransitions\n"^
     (*Transitions.to_string_check a.trans)*) "Transitions.to_string_check a ete remplacee par print_delta!!!!!!!")
END

(* Merges all symtables of l into acc *)

let rec merge_symtable (l: (symbol * table) list) (acc: symtable): unit =
  match l with
    | [] -> ()
    | (f, t)::rem ->
	(try
           let table = Hashtbl.find acc f in
	     Hashtbl.replace acc f (Statetable.union t table)
	 with Not_found -> Hashtbl.add acc f t
        );
        merge_symtable rem acc


(* Checks if there is at least one common symbol between list l1 and l2 *)

let rec matchone_symbol (l1: (symbol * table) list) (t2: symtable) =
  match l1 with
    | [] -> false
    | (f1, _) :: rem ->
	Hashtbl.mem t2 f1 || (matchone_symbol rem t2)


(* Filters a list of (symbol->table) s.t. the first list contains only
   constants and the second one only non constant transitions *)

let rec get_const (l: (symbol * table) list) (resconst: (symbol * table) list) (resnotconst: (symbol * table) list): (symbol * table) list * (symbol * table) list =
  match l with
    | [] -> (resconst, resnotconst)
    | (f, t)::rem ->
        if Transitions.constant_table t then
          get_const rem ((f, t)::resconst) resnotconst
        else
          get_const rem resconst ((f, t)::resnotconst)


(* Check if there is at least one common constant symbol between l1 and t2 *)
(* Symbols of l1 are only constants *)

let rec fast_const_test (l1: (symbol*table)list)(t2: symtable): bool=
  match l1 with
      [] -> false
    | (s,t)::rem ->
	(Hashtbl.mem t2 s)||(fast_const_test rem t2)

(* Ordering couples so that the smaller becomes the first one *)

let couple_order s1 s2=
  if compare s1 s2 >0 then (s2,s1)
  else (s1,s2)

(* Check or store the fact that the intersection of two states is not empty *)
let mem_couple s1 s2 t: bool=
  let (c1,c2)= couple_order s1 s2 in
    try (Stateset.mem c2 (Hashtbl.find t c1))
    with Not_found -> false

let add_couple s1 s2 t =
  let (c1, c2) = couple_order s1 s2 in
    try
      let setc1 = Hashtbl.find t c1 in
        Hashtbl.replace t c1 (Stateset.add c2 setc1)
    with Not_found ->
      Hashtbl.add t c1 (Stateset.singleton c2)


(* Get or store the FORMULA associated to the fact that the intersection of two states is not empty *)

let mem_couple_formula s1 s2 t: Formula.t=
  let (c1,c2)= couple_order s1 s2 in
    try (Hashtbl.find t (c1,c2))
    with Not_found -> Formula.mk_false

let add_couple_formula s1 s2 f t: unit =
  if Formula.unsat f then ()
  else
    let (c1, c2) = couple_order s1 s2 in
    try
      let fex = Hashtbl.find t (c1,c2) in
      (* If the existing formula is "weaker" than the new one we replace, i.e.
         A /\ B is stored and we found A (A/\B => A) ... or A is stored and we found True (A => True). *)
      if Formula.subsume fex f then
        Hashtbl.replace t (c1,c2) f        
      else ()
    with Not_found -> Hashtbl.add t (c1,c2) f



(* test of non emptyness of intersection of a list of couple of states
   of a single tree automaton  ** modulo epsilon ** *)
(* input list [(q1,q2);(q3,q4)] checks if q1/\q2 is not empty AND
  q3/\q4 not empty *)
(* Terminates also on cycling empty tree automata *)
(* Returns the formula to satisfy for the intersection not to be empty:
   - True: in this case we have a proof that whatever the approximation may be, the intersection 
   is not empty.
   - Phi different from True or False, then we know that when Phi is true the approximation is not empty...
     There may be a solution for Phi=True but it may not be already found.
   - False: the intersection is empty whatever the approximation may be. *)

let not_empty_intersection (l: (state * state) list)(delta: Transitions.t): Formula.t=
  (* table for inductive couple of states (for intersection of
     recursive transitions *)
  let inductive = Hashtbl.create Common.max_state_symb in
  (* [ok] table for saving formulas for each couple of states (s1,s2) having a non empty intersection *)
  let ok = Hashtbl.create Common.max_state_symb in
  let deltatable = delta.Transitions.transitions in
  (* Check that intersection between list of states lq1 and lq2 is not empty
     The result is the formula necessary to satisfy so that the intersection is not empty *)
  let rec check_state_list (lq1: state list) (lq2: state list) (acc: Formula.t): Formula.t =
    match lq1, lq2 with
      | [], [] -> acc
      | q1::rem1, q2::rem2 -> 
          let f= (check_state q1 q2) in
          Verbose.printf 3 "Check_state_list: Intersec: (q%i,q%i): %s\n" q1 q2 (Formula.to_string f);
          if Formula.unsat f then f
          else check_state_list rem1 rem2 (Formula.mk_and f acc)
      | _, _ ->
          failwith "Automaton:not_empty_intersection: state lists should have the same length"

  (* Check that intersection between q1 and q2 is not empty.
     The result is the formula necessary to satisfy so that the intersection is not empty *)
  and check_state q1 q2: Formula.t=
    let f= mem_couple_formula q1 q2 ok in
    if not(Formula.unsat f) then f
    else
      if mem_couple q1 q2 inductive then
      (* Couple has already been considered we are now dealing with
	 recursive transitions which are of no help for checking emptyness of an
	   intersection*)
      Formula.mk_false
        (* Otherwise, it has to be investigated more deeply *)
    else 
        if q1 = q2 then 
          begin
            add_couple_formula q1 q2 Formula.mk_true ok;
            Formula.mk_true
          end
        else
          let f1= (Transitions.is_succRE_formula delta q1 q2) in
          if Formula.is_sat f1 then
            begin
              add_couple_formula q1 q2 f1 ok;
              f1
            end
          else
            let f2= (Transitions.is_succRE_formula delta q2 q1) in
            if Formula.is_sat f2 then
              begin
                add_couple_formula q1 q2 f2 ok;
                f2
              end
            else
      (* We are considering all predecessors of q1 and q2 in order to
	 cover epsilon closure. *)
              let allq1 = (Transitions.get_predRE delta q1) in
              let allq2 = (Transitions.get_predRE delta q2) in
              let tableq1 = (Hashtbl.copy (Hashtbl.find deltatable q1)), Formula.mk_true in
              let tableq2 = (Hashtbl.copy (Hashtbl.find deltatable q2)), Formula.mk_true in
              let ltables1 =
                tableq1::
                  (List.map
	             (fun (q,phi) -> ((Hashtbl.find deltatable q)),phi)
	             allq1
                  )
              in
              let ltables2 =
                tableq2::
                  (List.map
	             (fun (q,phi) -> ((Hashtbl.find deltatable q)),phi)
	             allq2
                  )
              in
      let _ = add_couple q1 q2 inductive in
      let f= check_intersection_table_list ltables1 ltables2 in
      if Formula.is_sat f then add_couple_formula q1 q2 f ok;
      f

  (* Check that there is at least one common term recognized by a symbol table (seen as a list) [l1] and 
     table t2 *)
  and check_transition_list (l1: (symbol * table) list)(t2: symtable) : Formula.t =
    match l1 with
      | [] -> Formula.mk_false
      | (f, t) :: rem ->
	  try
            (let tab2 = Hashtbl.find t2 f in
             let t1= Statetable.elements t in
             let st2= Statetable.elements tab2 in
             let f=check_first_transition t1 st2 in
             if Formula.is_sat f then f
             else check_transition_list rem t2)
          with Not_found -> check_transition_list rem t2

  (* Check that there is at least statelist in list [l1] whose intersection is not empty w.r.t. statelist list [l2] *)
  and check_first_transition (l1: statelist list) (l2: statelist list): Formula.t=
    match l1 with 
        [] -> Formula.mk_false
      | sl::rem -> let f= check_first_transition2 sl l2 in
                   if Formula.is_sat f then f
                   else check_first_transition rem l2

  and check_first_transition2 (s1: statelist) (l2: statelist list): Formula.t=
    match l2 with 
        [] -> Formula.mk_false
      | fsl::rem -> 
          let f= check_state_list s1 fsl Formula.mk_true in
          if Formula.is_sat f then f
          else (check_first_transition2 s1 rem)

(* Check intersection of lists of sym tables. Results in the formula for the first transitions to match.
   not necessary the simplest formula (True) but the first *)

  and check_intersection_table_list (lt1: (symtable * Formula.t) list) (lt2: (symtable * Formula.t) list): Formula.t=
  let rec check_one_table (stconst: (symbol * table) list) (stnonconst: (symbol * table) list)
      (lt2: (symtable * Formula.t) list): Formula.t=
    match lt2 with
        [] -> Formula.mk_false
      | (stable,phi)::rem -> 
        (* Checks if there is at least one common symbol between list
           l1 and l2 for efficiency reasons *)
          if (matchone_symbol stconst stable) || (matchone_symbol stnonconst stable) 
          then
            (* there is at least one common symbol *)
            if (fast_const_test stconst stable) 
            (* If there is a common constant, we are done *)
            then phi
            else 
              let formula_trans_list= check_transition_list stnonconst stable in
              if Formula.is_sat formula_trans_list 
              then Formula.mk_and phi formula_trans_list
              else check_one_table stconst stnonconst rem 
          else
            check_one_table stconst stnonconst rem
  in
  match lt1 with 
      [] -> Formula.mk_false
    | (stable,phi)::rem -> 
          let (const, nonconst) = get_const (Hashtbl2.elements stable) [] [] in
          let f= (check_one_table const nonconst lt2) in
          if Formula.is_sat f then Formula.mk_and phi f
          else check_intersection_table_list rem lt2
  in
  let (lq1, lq2) = List.split l in
    check_state_list lq1 lq2 Formula.mk_true


(* utilise par inter_of' pour indiquer l'etat de traitement des couples d'etats *)
type visited = Open | Closed | Done
(* un couple d'etats est soit
   - Open : en cours de visite
   - Done : on a calcule la formule
   - Closed : on ne connait pas sa formule
              et n'est pas en cours de traitement
*)
(* Une deuxieme version sans sauvegarde des resultats... *)
let inter_of' (a1: automaton) (a2: automaton) : Formula.t =
  (* maxi is the maximal label number of a state in ai *)
  let last = a1.trans.epsilonR#last and last' = a2.trans.epsilonR#last in
    (* Les tuples (q, q', Φ) qui dénotent l'intersection a₁ × a₂
       aux états q et q', avec Φ la condition de a₁ nécessaire pour satisfaire
       la non-vacuité en (q, q'). c.f. le papier
       De plus, pour l'implémentation ne respecte pas l'algo du papier,
       le calcul est orienté des états finaux vers les co-accessibles,
       à l'inverse du papier... Normalement plus efficace, car il ne considère que
       les couples d'états co-accessibles.
    *)
  let visited = Array.make_matrix (last + 1) (last' + 1) Closed
  and phi = Array.make_matrix (last + 1) (last' + 1) Formula.mk_false
  and alpha = a1.trans.epsilonR#fclosure in

  (* Building formula for couple of states (q,q') exploring epsilonRE predecessors of q *)
  let rec phi_of_states q q' =
    Verbose.printf 3 "phi_of_states: Visiting couple of states (q%i, q%i')\n" q q';
    match visited.(q).(q') with
      | Open -> Formula.mk_false (* (q, q') deja  en cours d'exploration... *)
      | Done -> phi.(q).(q')
      | Closed ->
          (try
             (* For all predecessor q1 of q, we call phi_of_fun *)
             for q1 = 0 to last do
               Verbose.println 3 (fun _ -> "phi for "^(State.to_string q1)^" ->RE "^(State.to_string q)^"||"^(Formula.to_string alpha.(q).(q1))^"\n");
               if Formula.is_sat alpha.(q).(q1) then (* q₁ —α––>^*_RE q *)
                 let phi_q1_q' = match visited.(q1).(q') with
                   | Open -> Formula.mk_false
                   | Done -> phi.(q1).(q')
                   | Closed ->
                       visited.(q1).(q') <- Open;
                       let phi_q1_q' = phi_of_fun q1 q' in
                         visited.(q1).(q') <-
                           if Formula.tautology phi_q1_q' then Done else Closed;
                         phi_q1_q'
                 in
(*                 let _= print_string ("phi("^(State.to_string q)^")("^(State.to_string q')^"')= "^(Formula.to_string phi.(q).(q'))^"\n") in *)
                 phi.(q1).(q') <- Formula.mk_or phi.(q1).(q') phi_q1_q';
                 (* if not(Formula.subsume phi_q1_q' phi.(q1).(q')) then phi.(q1).(q') <- Formula.mk_or phi.(q1).(q') phi_q1_q'; *)
                 phi.(q).(q') <- Formula.mk_or phi.(q).(q')
                     (Formula.mk_and alpha.(q).(q1) phi_q1_q');
(*                 let _= print_string ("new phi("^(State.to_string q)^")("^(State.to_string q')^"')= "^(Formula.to_string phi.(q).(q'))^"\n") in *)
                   if Formula.tautology phi.(q).(q') then raise Break;
             done
           with Break -> phi.(q).(q') <- Formula.mk_true
          );
          Verbose.print 3 (fun _ -> "  Φ("^(State.to_string q)^", "^(State.to_string q')^")= "^(Formula.to_string phi.(q).(q'))^"\n");
          visited.(q).(q') <- Done;
          phi.(q).(q')

  (* Building formula for couple of states (q,q') exploring
     transitions f(...)->q and f(...)->q' *)
  and phi_of_fun q q' =
    Verbose.printf 3 "phi_of_fun: Visiting couple of states (q%i, q%i')\n" q q';
    try
      let st  = Hashtbl2.find a1.trans.transitions q
      and st' = Hashtbl2.find a2.trans.transitions q' in
        Verbose.print 3 (fun () -> "st= "^(Hashtbl2.to_string st));
        Verbose.print 3 (fun () -> "\nst'= "^(Hashtbl2.to_string st')^"\n");
      let phi_q_q' = ref Formula.mk_false in
        Hashtbl2.iter
          (fun f largs' ->
             try Statetable.iter
               (fun args -> Statetable.iter
                  (* we found a transition f(q1,...,qn) ->q and f(q'1, ..., q'n) ->q'
                     args= (q1,...,qn) and args'=(q'1, ..., q'n)*)
                  (fun args' ->
                     let alpha =
                       try (* α = /\₁ⁿ Φ_i pour [(q₁, q₁'); ...; (q_n, q'_n)] *)
                         List.fold_left2
                           (fun f qi qi' -> let phi_qi_qi' = phi_of_states qi qi' in
                              if not (Formula.is_sat phi_qi_qi') then raise Break;
                              Formula.mk_and phi_qi_qi' f)
                           Formula.mk_true args args'
                       with Break -> Formula.mk_false
                     in
                       phi_q_q' := Formula.mk_or !phi_q_q' alpha;
                       if Formula.tautology !phi_q_q' then raise Break;
                       (* phi_q_q' = ⊤ plus la peine de continuer *)
                  ) largs'
               ) (Hashtbl2.find st f)
             with Not_found -> () (* pas de transition de la forme q <- f(...) *)
          ) st';
        !phi_q_q'
    with
      | Break -> Formula.mk_true (* on a trouve ⊤ *)
      | Not_found -> Formula.mk_false (* st ou st' est vide donc il n'y a pas de transitions
                                         q <- f(...) ou pas de transition q' <- g(...)
                                         pour tout f ou g *)
  in
    (* We start from the couples (q,q') where q is a final state of a1 and q' a final state of q *)
    Stateset.fold
      (fun qf phi_inter ->
         Stateset.fold
         (fun qf' phi_inter ->
            Formula.mk_or (phi_of_states qf qf') phi_inter
         ) a2.fstates phi_inter
      ) a1.fstates Formula.mk_false

(* Peut etre utile pour debugguer le code de inter_of
   let str_of pref =
   List2.to_string ~left:(f ^ "(") ~right:")" ~sep:", "
   (fun q -> pref ^ (string_of_int q))
   in
   Format.printf "%s %s |- Φ = %s\n@?"
   (str_of "q" args) (str_of "q'" args') (Formula.to_string alpha);
*)




(*  0 => a raffiné;
 *  1 => pas de collision entre a et bad;
 * -1 => il existe un contre-example *)
let refine a bad =
  let phi = ref (inter_of' a bad) in
  let _= Verbose.println 3 (fun _ -> "Formula resulting of the intersection : "^
                              (Formula.to_string !phi)) in
    if not (Formula.is_sat !phi) then    (* A ∩ Bad = ∅ *)
      1                                 (* property holds *)
    else if (Formula.tautology !phi) then
      -1                                (* there exists a counter-example *)
    else (                              (* Refinement *)
      while Formula.is_sat !phi do
        Verbose.println 2 (fun _ -> "Refining Step!");
        Verbose.println 3 (fun _ -> "Refining Step: Φ = " ^ (Formula.to_string !phi));
        Transitions.prune a.trans !phi;
        (* print_string ("\n"^ (Transitions.to_string a.trans));
        print_string "\n"; *)
        phi := inter_of' a bad
      done;
      Verbose.println 3 (fun _ -> "\n" ^ (Transitions.to_string a.trans) ^ "\n");
      0;
    )

let rename (sl: state list) (q: state) (q':state)=
  if (List.mem q' sl) then q else q'

(* Applying a renaming (every state of l is renamed into q) to a list of transitions *)
let rec merge_one ((q,l): state * (state list)) (ltrans: Transitions.external_trans list)(acc: Transitions.external_trans list): Transitions.external_trans list=
  match ltrans with
    | [] -> acc
    | Trans (f,sl,q')::rem ->
        merge_one (q,l) rem
(*          (List2.add (Trans (f,List.map (rename l q) sl, (rename l q q'))) acc) *)
          ((Trans (f,List.map (rename l q) sl, (rename l q q')))::acc)
    | Epsilon (q1,q2,f)::rem ->
        (* renaming of q1 -> q2 may result in q1 -> q1 (or q2 -> q2) however this transition may be
           relevant since it may represent a rewriting step from q1 to q1. *)
        merge_one (q,l) rem
          (* (List2.add (Epsilon ((rename l q q1),(rename l q q2),f)) acc) *)
          ((Epsilon ((rename l q q1),(rename l q q2),f))::acc)

(* Applying one renaming on a state set. (every state of l is renamed into q) *)

let stateset_renaming_one ((q,l): state * (state list)) (set: Stateset.t): Stateset.t=
  List.fold_right (fun s2 set -> Stateset.replace s2 q set) l set

(* Applying a list of renaming (for every state list ls of lfusion, ls is of the
   form q::l2, where all states of l2 are renamed into q) every state of l is
   renamed into q) to a list of transitions *)

let rec merge (lfusion: state list list) (ltrans: Transitions.external_trans list): Transitions.external_trans list=
  match lfusion with
      [] -> ltrans
    | (q::sl)::rem -> merge rem (merge_one (q,sl) ltrans [])
    | _ -> failwith "Automaton.merge: badly formed connex component list"


let stateset_renaming (lfusion: state list list) (set: Stateset.t): Stateset.t=
  List.fold_right 
    (fun renaming set -> match renaming with
         [] | [_] -> failwith "Automaton.stateset_renaming: badly formed connex component list"
       | q::sl -> stateset_renaming_one (q,sl) set) lfusion set
              

(* 
   From an automaton A having epsilonR and epsilonE transitions, this function exports
   an epsilon-free tree automaton A' such that L(A)=L(A') and every possible run in A 
   is possible in A'.
   This is necessary for preserving R-closedness of tree automata.

   Export a tree automaton with epsilonE and epsilonR transitions to a tree
   automaton without epsilon transitions. We use mergin of cycles on epsilonR
   AND epsilonE, i.e. if q1 -> q2 -> q3 -> q1 then states q1, q2, and q3 are
   merged, i.e. renamed. Otherwise for all transition q1 -> q2 every transition
   c -> q1 is propagated to q2, i.e. we add c -> q2 to the tree
   automaton. Furthermore, we try to preserve R-closedness (of the automaton with
   RE-epsilon) on the tree automaton without epsilon: for all transition
   f(...q...) -> q' if q1 -> q, we add the transition f(...q1...) -> q'...
   otherwise the run f(...q1...) -> q is not possible in epsilon-free automaton 
   and may result in a tree automaton that is not R-closed.
*)


let export (a: automaton)=
  (* Les composantes connexes de <-RE*->. *)
  let cc= Transitions.get_strongly 
            (fun q q' -> (Transitions.is_succRE a.trans q q')&&(Transitions.is_succRE a.trans q' q))
            a.trans in
    Verbose.exec 3 (fun _ -> (Transitions.print_fclosure a.trans.epsilonE));
    Verbose.println 3 (fun _ -> "\nComposantes Connexes= "^
                       (List2.to_string (List2.to_string string_of_int) cc));
    (* We build the list of all states that disappear because of merging *)
    let erased_states= 
      List.fold_right (fun elt l -> 
                         match elt with
                             [] | [_] -> failwith "Bad strongly connected component"
                           | _::r -> List.rev_append r l) cc [] 
    in
      (* Fusion des classes d'équivalence sur toutes les transitions *)
    let all_merged= merge cc (Transitions.to_list a.trans) in
    let (all_sym, all_epsi)= Transitions.partition all_merged in
      (* We propagate epsilon transitions in upward direction, i.e. if f(...q1...) -> q0 and q2 -> q1, 
         we add f(...q2...) -> q0 *)
    let prop_up= Transitions.propagate_epsilon_up all_epsi all_sym in
      (* We propagate epsilon transitions in downward direction, i.e. if f(...q1...) -> q0 and q0 -> q2, 
         we add f(...q1...) -> q2 *)
    let final_prop= Transitions.propagate_epsilon prop_up erased_states a.trans in
      external_to_TA (stateset_renaming cc a.fstates) (List.rev_map (function t -> Trans t) final_prop)



(* Export a tree automaton with epsilonE and epsilonR transitions to a tree automaton with epsilon transitions 
   where all the epsilonE cycles are merged, i.e. if q1 -> q2 -> q3 -> q1 then
   states q1, q2, and q3 are merged, i.e. renamed. *)

let export_epsi (a: automaton) =
  (* Les composantes connexes de <-E*->. E seul! *)
  let cc= Transitions.get_strongly (fun q q' -> (Transitions.is_succE a.trans q q') &&
               (Transitions.is_succE a.trans q' q)) a.trans in
    Verbose.exec 3 (fun _ -> (Transitions.print_fclosure a.trans.epsilonE));
    Verbose.println 3 (fun _ -> "\nComposantes Connexes= "^
                         (List2.to_string (List2.to_string string_of_int) cc));

  (* La propagation des epsilon -RE> sur toutes les transitions de base *)
  let alltrans = Transitions.to_list a.trans in
  let all_merged = merge cc alltrans in
    external_to_TA (stateset_renaming cc a.fstates) all_merged



      (******************************************************************************)
      (*                                                                            *)
      (*                                   TESTS                                    *)
      (*                                                                            *)
      (******************************************************************************)

      IFDEF TEST THEN print_string "Automaton.ml\n\n" END

      IFDEF TEST THEN
let stt = new_s2t () and stt2 = new_s2t ()

let t1 = Fsym ("f", [ Fsym ("a", []);
                      Fsym ("g", [ Var "x"; Var "y"])
                    ])
  (* B: Cette ligne est presente deux fois... je serai tente de dire
     qu'il y a juste duplication sauf si ca cache quelque chose de plus
     tricky genre effet de bord pas decele
  *)
let (aut1, sv , _) = term_to_automaton [t1] ["x"; "y"] stt
let t2 =  Fsym ("f", [ Fsym("a", []);
                       Fsym("g", [
                              Fsym ("f", [ Fsym ("a", []); Fsym ("b", []) ]);
                              Fsym ("b", [])
                            ])
                     ])
let t3 = Fsym ("f", [ Fsym ("g", [ Fsym ("b", []) ]);
                      Fsym ("g", [ Fsym ("b", []) ])
                    ])
let t4 = Fsym ("f", [ Var "_";
                      Fsym ("g", [ Fsym ("b", []) ])
                    ])
  (* B: heu... y a 2 fois la meme ligne... effet de bord cache, ou il y
     a un truc que j'ai pas compris ???*)
let (aut1, sv ,_) = term_to_automaton [t1] ["x";"y"] stt
let (aut2, _, _) = term_to_automaton [t2] [] stt

let sv2 = Hashtbl.create Common.max_var
let _ =
  Hashtbl.add sv2 1 "x"; Hashtbl.add sv2 2 "y"

let (aut4, _, _) = term_to_automaton [t2] [] stt
let (aut5, _, ls) = term_to_automaton [t1; t2] ["x"; "y"] stt2

END

IFDEF TEST THEN
open Debug
let _ =
  print_string "Testing state to term table and printing\n";
  print_string (s2t_to_string stt2);
  asserteq (s2t_to_string stt2)
    "q8 -> f(a,g(f(a,b),b))\nq4 -> f(a,g(x,y))\n" "test 1";
  print_string ("d=" ^ Transitions.to_string Transitions.d ^ "\n\n");
  print_string "f(a, g(a,a)) -> q0 -> q1 = q2\n";
  asserteq (oriented_run Transitions.d Term.tn Transitions.s2) true "test 2";
  print_string "g(a,a) -/-> q0\n";
  asserteq (oriented_run Transitions.d Term.tm2 Transitions.s0) false "";
  print_string "det_run f(a,g(a,a)) --> f(q0,g(q0,q0))\n";
  asserteq (det_run Transitions.d Term.tn) Term.tq0 "";
  print_string "Add f(g(a,a),g(a,a)) to q1 in d\n";

  let (sl, _) = normalize_term d Term.tm false in
  let d2 = Transitions.make () in
    print_string ("New top state=" ^ (Statelist.to_string [sl]) ^ "\n\n");

    print_string ("New d=\n" ^ (Transitions.to_string d) ^ "\n");
    print_string "now, det_run f(g(a,a),g(a,a)) --> ok\n";
    asserteq (det_run Transitions.d Term.tm) (State sl) "";
    print_string ("We create a tree automaton for " ^
                    (Term.to_string Term.tmx) ^ "\n");

    let _ = normalize_term d2 Term.tmx true in
      print_string (Transitions.to_string d2);
      print_string ("########\n" ^ to_string aut1 ^ "\nshould be equal to\n" ^
                      "States q4 q3 q2 q1 q0 \nFinal States q4 \n
Transitions\ny -> q2\nx -> q1\nf(q0,q3) -> q4\na -> q0\ng(q1,q2) -> q3\n\n");

      asserteq (string_equal (to_string aut1)
                  "States q0 q1 q2 q3 q4
Final States q4
Transitions
f(q0,q3) -> q4
g(q1,q2) -> q3
y -> q2
x -> q1
a -> q0") true "Equality of tree automata produced by normalization";

      asserteq (Hashtbl2.equal_level1 (=) sv sv2) true "statevar tables"
END


IFDEF TEST THEN
let aut10 =
  { trans = Transitions.make ();
    fstates = Stateset.empty
  }
let s0 = make_state aut10.trans and s1 = make_state aut10.trans
and s2 = make_state aut10.trans and s3 = make_state aut10.trans
and s4 = make_state aut10.trans and s5 = make_state aut10.trans
and s6 = make_state aut10.trans and s7 = make_state aut10.trans
and s8 = make_state aut10.trans 
  (* In the following automaton f(g(a), g(a)) is in the intersection
     between states q0 and q4

     f(q0,q0) -> q0
     g(q1) -> q0
     b -> q1
     a -> q3
     q3 -> q1 Eq(5,6)
     g(q6) -> q6

     f(q6, q5) -> q4
     g(q4) -> q6
     g(q4) -> q5
     g(q7) -> q7
     a -> q4
     b -> q4 *)

let _=
  Transitions.add aut10.trans ("g",[s3],s8);
  Transitions.add aut10.trans ("f",[s0;s0],s0);
  Transitions.add aut10.trans ("g",[s1],s0);
  Transitions.add aut10.trans ("g",[s6],s6);
  Transitions.add aut10.trans ("b",[],s1);
  Transitions.add aut10.trans ("a",[],s3);
  Transitions.addR aut10.trans (s3,s1,Formula.mk_atom s6 s5 Eq);
  Transitions.add aut10.trans ("f",[s6;s5],s4);
  Transitions.add aut10.trans ("g",[s4],s6);
  Transitions.add aut10.trans ("g",[s4],s5);
  Transitions.add aut10.trans ("g",[s7],s7);
  Transitions.add aut10.trans ("a",[],s4);
  Transitions.add aut10.trans ("b",[],s4)

let ind = Hashtbl.create Common.max_state_symb

ENDIF


IFDEF TEST THEN
begin
  print_string "Test all_run\n";
  print_string ("On aut10=" ^ (Transitions.to_string aut10.trans) ^ "\n");
  (*
     f(g(b),g(b)) -> f(g(q1),g(q1)) -> f(q0,q0) -> q0
     -> f(g(q4),g(q4)) -> f(q6,q5) -> q4
  *)
  asserteq (all_run aut10.trans t3) [0; 4] "f(g(b),g(b)) is recognized by two states q0 and q4";
  asserteq (all_run aut10.trans t4) [0; 4] "f(_,g(b)) is recognized by two states q0 and q4";
  print_string "Test not_empty_intersection\n";
  asserteq (mem_couple 1 2 ind) false "1,2 not present in empty table";
  add_couple 1 2 ind;
  add_couple 3 1 ind;
  asserteq (mem_couple 2 1 ind) true "1,2 with order";
  asserteq (mem_couple 1 2 ind) true "1,2 without order";
  asserteq (mem_couple 1 3 ind) true "1,3 without order";
  asserteq (mem_couple 3 1 ind) true "1,3 with order"; 
  print_string "Test intersection results:\n";
  print_string ("0,4:"^(Formula.to_string (not_empty_intersection [(s0, s4)] aut10.trans))^"\n");
  print_string ("5,6:"^(Formula.to_string (not_empty_intersection [(s5, s6)] aut10.trans))^"\n");
  print_string ("0,0:"^(Formula.to_string (not_empty_intersection [(s0, s0)] aut10.trans))^"\n");
  print_string ("6,0:"^(Formula.to_string (not_empty_intersection [(s6, s0)] aut10.trans))^"\n");
  print_string ("4,1:"^(Formula.to_string (not_empty_intersection [(s4, s1)] aut10.trans))^"\n");
  print_string ("8,0:"^(Formula.to_string (not_empty_intersection [(s8, s0)] aut10.trans))^"\n");
  asserteq (not_empty_intersection [(s0, s4)] aut10.trans) Formula.mk_true "intersection not empty";
  asserteq (not_empty_intersection [(s4, s0)] aut10.trans) Formula.mk_true "should work in both directions"; 

  let form= Formula.mk_atom s6 s5 Eq and form2= (not_empty_intersection [(s8, s0)] aut10.trans) in
  asserteq (Formula.equival form form2) true "intersection is not empty if formula Eq q6 q5 is true";

  asserteq (not_empty_intersection [(s1, s5)] aut10.trans) Formula.mk_false "intersection empty";
  asserteq (not_empty_intersection [(s5, s1)] aut10.trans) Formula.mk_false "other direction";
  asserteq (not_empty_intersection [(s6, s7)] aut10.trans) Formula.mk_false "check termination of intersection on cycling empty langages";
  (* in aut10 we have f(q6,q5) -> q4 and g(q7) -> q7. We add q7->q5 and try
     to normalize modulo E f(q6,g(q7)) ... it sould be normalized into q4 with
     formula q7->q5 *)
  let aut11 = copy aut10 in
    Transitions.addE aut11.trans s7 s5;
    (* transitive_Rclosure aut11.trans; *)
    print_string "Test closest_transitions\n";
    asserteq (closest_transitions "g" [s7] aut11.trans) [(("g", [s7], s7), Formula.mk_true)] "closest_trans 1";
    Transitions.print_delta aut11.trans;
    asserteq (closest_transitions "f" [s6;s7] aut11.trans) [(("f",[s6;s5],s4),Formula.mk_atom s7 s5 Eq)] "closest_trans 2";
    print_string "Test normalize_termE\n";
    let t11 =
      Fsym ("f", [ State s6; Fsym ("g", [State s7]) ])
    in
      asserteq (normalize_termE t11 aut11.trans) (s4, Formula.mk_atom s7 s5 Eq, []) "normalize_termE 1"
end

END

IFDEF TEST THEN
print_string "Test automaton.ml : inter_of /0";
  (* On construit
     L(a₁) = { f(a, b); f(b, a); f(a, a); f(b, b) }
  *)
  let delta = Transitions.make () in
  let q0 = Transitions.make_state delta and q1 = Transitions.make_state delta
                                        and q2 = Transitions.make_state delta
  in
  let a1 = { trans = delta; fstates = Stateset.singleton q0 }
  and _ =
    Transitions.add delta ("f", [q1; q2], q0);  (* f(q1, q2) -> q0      *)
    Transitions.add delta ("a", [], q1);        (* a -> q1              *)
    Transitions.add delta ("b", [] ,q2);        (* b -> q2              *)
    Transitions.addE delta q1 q2;
    Transitions.addE delta q2 q1                (* q₁ <——–> q₂ ∈ EpsE   *)
  in
  print_string ("A1 = \n" ^ (Transitions.to_string delta));
  print_string "\n";
  (* On construit :
     L(a₂) = { f(b, b) }
  *)
  let delta2 = Transitions.make () in
  let q0' = Transitions.make_state delta2 and q1' = Transitions.make_state delta2 in
  let a2 = { trans = delta2; fstates = Stateset.singleton q0' } in
  let _ =
    Transitions.add delta2 ("f", [q1'; q1'], q0');    (* f(q1', q1') -> q0' *)
    Transitions.add delta2 ("b", [], q1')             (* b -> q1' *)
  in
  print_string ("A2 = \n" ^ (Transitions.to_string delta2));
  print_string "\n";
  let phi = inter_of' a1 a2 in
    asserteq phi (Formula.mk_atom q2 q1 Eq) "correct formula";
    Transitions.prune a1.trans phi;
    print_string ("\n" ^ (Transitions.to_string delta));
    print_string "\n";
      let phi = inter_of' a1 a2 in
        Format.printf "Φ = %s\n@?" (Formula.to_string phi);

print_string "Test automaton.ml : inter_of /1\n";
(* On complète a1:
   L(a₁) = { f(a, b); f(a, a); g(a, a); g(b, a) }
*)
let q3 = Transitions.make_state delta in
let _ =
  Transitions.add delta ("g", [q2; q1], q3);            (* g(q2, q1) ––> q3      *)
  Transitions.addR delta (q3, q0, Formula.mk_true)      (* q3 —⊤––> q0 ∈ EpsR   *)
in
(* Declaration de l'automate a2:
   L(a₂) = { g(b, a) }
*)
let delta2 = Transitions.make () in
let q0' = Transitions.make_state delta2 and q1' = Transitions.make_state delta2
                                        and q2' = Transitions.make_state delta2
in
let _ =
  Transitions.add delta2 ("g", [q1'; q2'], q0');         (* g(q1',q2') —> q0'   *)
  Transitions.add delta2 ("b", [], q1');                 (* b —> q1'            *)
  Transitions.add delta2 ("a", [], q2');                 (* a —> q2'            *)
in
let a2 = { trans = delta2; fstates = Stateset.singleton q0' } in
  print_string "##################################################\n";
  print_string ("A1 =\n" ^ (Transitions.to_string delta));
  print_string "##################################################\n";
  print_string ("A2 =\n" ^ (Transitions.to_string delta2));
  print_string "\n";
  let phi = inter_of' a1 a2 in
    asserteq phi Formula.mk_true "correct formula";
    Format.printf "Obtained: %s\n@?" (Formula.to_string phi)
END

IFDEF TEST THEN
print_string "Test stateset_renaming\n";
let delta= Transitions.make () in 
let q0 = Transitions.make_state delta and q1 = Transitions.make_state delta and
    q2 = Transitions.make_state delta and q3 = Transitions.make_state delta in
let set1= Stateset.add_list [q0;q1;q2;q3] Stateset.empty and set2=Stateset.add_list [q1;q2] Stateset.empty in
let set3= stateset_renaming [[q1;q0];[q2;q3]] set1 in
  begin
    print_string ("set1="^(Stateset.to_string set1)^"\n");
    print_string ("set2="^(Stateset.to_string set2)^"\n");
    print_string ("set3="^(Stateset.to_string set3)^"\n");
    asserteq set3 set2 "stateset_renaming renames according to a list of renaming"
  end

END

IFDEF TEST THEN
print_string "Test automaton.ml : Papier : inter_of /10";
(* Test sur les transitions issues de l'approximation de la première étape de complétion*)

let delta = Transitions.make () in
let q0 = Transitions.make_state delta and q1 = Transitions.make_state delta and
    q2 = Transitions.make_state delta and q3 = Transitions.make_state delta and
    q4 = Transitions.make_state delta and q5 = Transitions.make_state delta and
    q6 = Transitions.make_state delta and q7 = Transitions.make_state delta in
let _ =
  Transitions.add delta ("f", [q1], q0);                (* f(q1) —–> q0 *)
  Transitions.add delta ("O", [],   q1);                (*     O —–> q1 *)
  Transitions.add delta ("s", [q1], q2);                (* s(q1) —–> q2 *)
  Transitions.add delta ("s", [q2], q3);                (* s(q2) —–> q3 *)
  Transitions.add delta ("f", [q3], q4);                (* f(q3) —–> q4 *)
  Transitions.add delta ("s", [q3], q5);                (* s(q3) —–> q5 *)
  Transitions.add delta ("s", [q5], q6);                (* s(q5) —–> q6 *)
  Transitions.add delta ("f", [q6], q7);                (* f(q6) —–> q7 *)

  Transitions.addR delta (q4, q0, Formula.mk_true);     (*    q4 —–> q0 ∈ EpsR *)
  Transitions.addR delta (q7, q4, Formula.mk_true);     (*    q7 —–> q4 ∈ EpsR *)

  Transitions.addE delta q2 q3;
  Transitions.addE delta q3 q2;
  Transitions.addE delta q3 q5;
  Transitions.addE delta q5 q3;
  Transitions.addE delta q6 q5;
  Transitions.addE delta q5 q6
in
(* Declaration de l'automate Propriété *)
let delta2 = Transitions.make () in
let q0' = Transitions.make_state delta2 and q1' = Transitions.make_state delta2
                                        and q2' = Transitions.make_state delta2 in

  Transitions.add delta2 ("f", [q1'], q0');         (* f(q1') —–> q0' *)
  Transitions.add delta2 ("O", [],    q2');         (*      O —–> q2' *)
  Transitions.add delta2 ("s", [q2'], q1');         (* s(q2') —–> q1' *)
  Transitions.add delta2 ("s", [q1'], q2');         (* s(q1') —–> q2' *)
  print_string "##################################################\n";
  print_string ("A2=\n" ^ (Transitions.to_string delta2) );
  print_string "\n";
  print_string "##################################################\n";
  print_string ("A1=\n" ^ (Transitions.to_string delta));
  print_string "\n";
  let phi = inter_of'
    { trans = delta;  fstates = Stateset.singleton q0  }
    { trans = delta2; fstates = Stateset.singleton q0' }
  in
    Format.printf "Obtained: %s\n@?" (Formula.to_string phi)
END

