IFDEF TEST THEN
open Debug
END
open Common
open Transitions
open Term
open Substitution
open SubstitutionPhi
open Automaton
open Rewriting
open Equation
open Definitions

type substitution = Substitution.t
type substitutionset = SubstitutionSet.t
type substable = SubstitutionTable.t

(* condition_tables map states of lhs of conditional rule to sets of
   qredex states we reconsider every conditional rule at every
   completion step
*)
type condition_table = (state, Stateset.t) Hashtbl.t

let new_condition_table () =
  Hashtbl.create Common.max_condition

let condition_table_to_string =
  Hashtbl2.to_string ~sep:"\n" ~left:"" ~right:""
    ~format:
    (fun _ _ k e -> (State.to_string k) ^ " -> " ^ (Stateset.to_string e))



(* in exteq_table each qlhs of a pattern is associated to either
   Unique (the pattern is alone [f(x)] -> ... or Equal like [f(x)=g(y)] or
   Any like [f(x),g(x)] in that case the state gives the qrhs. The
   term*term list is the list of merging equations *)

type exteq =
  | Unique of (term * term) list
  | Equal of state * (term * term) list
  | Any of state * (term * term) list


(* Propagation of new substitutions should apply modulo epsilon *)
(* Propagation of epsilon should apply on substitutions! and propagate
   upwards to final states of rules (if possible) *)
(* For instance if (qg,q2) {{x->q1}} and q2 -> q3 by epsilon then
   (qg, q3) should contain at least {{x->q1}} *)


(** Returns the set of substitutions corresponding to the matching of
    state [q1] on [q2] with substable [s] and variable list [vl]
    (states recognizing variables in lhs)
*)
let match_one (q1: state) (q2: state) (s: substable) (vl: statevar): substitutionset =
  try (* If q1 is a variable *)
    let varname = Automaton.statevar_find vl q1 in
      SubstitutionSet.singleton
        (SubstitutionPhi.addBoth varname (State q2) Formula.mk_true SubstitutionPhi.empty)
        (* q1 is a regular state... so we look for solutions of (q1,q2) in s*)
  with
    | Not_found ->
        try SubstitutionTable.find (q1, q2) s with
            (* Otherwise the set of solution is empty *)
          | SubstitutionTable.Couple_not_found -> SubstitutionSet.empty


(** Returns for a list of transitions of lhs and a new trans the list
    association [(q1, q2) -> new] set of substitutions to be added to (q1, q2).
    vl is the variable (given as states) list *)
let rec filter_set
    (l: external_symtrans list) ((f, sl, dest): external_symtrans)
    (s: substable) (vl: statevar) (res: ((state * state) * substitutionset) list)
    : ((state * state) * substitutionset) list =
  match l with
    | [] -> res
    | (_, sl2, dest2)::rem ->
	let value_list =
	  List.map2 (fun q1 q2 -> match_one q1 q2 s vl) sl2 sl
	in
	let new_subst =
          ((dest2, dest), SubstitutionSet.combine value_list)
       
(*      in
       let _= print_string ("Matching "^(Transitions.symextlist_to_string [(f,sl2,dest2)])^" on "^
                             (Transitions.symextlist_to_string [(f,sl,dest)])^" and result is \n"^
                             "value list="^(SubstitutionSet.list_to_string value_list)^"\n"^
                             "("^(State.to_string dest2)^","^(State.to_string dest)^") -> "^
			(SubstitutionSet.to_string (SubstitutionSet.combine value_list))^"\n")
*)
        in
	  filter_set rem (f, sl, dest) s vl (new_subst::res)



(** Filtering function for a term automaton lhs (gathering all
    left-hand sides of rules) and a tree automaton a representing a
    regular language of terms. Does not compute the product
    automaton. vl is the list of (state * variable) variables in
    lhs. nt is the list of new transitions. ns is the list of new
    substitution by couple of states produced by nt. *)

(* Attention aux constantes ex: mb gauche g(a) ... subst vide *)

(* This function does not propagate new substitutions to fathers etc. *)
(* This function does not modify s!!! *)
(* This function only cares for producing new substitutions w.r.t. matching of lhs on Delta
   (no epsilon resolving), i.e. substitution local to states of Delta. Propagation w.r.t. epsilon   transitions is done afterwards by "propagate" functions *)

(** The iseq flag is there only to use unoptimized new substitution production in case
   of equations: a transition may need to be reconsidered even if corresponding substitutions
   are already in the substable because they may be present there because of merging.
   They do not need to be reconsidered for TRS because merging of substitutions also
   applies on rhs hence new solutions, but it is necessary for equation because it may lead
   to new equation applications. See TU-ex7.txt.
*)

let rec filter
    (lhs: automaton) (nt: external_symtrans list)
    (ns: ((state * state) * substitutionset) list)
    (s: substable) (vl: statevar) (iseq: bool)
    : ((state * state) * substitutionset) list =
  match nt with
    | [] -> ns
    | (f, [], dest)::rem ->
	(* Constant symbols *)
	(* There should be at most one transition of the form a->qa in
	   the lhs.  not modulo epsilon!  *)
	let destlhs = Transitions.get_trans_sym_constant f lhs.trans in
	  (match destlhs with
	     | [] -> (* There is no transitions labelled by f in lhs *)
		 filter lhs rem ns s vl iseq
	     | [ (_, [], dest2) ] ->
		 (* There is exactly one transition labelled by f in lhs *)
		 if
                   SubstitutionTable.mem (dest2,dest) (SubstitutionPhi.addPhi Formula.mk_true SubstitutionPhi.empty ) s
		 then
		   (* Found substitution is not new *)
      		   filter lhs rem ns s vl iseq
		 else (* Found substitution is new *)
		   filter lhs rem
		     (
                       ( (dest2, dest),
                         (SubstitutionSet.singleton
			    (SubstitutionPhi.addPhi Formula.mk_true SubstitutionPhi.empty))
                       ) :: ns
                     ) s vl iseq
	     | _ -> (* Any other case is impossible *)
		 failwith "Completion:filter:malformed lhs constant automaton"
          )
    | (f, sl, dest)::rem -> (* General transitions *)
        (* Get all the transitions of lhs with f as top symbol. Not modulo
           epsilon *)
        let all_ftrans = Transitions.simple_get_trans_f f lhs.trans in

        (* Get all the substitutions produced by the new transitions *)
        let new_substs = filter_set all_ftrans (f, sl, dest) s vl [] in
        let _ = Verbose.println 3
          (fun _ -> "filter: new_substs= " ^
             (SubstitutionTable.list_to_string new_substs) ^ "\n"
          )
        in
        (* Filter-out those who are already covered by s, for each es
           (set of substitution) retrieve in es all the substs that
           are already in s (done by find and diff *)

        (* Necessary to reconsider all substitutions for filtering correctly
           on fixpoint_merging for equations hence we remove the optimization: *)
        let new_not_covered =
          if iseq then
            new_substs
          else
            List.fold_left
	      (fun l (couple, es) ->
	         let newset =
	           try
                     SubstitutionSet.diff es (SubstitutionTable.find couple s)
                   with
	             | SubstitutionTable.Couple_not_found -> es
                 in
	           if SubstitutionSet.is_empty newset then l else
                     (couple, newset)::l

	      ) [] new_substs 
        in
        let _=
          Verbose.println 3 
            (fun _ -> "filter: new_not_covered= "^
               (SubstitutionTable.list_to_string new_not_covered)^"\n")
        in
          filter lhs rem (new_not_covered @ ns) s vl iseq



(** Filter_epsi propagate the q1 -> q2 transitions. This is necessary
    when an epsilon transition q1 -> q2 appears some steps after that
    q1 has been filled. Thanks to this epsilon transition a NEW
    substitution may occur in q2 and may thus be declared.  It will be
    propagated further into by propagation if q2 ->* q3 ... note that
    adding q1 -> q2 and q2 -> q3 in ne won't propagate substs of q1 to
    q3... This cannot occur anyway by a single step of completion!!
*)

(* This function does not propagate new substitutions to fathers etc. *)
(* This function does not modify s!!! *)
let rec filter_epsi
    (ne: external_epsilon list)
    (ns: ((state * state) * substitutionset) list)
    (s: substable) : ((state * state) * substitutionset) list =
  match ne with
    | [] -> ns
    | (q1, q2, f)::rem ->
	(* We add all substs of q1 to q2 with formula f *)
        if f = Formula.mk_false then (* should not occur, but... *)
          filter_epsi rem ns s
        else
	  let res =
	    try
              let all_new_sub = Hashtbl2.elements (Hashtbl.find s q1) in
                (* all_new_sub is a list of couples (q, es) where q is a state
                   of the lhs automaton and es is a substitution set associating
                   to each substitution a formula *)
		List.fold_left
                  (fun acc (q, es) ->
		     let new_es =
                       SubstitutionTable.add_set
                         (SubstitutionSet.addPhi_to_all f es) (q, q2) s
                     in
		       if SubstitutionSet.is_empty new_es then
                         acc
		       else
                         ((q, q2), new_es)::acc
                  ) ns all_new_sub
	    with
              | Not_found -> ns
          in
	    filter_epsi rem res s

(* Pour Propagate: Hypothese faite ici l'ordre dans lequel les external trans
   arrivent n'importent pas pour la coherence des substitutions calculees par
   intersection locale puis propagation.
   Ex:

   f(qg, qf2) -> qf1		f(q1, q3) -> q0
   g(qvg1) -> qg                g(q2) -> q1
   f(qvf1, qvf2) -> qf2		f(q5, q6) -> q3

   (qg,q1)={{qvg1->q2}}

   On suppose que new1=f(q1,q3)->q0 et new2=f(q5,q6) -> q3

   Si on ajoute new1 avant new2: (qf,q0) reste vide (pas de subst possible), et
   en ajoutant new2 + propage on obtient: (qf2,q3) = {{qvf1->q5, qvf2->q6}} et en
   propageant au dessus: (qf,q0)={{qvg1->q2, qvf1->q5, qvf2->q6}}

   Ce qui est identique a ce que l'on obtiendrait par l'operation inverse: new2 avant
   new1.

   par contre, il faut garder en tete que les nouvelles epsilon, ex: q1 -> q2
   peuvent ajouter des substitutions a q2 (ET TOUS SES DESCENDANTS!) et il faut
   repropager a partir de ces nouvelles substitutions. *)


let match_in_list l1 l2 e1 e2 : bool =
  try
    List.exists2 (fun elt1 elt2 -> elt1 = e1 && elt2 = e2) l1 l2
  with Invalid_argument _ ->
    failwith "Completion:match_in_list: lists do not have the same size!"

(* For a given symbol f, destination states qlhsdest (lhs) and qdest (redex),
   corresponding tables tlhs and t2 (redex), and parameter states qlhs and q (redex)
   this function produces the list of external transitions such that at least one
   position of qlhs corresponds to one position of q *)
let intersec_by_table
    (f: symbol) (qlhsdest: state) (tlhs: table) (qlhs: state)
    (q: state) (qdest: state) (t2: table)
    (res: (external_symtrans * external_symtrans) list)
    : (external_symtrans * external_symtrans) list =
  Statetable.fold
    (fun elt1 l1 ->
       Statetable.fold
	 (fun elt2 l2 ->
	    if match_in_list elt1 elt2 qlhs q then
	      ((f, elt1, qlhsdest), (f, elt2, qdest))::l2
	    else
              l2
         ) t2 l1
    ) tlhs res

let rec intersec_by_table_list f qlhsdest t qlhs q lred res =
  match lred with
    | [] -> res
    | (qdest, t2)::rem ->
	intersec_by_table_list
          f qlhsdest t qlhs q rem
          (intersec_by_table f qlhsdest t qlhs q qdest t2 res)


(** intersec_by_dest computes the list of couples of transition
    lhs * redex where the top symbol is f and the position of qlhs in lhs
    trans is the same a q in redex trans *)
let rec intersec_by_dest
    (f: symbol) (qlhs: state) (q: state)
    (llhs: (state * table) list) (lred: (state*table) list)
    (res: (external_symtrans * external_symtrans) list)
    : (external_symtrans * external_symtrans) list =
  match llhs with
    | [] -> res
    | (q2, t)::rem ->
	intersec_by_dest
          f qlhs q rem lred
          (intersec_by_table_list f q2 t qlhs q lred res)


(** intersec computes the list of couples of transition lhs*redex where the
    top symbol is the same and the position of qlhs in lhs trans is the
    same a q in redex trans *)
let rec intersec
    (qlhs: state) (q: state)
    (llhs: (symbol * (state, table) Hashtbl.t) list)
    (sd: satable)
    (res: (external_symtrans * external_symtrans) list)
    : (external_symtrans * external_symtrans) list =
  match llhs with
    | [] -> res
    | (f, t)::rem ->
	try
	  let lred_by_dest =
            Hashtbl2.elements (Hashtbl.find (Hashtbl.find sd q) f)
          and llhs_by_dest =
            Hashtbl2.elements t
          in
	    intersec
              qlhs q rem sd
              (intersec_by_dest f qlhs q llhs_by_dest lred_by_dest res)
        with
          | Not_found -> intersec qlhs q rem sd res


(* Propagation of new substitution w.r.t. list of couples of transitions *)

(** For propagation on father transitions, it is enough to propagate new
    substitutions of (qlhs,q) with old substitution of other states.
    It is possible that several occurences of (qlhs,q) occur in ls1 ls2 but
    since rules are linear qlhs cannot contain a variable and thus it it not
    necessry to combine old solutions of (qlhs,q) with new ones (it can only
    be the trivial substitution. *)
let rec propagate_one_layer
    (linter: (external_symtrans * external_symtrans) list) (qlhs: state)
    (q: state) (es: substitutionset) (s: substable) (vl: statevar)
    (res: ((state * state) * substitutionset) list)
    : ((state * state) * substitutionset) list =
  match linter with
    | [] -> res
    | ((f, ls1, dest1), (_, ls2, dest2))::rem ->
	let newsub =
          SubstitutionSet.combine
            (List.map2
	       (fun qlhs2 ql2 ->
		  if qlhs2 = qlhs && ql2 = q then
		    es
		  else
                    match_one qlhs2 ql2 s vl
               ) ls1 ls2
            )
	in
	  if SubstitutionSet.is_empty newsub then
	    (* One couple of states as no entry in s, and thus
	       the list of subst is empty in that case*)
	    propagate_one_layer rem qlhs q es s vl res
	  else
	    propagate_one_layer rem qlhs q es s vl (((dest1, dest2), newsub)::res)




(** propagation of a new substition w.r.t. a substable. New substitutions gives
   all the points where there are new values i.e. (q1,q2) -> {{x -> q1}} This has
   to be propagated for every Fathers of (q1,q2) w.r.t. the current values of
   brothers of (q1,q2). For instance if in a left-hand side of the rule we have:
   f(q1, q3) -> q0, there is a new value on couple (q1,q4) hence, one have to
   find the possible values for (q3,?) and propagate the combination of
   substitutions to (q0,?) etc... The result is the list of couples (q1,q2)
   where q1 is a final state of a lhs. Final is the set of final states of the
   lhs. Note that there may be a father to final states in automata for the lhs
   (but no cycles in the transition relation).

   We also propagate new substitutions to successors modulo epsilon of
   states. However, if (qr,q1) is new and q1->epsilon*q2 then (qr,q2) is
   new but it is not added to the set of results as is (otherwise we
   would not share the fact that successors w.r.t. R are also successors of
   q2. We only propagate this "new" information to fathers of q2... those
   may also be added to the result if they provide a new substitution based
   on (qr,q2)... Except for interactive pattern detection where this information has
   to be propagated on every epsilon transition so as to build ALL the possible matches.

   qlhs: state of the lhs automaton alhs to consider
   q: state of the Aredex automaton a to consider
   es: state of new substitutions occurring for (qlhs,q)
   s: the current substitution table
   successor: is the current state a successor of another state
   vl: the list of states of alhs representing variables
   newsub: the list of new substitutions to consider

*)

let rec bottomup_propagate
    (qlhs: state) (q: state) (es: substitutionset) (alhs: automaton)
    (a: automaton) (s: substable) (successor: bool) (vl: statevar)
    (iseq: bool) (newsub: ((state * state) * substitutionset) list)
    : ((state * state) * substitutionset) list =
  try
    let satablelhs = Hashtbl.find alhs.trans.satable qlhs in
    let llhs = Hashtbl2.elements satablelhs in
    let satablered = a.trans.satable in
      (* List of couples of transition lhs*redex where the top symbol is the
	 same and the position of qlhs in lhs trans is the same a q in redex
	 trans *)
    let linter: (external_symtrans * external_symtrans) list =
      intersec qlhs q llhs satablered []
    in
    let new_combined_subst: ((state * state) * substitutionset) list =
      propagate_one_layer linter qlhs q es s vl []
    in

    (* Add the substitution corresponding to (qlhs,q) to s and
       get_back the new elements *)
    (* Necessary to reconsider all substitutions for filtering
       correctly on fixpoint_merging hence we remove the optimization
       for equations *)
    let newes =
      if iseq then (
        ignore (SubstitutionTable.add_set es (qlhs, q) s);
        es
      ) else
        SubstitutionTable.add_set es (qlhs, q) s
    in

    let newnewsub =
      (* qlhs is a final state and q is not considered as a successor of another
	 q' then (qlhs,q)->es is a new set of subst for final states and if
	 it contains at least a new susbt (newes contains only new subst) *)
      if Stateset.mem qlhs alhs.fstates && not successor &&
        not (SubstitutionSet.is_empty newes) then
        ((qlhs, q), newes)::newsub
      else
        newsub
    in
      breadth_propagate new_combined_subst alhs a s vl iseq newnewsub
  with Not_found -> (* Cannot go up, qlhs is a final state *)
    (* Necessary to reconsider all substitutions
       for filtering correctly on fixpoint_merging
       hence we remove the optimization for equations.
    *)
    let newes =
      if iseq then (
        ignore (SubstitutionTable.add_set es (qlhs,q) s);
        es
      ) else
	SubstitutionTable.add_set es (qlhs,q) s
    in
      if (Stateset.mem qlhs alhs.fstates) && (not successor) then
	(* qlhs is a final state then (qlhs,q)->es is a new set of subst for
	   final states and if it contains at least a new susbt (newes contains
	   only new subst) *)
        ((qlhs, q), newes) :: newsub
      else
        newsub


(** propagation of a list of new substition w.r.t. a substable. New substitutions
    gives all the points in the intersection between alhs and aredex where there
    are new values i.e. (q1,q2) -> {{x -> q1}} *)
and breadth_propagate
    (ls: ((state * state) * substitutionset) list) (alhs: automaton)
    (a: automaton) (s: substable) (vl: statevar)(iseq: bool)
    (newsub: ((state * state) * substitutionset) list)
    : ((state * state) * substitutionset) list =
  match ls with
    | [] -> newsub
    | ((q1, q2), es) :: rem ->
	(* if there is no father to q1 (q1 is the top of a rule) then we do not
	   propagate new substitutions to successors (the resulting values will
	   be implicitely propagated by epsilon closure)*)
	(* Find lhs transitions concerned by new values in q1 *)
	(* note that new in (q1,q2) => new in (q1,q3) if q1 ->epsilon* q3 *)
	let all_succ = Transitions.get_succRE a.trans q2 in
	  (* We apply bottomup_propagate to q2, where q2 is not considered
	     as a successor of himself *)
	let first = bottomup_propagate q1 q2 es alhs a s false vl iseq newsub in
	  (* We apply bottomup_propagate to all successors of q2 (with
	     a shared list of new substitutions (fold). Only if we are not
	     in the case of an equation! But it is applied if we are
	     in noepsi mode! *)
	let bup =
          if iseq && not(!completion_type = Noepsi) then
            first
          else
	    List.fold_left
	      (fun sol (state, formula) ->
	         bottomup_propagate
                   q1 state
                   (SubstitutionSet.addPhi_to_all formula es)
                   alhs a s true vl iseq sol
              )
	      first
	      all_succ
        in
	  breadth_propagate rem alhs a s vl iseq bup


(** returns the list of new substitutions where the first state
   corresponds to the lhs of the rule (resp. lhs or rhs of equation)
   and the second state corresponds to the redex tree
   automaton. Breadth_propagates adds new substitutions in the
   substable.

   iseq flag is to switch on/off the fact that ne (epsilon) have to be
   taken into account to propagate the substitution or not. True
   (matching for equations) not taken into account, false (regular
   matching for lhs of rules for instance): taken into account.  This
   flag is also set to false for matching equations when ne only
   contains epsilonE rules.

   Note that adding q1 -> q2 and q2 -> q3 in ne, at the same time,
   won't propagate substs of q1 to q3... This cannot occur anyway by a
   single step of completion!!
*)

let complete_filter
    (lhs: automaton) (aredex: automaton) (nt: external_symtrans list)
    (ne: external_epsilon list)(s: substable) (vl: statevar)(iseq: bool)
    : ((state * state) * substitutionset) list =

  (* filter and filter_epsi functions do not modify the substitution table s *)
  let new_sub = filter lhs nt [] s vl iseq in  (* new sub for new normal trans *)
    Verbose.println 3
      (fun _ -> "Complete filter: new_sub= " ^ (SubstitutionTable.list_to_string new_sub));
    (* We do not propagate substs using epsilon for equations *)
    (* new sub for new epsi trans *)
    let new_sub2 =
      if iseq && not(!completion_type = Noepsi) then
        new_sub
      else
        filter_epsi ne new_sub s
    in
      Verbose.println 3
        (fun _ -> "Complete filter: new_sub2= " ^ (SubstitutionTable.list_to_string new_sub2));
(*      let _ = print_string ("For ne="^(Transitions.epsiextlist_to_string ne)^"\n new_sub2= "^
                            (SubstitutionTable.list_to_string new_sub2)^"\n s="^(SubstitutionTable.to_string s)^"\n") in *)

      (* modification is performed by breadth_propagate *)
      breadth_propagate new_sub2 lhs aredex s vl iseq []


(** Applies a substitution list to a list of equations el, i.e. the
    rhs of an extended equation l1=r1...ln=rn, to produce a list of
    state couples for merging *)
let rec apply_subst_equationlist
    (el: (term * term) list) (sl: substitution list)
    (d: Transitions.t) (res: (state * state) list) : (state * state) list =
  match sl with
    | [] -> res
    | s::rem ->
	let newres =
	  List.fold_right
	    (fun (lhs, rhs) l ->
	       let qlhs = Automaton.all_run d (Rewriting.subst_apply s lhs) in
		 if qlhs = [] then l else
		   let qrhs = Automaton.all_run d (Rewriting.subst_apply s rhs) in
		     if qrhs = [] then l else
		       let all_combination =
			 List.filter (fun (q1, q2) -> q1 <> q2) (List2.prod qlhs qrhs)
		       in
		         Verbose.exec 1
			   (fun _ -> List.iter
			      (fun (q1, q2) ->
			         print_string ("Merging states: "^(State.to_string q1)^" and "^(State.to_string q2)^"\n")
                              ) all_combination
                           )
                         ;
			 List2.union all_combination l
            ) el res
	in
	  apply_subst_equationlist el rem d newres




(* This table gives for each lhs/rhs of an equation the set of states
   related (in redex automaton) ... used to find in susbstset the
   corresponding new values... *)
type equation_subst = (state, Stateset.t) Hashtbl.t

let equation_subst_to_string =
  Hashtbl2.to_string ~sep:"\n" ~left:"" ~right:""
    ~format:
    (fun _ _ k e -> (State.to_string k) ^ " -> " ^ (Stateset.to_string e))


let new_equation_table () = Hashtbl.create Common.max_equation

(* Get the set of substitutions to apply for a given exteq and set of substitution.
   given seq1 the top state of lhs of equation, sredex1 the top state of matching equation
   in aredex automaton. *)

(* An equation set maps an equation lhs state (resp. rhs) to the corresponding
   lhs state (resp. lhs). There can be more than one! *)

type equation_table = (state, state list) Hashtbl.t

type exteq_table = (state, exteq list) Hashtbl.t
let new_exteq_table () = Hashtbl.create Common.max_equation

let equation_table_to_string =
  Hashtbl2.to_string ~sep:"\n" ~left:"" ~right:""
    ~format:
    (fun _ _ q e -> (State.to_string q) ^ " -> " ^
       (List2.to_string ~sep:"" ~left:"{" ~right:"}" State.to_string e)
    )

let exteq_to_string =
  let str_of rl =
    List2.to_string ~sep:"," ~left:"[" ~right:"]"
      (fun (t1, t2) -> (Term.to_string t1) ^ "=" ^ (Term.to_string t2))
      rl
  in
    function
      | Unique rl -> "Unique" ^ (str_of rl)
      | Equal (q, rl) -> "Equal(" ^ (State.to_string q) ^ ")" ^ (str_of rl)
      | Any (q, rl) -> "Any(" ^ (State.to_string q) ^ ")" ^ (str_of rl)

let exteq_table_to_string =
  Hashtbl2.to_string ~sep:"\n" ~left:"" ~right:""
    ~format:
    (fun _ _ q l -> (State.to_string q) ^ " -> " ^
       (List2.to_string ~sep:"\n" ~left:"" ~right:"" exteq_to_string l)
    )


let filter_subst_exteq
    (seq1: state) (sredex1: state) (es: substitutionset)
    (seq: substable) (substeq: equation_subst) (stt: s2t)
    (exteq: exteq) : (substitutionset * ((term * term) list)) =
  match exteq with
    | Unique rl ->
        (try
           Verbose.println 1
             (fun _ -> "Equation [" ^ (Term.to_string (Hashtbl.find stt seq1)) ^ "] => [" ^
                (List2.to_string ~sep:" " ~left:"" ~right:"]"
                   (fun (t1, t2) -> (Term.to_string t1) ^ "=" ^ (Term.to_string t2))
                   rl
                ) ^ " applied on state " ^ (State.to_string sredex1)
             )
	 with
           | Not_found ->
               failwith ("Term state " ^ (State.to_string seq1) ^ " does not have image in stt table!")
        );
	(es, rl)

    | Equal (seq2, rl) -> (* for 'Equal' the redex state must be the same *)
	(try
           let othersub = Hashtbl.find (Hashtbl.find seq sredex1) seq2 in
	     (try
                Verbose.println 1
		  (fun _ ->
                     "Equation [" ^ (Term.to_string (Hashtbl.find stt seq1)) ^ "=" ^
                       (Term.to_string (Hashtbl.find stt seq2)) ^ "] => [" ^
                       (List2.to_string ~sep:" " ~left:"" ~right:""
                          (fun (t1, t2) -> (Term.to_string t1) ^ "=" ^ (Term.to_string t2))
                          rl
                       ) ^ "] applied on state " ^ (State.to_string sredex1)
                  )
	      with
                | Not_found ->
                    failwith ("Term state " ^ (State.to_string seq1) ^ "or " ^
                                (State.to_string seq2) ^" does not have image in stt table!")
	     );
	     (SubstitutionSet.extended_intersection es othersub, rl)
	 with
           | Not_found -> (SubstitutionSet.empty, rl)
        )
    | Any (seq2, rl) -> (* for 'Any' the redex state can be anyone *)
        (try
	   let new_redex_for_seq2 = Hashtbl.find substeq seq2 in
	     (try
                Verbose.println 1
		  (fun _ -> "Equation [" ^ (Term.to_string (Hashtbl.find stt seq1)) ^ "," ^
		     (Term.to_string (Hashtbl.find stt seq2)) ^ "] => [" ^
                     (List2.to_string ~sep:" " ~left:"" ~right:""
                        (fun (t1, t2) -> (Term.to_string t1)^"="^(Term.to_string t2))
                        rl
                     ) ^ "] applied on state " ^ (State.to_string sredex1) ^
                     " for lhs and states " ^
		     (List2.to_string
                        ~sep:"," ~left:"" ~right:"" State.to_string
                        (Stateset.elements new_redex_for_seq2)
                     ) ^ " for rhs\n"
                  )
	      with
                | Not_found ->
                    failwith ("Term state "^(State.to_string seq1)^"or "^(State.to_string seq2)^" does not have image in stt table!")
             );
	     let eslist =
               Stateset.fold
	         (fun sred l ->
		    try (SubstitutionTable.find (seq2,sred) seq)::l with
		      | SubstitutionTable.Couple_not_found -> l
                 ) new_redex_for_seq2 []
	     in
	       List.fold_right
		 (fun es2 esres ->
		    let inter = SubstitutionSet.extended_intersection es es2 in
		      if SubstitutionSet.is_empty inter then esres
		      else SubstitutionSet.union inter esres
                 ) eslist SubstitutionSet.empty, rl
         with Not_found ->
           SubstitutionSet.empty, rl)



(* applies a list of labelled substitution (labelled by qlhs and qredex states) on
   an exteq_table eet, w.r.t state merging table mt  *)

let rec apply_subst_extendedeq
    (ls: ((state * state) * substitutionset) list)
    (eet: exteq_table)(seq: substable)(substeq: equation_subst)(d: Transitions.t)
    (s2t: s2t)(res: (state * state) list): ((state * state) list) =
  match ls with
    | [] -> res
    | ((seq1, sredex1), es)::rem ->
	try
	  let eelist = Hashtbl.find eet seq1 in (* get the list of exteq associated to seq1 *)
	  let subst_to_apply =
            List.map
              (filter_subst_exteq seq1 sredex1 es seq substeq s2t)
              eelist
          in
	  let newres =
	    List.fold_right
	      (fun (newes, rl) l ->
		 if SubstitutionSet.is_empty newes then l else
		   apply_subst_equationlist rl
                     (* We don't need Formula annotation for equation
                        application, it is removed. *)
                     (List.map
                        (fun subst -> subst.value)
                        (SubstitutionSet.elements newes)
                     ) d l
              ) subst_to_apply res
	  in
	    apply_subst_extendedeq rem eet seq substeq d s2t newres
	with
          | Not_found -> (* The root state is not part of an exteq *)
	      apply_subst_extendedeq rem eet seq substeq d s2t res



(* Apply subst applies substitution to lhs/rhs of equations and checks if the
   equation applies. This function is also supposed to update substeq and seq
   w.r.t. mt. This function does not produce trivial merge equation.
   If the lhs or rhs is just a variable, we simply merge with this state *)
let rec apply_subst_eq
    (ls: ((state * state) * substitutionset) list) (delta: Transitions.t)
    (seq: substable) (et: equation_table) (substeq: equation_subst)
    (sv: statevar) (stt:Automaton.s2t) (res2: (state * state) list)
    : (state * state) list =

  (* qrhs is the alhs state for the right hand side (or lhs) of equation
     q2 is the state matched by the lhs (or rhs) in the aredex *)
  let rec aux (qlhs: state) (qrhs: state) (q2: state) (es: substitutionset) : (state * state) list =
    (* if rhs is a variable we merge the result of es applied to var of qhs,
       with q2 *)
    try
      let variable = Automaton.statevar_find sv qrhs in
	List.fold_left
	  (fun l s ->
	     if s = q2 then l else (
	       (try                     	       (* equation applies *)
		  Verbose.println 1
                    (fun _ -> "Equation " ^ (Term.to_string (Hashtbl.find stt qlhs)) ^ " = " ^
		       (* (Variable.to_string variable) *)
		       (Term.to_string (Hashtbl.find stt qrhs)) ^ " applies on states: " ^
		       (State.to_string q2) ^ " and " ^ (State.to_string s))
	        with
                  | Not_found ->
                      failwith ("Term state "^(State.to_string qlhs)^" does not have image in stt table!")
               );
               (s, q2)::l)
          ) []
          (* We don't need the formula information, it is removed. We only need the state corresponding to the term *)
          (List.map
             (fun (t, _) -> Term.get_state t)
             (Rewriting.substset_apply es (Var variable))
          )
    with
      | Not_found ->
          try
            (* Look for aredex states associated with substs associated with qrhs in aredex *)
            let new_state_redex = Hashtbl.find substeq qrhs in
            let new_state_list = Stateset.elements new_state_redex in
	      (* List associating with states in aredex (for rhs match) the
	         corresponding sets of substitution *)
	      List.fold_left
	        (fun l qredex ->
	           try
	             let es2 = SubstitutionTable.find (qrhs, qredex) seq in
		       if qredex <> q2 &&
		         SubstitutionSet.matching_substitutionset es2 es
		       then (
		         (* a new merge position has been found! *)
		         (try
			    Verbose.println 1
                              (fun _ -> 
                                 "Equation " ^ (Term.to_string (Hashtbl.find stt qlhs)) ^ " = " ^
			         (Term.to_string (Hashtbl.find stt qrhs))^" applies on states: "^
			         (State.to_string q2)^" and "^(State.to_string qredex)
                              )
		          with
                            | Not_found ->
                                failwith ("Term state " ^ (State.to_string qlhs) ^
                                            " or " ^ (State.to_string qrhs) ^
                                            " does not have image in stt table!")
                         ); (q2, qredex)::l
                       ) else l
	           with
                     | SubstitutionTable.Couple_not_found -> l
                ) [] new_state_list
          with Not_found -> []
  in
    match ls with
      | [] -> res2
	  (* q1 is the qlhs, q2 is the qredex *)
      | ((q1, q2), es)::rem ->
	  let new_res2 =
	    (* memb is the list of rhs of corresponding equation *)
	    try
              let memb = Hashtbl.find et q1 in
		List.append (List.flatten (List.map (fun elt_memb -> aux q1 elt_memb q2 es) memb)) res2
	    with
              | Not_found -> res2 in
	    apply_subst_eq rem delta seq et substeq sv stt new_res2

(* This table gives for each lhs state the list of corresponding rhs (there
   could be more than one *)

type rule_table = (state, (term * condition list) list) Hashtbl.t

let new_rule_table () = Hashtbl.create Common.max_rule_node

let rule_table_to_string =
  Hashtbl2.to_string ~sep:"\n" ~left:"" ~right:""
    ~format:
    (fun _ _ q e -> (State.to_string q) ^ " -> " ^
       (List2.to_string ~sep:"," ~left:"{" ~right:"}"
          (fun (t, ctl) -> "(" ^ (Term.to_string t) ^ " if " ^
	     (List2.to_string ~sep:" /\ " ~left:"" ~right:""
                (function
		   | Joinable (t1, t2) ->
		       Term.to_string t1 ^ "<->" ^ (Term.to_string t2)
		   | _ -> failwith
                       "Completion:rule_table_to_string: Not supported yet"
                ) ctl
             ) ^ ")"
          ) e
       )
    )


(* completion_step type:
   - alhs is the automaton of all left-hand sides of rules and equations,
   - aredex is the tree automaton being completed
   - svlhs is the list of all state names associated to variables in lhs of rules in alhs
   - sveq same for equations
   - et: equation table associating the state of an equation member to the state of the other
   - eet: extended (with patterns) equations
   - rt: rule_table
   - ct: condition_table
   - state2lhs: table associating a state to a lhs term for rules
   - state2eq: table associating a state to a term for equations
   - pattern: table associating state names to the pattern (term) for explicit pattern detection
   - subst_for_eq: found equivalent states w.r.t to a state
   - sublhs: tabled substitutions for lhs of rules
   - subeq: tabled substitution for equations left and right hand side same for lhs of rules
   - new_trans: new "regular" transitions that are part of aredex and not of the previous completed T.A
   - new_epsi: same for epsilon transitions
*)

type epsiset = Epsiset.t

type step =
    { alhs: automaton;
      aeq: automaton;
      aredex: automaton;
      svlhs: statevar;
      sveq: statevar;
      et: equation_table;
      eet: exteq_table;
      rt: rule_table;
      ct: condition_table;
      state2lhs: Automaton.s2t;
      state2eq: Automaton.s2t;
      pattern: Automaton.s2t;
      (* rejected: epsiset; *)
      subst_for_eq: equation_subst;
      sublhs: substable;
      subeq: substable;
      new_trans: external_symtrans list;
      new_epsi: external_epsilon list
    }


let to_string (c: step): string=
  "---------\nalhs= "^(Automaton.to_string c.alhs)^
    "\naeq= "^(Automaton.to_string c.aeq)^
    "\n\naredex= "^(Automaton.to_string c.aredex)^
    "\n\nsv= "^(Automaton.statevar_to_string c.svlhs)^
    "\n\neqsv= "^(Automaton.statevar_to_string c.sveq)^
    "\n\net= "^(equation_table_to_string c.et)^
    "\n\neet= "^(exteq_table_to_string c.eet)^
    "\n\nrt= "^(rule_table_to_string c.rt)^
    "\n\nct= "^(condition_table_to_string c.ct)^
    "\n\npattern= "^(Automaton.s2t_to_string c.pattern)^
    (*  "\n\nproperty= "^(match c.property with Aut(a) -> ("Automaton =\n"^(Automaton.to_string a)^"\n")
        |Pat -> "Pattern only\n")^ *)
    (* "\n\nrejected= "^(Epsiset.to_string c.rejected)^ *)
    "\n\nstate2lhs= "^(Automaton.s2t_to_string c.state2lhs)^
    "\n\nstate2eq= "^(Automaton.s2t_to_string c.state2eq)^
    "\n\nsubst_for_eq= "^(equation_subst_to_string c.subst_for_eq)^
    "\n\nsublhs substable= "^(SubstitutionTable.to_string c.sublhs)^
    "\n\nsubeq substable= "^(SubstitutionTable.to_string c.subeq)^
    "\n\nnew_trans= "^(Transitions.symextlist_to_string c.new_trans)^
    "\n\nnew_epsi= "^(Transitions.epsiextlist_to_string c.new_epsi)^
    "\n\n--------\n\n"


(* Fill pt and retrieves the corresponding final states *)
let rec fill_pattern (l: term list) (sl: state list) (pt: Automaton.s2t) : state list =
  match l, sl with
    | [], _ -> sl
    | t::l, s::sl ->
	Hashtbl.add pt s t;
	fill_pattern l sl pt
    | _ -> failwith "Completion:fill_pattern: no other matching possible"


(* Fill the rtable and etable from the list of rules and list of equation
   + the list of corresponding states. There should be one state per rule
   and two per equation *)
let rec fill_tables (trs: trs) (et: equational_theory) (fs: state list)
    (rtable: rule_table) (etable: equation_table) (eetable: exteq_table)
    (ctable: condition_table): unit =
  let update table si sj =
    try
      let l = Hashtbl.find table si in
	Hashtbl.replace table si (List2.add sj l)
    with
      | Not_found -> Hashtbl.add table si [sj]
  in
    match trs, et, fs with
      |  [] , [] ,  [] -> ()
      | _::_, _  ,  []
      |  [] ,_::_,  ([] | [_]) (* ca ne se resume pas a *)
      |  [] , [] , _::_ -> failwith "Completion:fill_tables: bad number of states in fs!"
      |  [] ,e::rem,s1::(s2::rem2) ->
	   (* For equations *)
	   (* We add the corresponding links either in the etable or in the eetable
	      s1 -> s2 and s2 -> s1 *)
	   (match e with
	      | Regular _ ->                    (* f(x) = (y) *)
                  update etable s1 s2;
                  update etable s2 s1
	            (* [f(x)] -> ... The case where there is only one
		       term in pattern is modeled by an [f(x)=_] *)
	      | Exteq (lhs, Var "_", rl) ->
                  update eetable s1 (Unique rl)

	      | Exteq (lhs, rhs, rl) ->         (* [f(x)=g(y)]-> ...*)
		  update eetable s1 (Equal (s2, rl));
                  update eetable s2 (Equal (s1, rl))

	      | Extany (lhs, rhs, rl) ->        (* [f(x),g(y)]-> ... *)
                  update eetable s1 (Any (s2, rl));
                  update eetable s2 (Any (s1, rl))
           );
           fill_tables trs rem rem2 rtable etable eetable ctable

      | (_, r, cond)::rem, _, (s::rem2) ->
          (try
	     let l = Hashtbl.find rtable s in
	       Hashtbl.replace rtable s ((r, cond)::l)
           with
             | Not_found -> Hashtbl.add rtable s [(r, cond)]
          );
	  if cond <> [] then
            Hashtbl.add ctable s (Stateset.empty);
	  fill_tables rem et rem2 rtable etable eetable ctable


(** Reset the completion step for refinement. c is the completion step
    where all guilty transitions have been suppressed and rejected is
    the list of guilty transitions. This function reinit all
    substitution tables and reinitialize new_trans and new_epsi with
    values coming from c *)
let reset_step (c: step) : step =
  let (trans, epsi) = Transitions.partition (Transitions.to_list_noE c.aredex.trans) in
    { c with
        ct = new_condition_table ();
        subst_for_eq = Hashtbl.create Common.max_equation;
        subeq = SubstitutionTable.new_substable ();
        sublhs = SubstitutionTable.new_substable ();
        (* Here an adaptation may be necessary as in check_intersection function, in order
           to catch all the substitutions, adding of epsilon cannot be done at once. *)
        new_trans = trans;
        (* Not necessary to reconsider epsilon transitions since they are already in
           the set of transition of aredex and adding new substitutions to states will be
           propagated using aredex *)
        new_epsi = []; 
(*         new_epsi = epsi; *)
    }



let new_step
    (a: automaton) (trs: trs) (et: equational_theory)
    (pattern: term list) (trsvl: variable list)
    (eqvl:variable list) : step =
  (* we only add lhs of rules in alhs *)
  let realet = if !approx_type = Equationnal then et else [] in
  let s2lhs = Automaton.new_s2t ()
  and s2eq = Automaton.new_s2t () in
  let all_terms_lhs =
    List.fold_right (fun (lhs, _, _) e -> lhs::e) trs []
  in
    (* we add lhs and rhs for equations in alhs *)
  let all_terms_eq =
    List.fold_right
      (function
         | Regular (lhs, rhs, _) | Exteq (lhs, rhs, _) | Extany (lhs, rhs, _) ->
	     fun l -> lhs::rhs::l
      ) realet []
  in
  let (alhs, svlhs, fslhs) =
    Automaton.term_to_automaton (pattern @ all_terms_lhs) trsvl s2lhs
  in
  let (aeq, sveq, fseq) =
    Automaton.term_to_automaton all_terms_eq eqvl s2eq
  in
  let rtable = new_rule_table () and etable = new_equation_table ()
                                 and eetable = new_exteq_table ()
                                 and ptable = Automaton.new_s2t ()
                                 and ctable = new_condition_table ()
  in
  let new_fslhs = fill_pattern pattern fslhs ptable in
    fill_tables trs realet (new_fslhs @ fseq) rtable etable eetable ctable;

    let (trans, epsi) =
      Transitions.partition (Transitions.to_list a.trans)
    in

    let subst_for_eq = Hashtbl.create Common.max_equation in
    let subeq = SubstitutionTable.new_substable () in
      { alhs = alhs;
        aeq = aeq;
        aredex = a;
        svlhs = svlhs;
        sveq = sveq;
        et = etable;
        eet = eetable;
        rt = rtable;
        ct = ctable;
        pattern = ptable;
        (* rejected = Epsiset.empty; *)
        state2lhs = s2lhs;
        state2eq = s2eq;
        subst_for_eq =subst_for_eq;
        subeq = subeq;
        sublhs = SubstitutionTable.new_substable ();
        new_trans = trans;
        new_epsi = epsi
      }

(* For a given list of (right hand side + a list of conditions) and set of
   substitutions es, this function returns the list of triples (rhs, substituted
   right hand, formula) for all cases where conditions are satisfied. *)
let rec check_conditions
    (l: (term * condition list) list) (es: substitutionset)
    (delta: Transitions.t) (acc2: (term * term * Formula.t) list)
    : (term * term * Formula.t) list =

  let aux
      (es: substitutionset) ((rhs, cond) : term * condition list)
      (acc: (term * term * Formula.t) list)
      : (term * term * Formula.t) list =
    SubstitutionSet.fold
      (fun sub formula (acc: (term * term * Formula.t) list) ->
         (* substcond= list of couples of states (q1,q2) for which we have to decide non
            emptiness of the intersection *)
         let substcond =
           List.map
	     (function
		| Joinable (c1, c2) ->
		    Term.get_state (Rewriting.subst_apply sub c1),
		    Term.get_state (Rewriting.subst_apply sub c2)
		| _ -> failwith "Completion:check_conditions: Not supported yet"
             ) cond
         in
           (* for a conditional rule, we add to the formula the formula such that the intersection is not
              empty *)
         let f= Automaton.not_empty_intersection substcond delta in
         if Formula.is_sat f then
             (rhs, Rewriting.subst_apply sub rhs, Formula.mk_and f formula) :: acc
	   else
             acc
      )
      es acc2
  in
    match l with
      | [] -> acc2
      | (rhs, cond)::rem ->
	  check_conditions rem es delta (aux es (rhs, cond) acc2)


(** Adds an entry to qlhs if it is already in the table (it is a
    conditional rule) *)
let update_condition_table (qlhs: state) (qredex:state) (ct: condition_table) =
  try
    let stateset = Hashtbl.find ct qlhs in
      Hashtbl.replace ct qlhs
        (Stateset.union stateset (Stateset.singleton qredex))
  with
    | Not_found -> ()


(* Depending on the strategy we use new or existing states. Depending on the
   completion type, we use either epsilon transitions for top rewriting or not.
   Note that, specific normalization add the new rules to delta! *)

let specific_normalization
    (t:term) (f: Formula.t) (dest:state) (delta: Transitions.t)
    (cs: strategy) (ct: comptype) : external_trans list =
  match cs with
    | Prior ->
	(* If the transition exists in Delta we reuse it *)
	if ct = Noepsi then
          Automaton.oldnormalize_term false t dest delta
	else
          let (top_state, phi, new_trans) = Automaton.normalize_termE t delta in
(*          let _= print_string ("Formule= "^(Formula.to_string phi)^"\n") in *)
	    Epsilon (top_state, dest, Formula.mk_and f phi) :: new_trans
    | New -> (* we use only new states *)
	if ct = Noepsi then
          Automaton.oldnormalize_term true t dest delta
	else
          let (top_state, new_trans) = Automaton.normalize_new_states t delta in
	    Epsilon(top_state, dest, f) :: new_trans


(* OPTIMISATION possible here
   Checking if a term is already recognized, is performed in a non-deterministic way.
   If it is not precise enough comment the following code...
   Note that in that case normalisation may generate new transitions not be
   deterministic. For instance in example c:
   we have the following transition to normalize: f(s(q3),s(q4)) -> q8 but
   f(s(q3),s(q4)) is normalized to q5 (using deterministic normalization (first
   transition to come is ok :-)) and thus q5 -> q8 is added to delta though
   f(s(q3), s(q4)) ->* q8

   Code to comment:
   if Automaton.oriented_run delta t q2 then
   let _= print_string ("Transition covered!\n") in
   []
   else *)


(* Apply subst applies substitution to rhs of rules.
   This function also adds new transitions to delta.
   For a new transition t ->phi q, if t->phi' q then nothing is add whatever the relation
   between phi and phi' (even if phi is top... maybe to be adapted later) or even if (more generally)
   phi' => phi. For instance, if phi=A and phi'=A/\B, phi is stronger since we know that A is
   enough to have the term.
*)

let rec apply_subst_lhs
    (ls: ((state * state) * substitutionset) list) (delta: Transitions.t)
    (rt: rule_table) (ct: condition_table)
    (stt: s2t) (res1: external_trans list) : external_trans list =
  match ls with
    | [] -> res1
	(* q1 is the qlhs, q2 is the qredex *)
    | ((q1, q2), es)::rem ->
	(* If it is a conditional rule we signal the new hit on q2 *)
	update_condition_table q1 q2 ct; 
	let new_res1 =
	  try
            let erule= (Hashtbl.find rt q1) in
	      (* New list of (rhs, substituted right-hand side,formula) *)
	    let lsubstforms = check_conditions erule es delta [] in
	    let lhs =
              try Term.to_string (Hashtbl.find stt q1) with
	        | Not_found ->
                    failwith ("Term state " ^ (State.to_string q1) ^
                                " does not have image in stt table!")
	    in
	      Verbose.println 1
                (fun _ -> List2.to_string ~sep:"\n" ~left:"" ~right:""
		   (fun (r,rs,f) ->
                      ("Rule "^lhs^" -> "^(Term.to_string r)^" applied on state "^
                         (State.to_string q2)^" with substituted form: "^(Term.to_string rs)^" and formula="^ (Formula.to_string f))
                   ) lsubstforms
                );
	      (* We construct the list of new transitions for es *)
	      List.append
		(List.flatten
		   (List.map
                      (fun (_, t, f) ->
			 match t with
			   | Fsym(_, _) ->
                               if Automaton.oriented_run delta t q2 then (
                                 Verbose.println 1 (fun _ -> "Transition covered!");
                                 []
                               ) else (
                                 Verbose.println 1
                                   (fun _ -> "Transition to normalize: " ^ (Term.to_string t) ^ " -> " ^
                                      (State.to_string q2) ^ " || " ^ (Formula.to_string f)
                                   );
                                 (* (* We normalize new term t w.r.t. delta if it is covered for
                                    merge_loop mode it is enough *)
                                    if !(Common.merge_loop) && (Automaton.oriented_run t q2 delta) then (
                                    if !Common.verbose>1 then
                                    print_string ("Transition covered!\n"); [] ) *)

				 (* Depending on the strategy we use new or existing states. Depending on the
				    completion type, we use either epsilon transitions for top rewriting or not.
				    Note that, specific normalization add the new rules to delta! *)
				 let normedtrans=
				   specific_normalization t f q2 delta !completion_strategy !completion_type
                                 in
				   Verbose.println 1
                                     (fun _ -> "Adding new_trans=" ^ (Transitions.list_to_string normedtrans));
				   normedtrans
			       )
			   | State q3 -> [ Epsilon (q3, q2, f) ]
			   | _ -> failwith "Completion:apply_subst:cannot be anything else"
                      ) (* End of fun (_, t, f) -> ... *)
                      lsubstforms
                   )
                )
		res1
          with Not_found -> res1
        in
	  apply_subst_lhs rem delta rt ct stt new_res1


(* Apply subst applies substitution to patterns and check if it applies.
   If so, we return the pattern and the found substitution set*)
let rec apply_subst_pattern
    (ls: ((state * state) * substitutionset) list) (delta: Transitions.t)
    (pat: Automaton.s2t) (res1: (term * state * substitutionset) list)
    : (term * state * substitutionset) list =
  Verbose.println 3 (fun _ -> "ls= " ^ (SubstitutionTable.list_to_string ls));
  match ls with
    | [] -> res1
	(* q1 is the qlhs, q2 is the qredex *)
    | ((q1, q2), es)::rem ->
	if (SubstitutionSet.is_empty es) then
	  apply_subst_pattern rem delta pat res1
	else
	  let new_res1 =
	    try
	      let term = Hashtbl.find pat q1 in
		(term, q2, es)::res1
	    with Not_found -> res1
          in
	    apply_subst_pattern rem delta pat new_res1



(* Multiple adding of epsilon transition. Result is the list of uncovered
   new epsilon transitions *)

let rec multiple_add_epsilon (le: external_trans list) (d: Transitions.t) (uncovered_epsi: external_epsilon list) =
  match le with
    | [] -> uncovered_epsi
    | Epsilon t :: rem ->
	if Transitions.addR' d t then (* B: on essaye d'ajouter la transition a d pour tester si elle uncovered...
                                                 Pourquoi cela marche-t-il ???
                                              *)
	  (* let (q1, q2, _) = t in  *)
	    begin
              (*
	      Transitions.new_epsilon_transitions :=
                List2.union [(q1, q2)] !Transitions.new_epsilon_transitions;
              *)
	      multiple_add_epsilon rem d (t::uncovered_epsi);
	    end
	else
	  multiple_add_epsilon rem d uncovered_epsi
    | _ -> failwith "Completion:multiple_add_epsilon: cannot be another kind of transition"



(** Returns all top substitution of conditional rules *)
let rec add_conditional_sub
    (ct: (state * Stateset.t) list) (slhs: substable)
    (acc: ((state * state) * substitutionset) list)
    : ((state * state) * substitutionset) list =
  match ct with
    | [] -> acc
    | (s, stateset)::rem ->
	let new_acc =
	  Stateset.fold
	    (fun elt l ->
               try ((s, elt), (SubstitutionTable.find (s, elt) slhs)) :: l with
	         | SubstitutionTable.Couple_not_found ->
                     failwith "Completion:add_conditional_sub:Subsitution not found in table for conditional rule!"
            ) stateset acc
	in
          add_conditional_sub rem slhs new_acc


(* Propagation of new subst on equation_subst table:
   for a solution: ((q1,q2),es), if q1 is the top symbol of an equation
   this function adds q2 to the set of states associated with q1 *)

let rec propagate_on_substeq
    (l: ((state * state) * substitutionset) list)
    (seq: equation_subst) (et: equation_table) (eet: exteq_table) : unit =
  List.iter
    (fun ((q1, q2), _) ->
	(* We check if q1 is a member of an equation, if so we add the corresponding state *)
       if (Hashtbl.mem et q1) || (Hashtbl.mem eet q1) then
	 let state_set =
           try Hashtbl.find seq q1 with
             | Not_found -> Stateset.empty
         in
	   Hashtbl.replace seq q1 (Stateset.add q2 state_set)
    ) l



(** Merging fixpoint: applies successively merging equation up to a
    normal form. The result is the list of new epsilonE transitions
    added by (possibly) several equation applications. For instance, if
    the application of one equation raises new matches for another
    equation, this is found... and solved.  The result is a list of
    epsilon q1->q2 labelled by the formula q1->q2. This is necessary
    for the propagation of substitutions. *)
(* Merging fixpoint transforms aut by adding approximation transitions
   (i.e. epsilonE transitions) *)
let fixpoint_merging
    (init_sym: external_symtrans list) (epsiR: external_epsilon list)
    (aut: automaton) (eq: equation_table) (eet: exteq_table)
    (aeq: automaton) (eqvar: statevar) (es: equation_subst)
    (subeq: substable) (stt:Automaton.s2t) (* rejected:epsiset *)
    : (external_epsilon list) =

  let rec fpmerging_aux
      (sym_to_consider: external_symtrans list) (epsiR: external_epsilon list)
      (epsiE: external_epsilon list) (resE: external_epsilon list) : external_epsilon list =

    match sym_to_consider, epsiR, epsiE with
        (* - first iteration of fpmerging_aux there is an empty set of epsiE to consider
           - next iterations of fpmerging_aux there are emptysets of sym and epsiR and some epsiE found
           by previous steps and that may lead to new matching solutions and new epsiE etc... until
           a fixpoint is found and no new epsiE is found.

           For iterations with epsiE, all substitutions have to be considered. Substitutions produced
           by epsilonE transitions must not be rejected (hence the the iseq flag is set to false) *)

      | [], [], [] -> resE
      | _::_, _, _::_ -> failwith "Fixpoint merging impossible case 1"
      | _, _::_, _::_ -> failwith "Fixpoint merging impossible case 2"
      | _, _, _ ->

          (* For iterations with epsiE, all substitutions have to be considered. Substitutions produced
             by epsilonE transitions must not be rejected (hence the the iseq flag is set to false) *)
	  let soleq =
            if epsiE = [] then
              complete_filter aeq aut sym_to_consider epsiR subeq eqvar true
            else
              complete_filter aeq aut [] epsiE subeq eqvar false
          in
	  let _ = propagate_on_substeq soleq es eq eet in

	  let states_to_merge1 = apply_subst_eq soleq aut.trans subeq eq es eqvar stt [] in
	  let states_to_merge2 = apply_subst_extendedeq soleq eet subeq es aut.trans stt states_to_merge1 in
            (* For each couple (q1,q2), we add q1->q2 and q2->q1 only if they are not already covered
               by aut. Note that it is not useful to keep q1->q2 if q1->*q2.
               Transitions.add_approximation should add it to the automaton (to be corrected).
               Not necessary to add new matching substitutions. We also do not add q1->q2 if it is part
               of the rejected epsilon transitions *)

            (* Probleme ajout epsilon e *)
	  let (new_epsilonE: external_epsilon list) =
            List.fold_right
              (fun ((q1: state), (q2: state)) l ->
                 let tmp =
                   if Transitions.addE' aut.trans q1 q2 then
                     (q1, q2, Formula.mk_atom q1 q2 Eq)::l
                   else
                     l
                 in
                   if Transitions.addE' aut.trans q2 q1 then
                     (q2, q1, Formula.mk_atom q2 q1 Eq) :: tmp
                   else
                     tmp
              ) states_to_merge2 []
          in
            Verbose.exec 2
	      (fun _ ->
                 if new_epsilonE = [] then
                   print_string "\nNo approximation transitions to add!\n"
                 else
                   print_string ("\nApproximation transitions="^
                                   (List2.to_string ~sep:"\n" ~left:"" ~right:"\n"
                                      Transitions.epsiext_to_string new_epsilonE)
                                )
              )
	    ;
            fpmerging_aux [] [] new_epsilonE (new_epsilonE @ resE)
  in
    fpmerging_aux init_sym epsiR [] []




(** filter substitutions associated to final states pred is a list
    associating a predecessor of a final state to the final state that
    is reachable by ->epsilonR and the formula necessary to reach this
    state.  fspred is the state of final states predecessors (to
    accelerate a bit the check (i)) instead of crawling in pred.        *)
let filter_final
    (ls: (term * state * substitutionset) list)
    (pred: (state * (state * Formula.t)) list)
    (fspred: Stateset.t)
    : (term * state * substitutionset) list =
  List.fold_right
    (fun (t, s, ss) l ->
       try
         if Stateset.mem s fspred then  (* i *)
           let (final, phi) = List.assoc s pred in
             (* B: ici y a un acces aleatoire dans une liste*)
             (t, final, SubstitutionSet.addPhi_to_all phi ss)::l
         else
           l
       with Not_found -> l
    ) ls []


(* One step of completion, performs a full copy of current completion step
   and returns the set of substitution matching the pattern terms (if any).
   Also perform the transitive Rclosure of the new aredex. *)
let do_one_step (c: step)
    : step * (term * state * substitutionset) list =
  (* 1) We make copies of mutable structures for the new completion step *)
  let new_subst_for_eq = Hashtbl.copy c.subst_for_eq in
(*  let _ = print_string "1 "; flush stdout in *)
  let new_aredex = Automaton.copy c.aredex in
(*  let _ = print_string "2 "; flush stdout in *)
  let new_slhs = SubstitutionTable.copy c.sublhs in
(*  let _ = print_string "3 "; flush stdout in *)
  let new_seq = SubstitutionTable.copy c.subeq in
(*   let _ = print_string "4 "; flush stdout in *)
  let new_ct = Hashtbl.copy c.ct in
(*  let _ = print_string "5 "; flush stdout in *)
    (* Transitions.reset_new_epsilons(); *)

  (* 2) We search for equation applications and add new epsilonE transitions to consider,
     due to equations. This modifies new_aredex (add new epsilonE transitions) *)
  let new_epsi_from_merge =
    fixpoint_merging c.new_trans c.new_epsi new_aredex c.et c.eet
      c.aeq c.sveq new_subst_for_eq new_seq c.state2eq (* c.rejected*) in
(*  let _ = print_string "5.5 "; flush stdout in  *)
    flush stdout;
    Verbose.println 3
      (fun _ -> "New epsilonE transitions due to equations: "^
         (Transitions.epsiextlist_to_string new_epsi_from_merge));
(*    let _ = print_string "6 "; flush stdout in *)
    let _ = flush stdout in
      if c.new_trans = [] && c.new_epsi = [] && new_epsi_from_merge = [] then
	({ subst_for_eq = new_subst_for_eq;
	   aredex = c.aredex; (* here ... if no new transitions... no need to merge, hence
                                 we output the previous automaton and not: new_aredex that
                                 may contain unwanted mergings? *)
	   sublhs = new_slhs;
	   subeq = new_seq;
	   new_trans = [];
	   new_epsi = [];
	   et = c.et;
	   eet = c.eet;
	   rt = c.rt;
           (* rejected = c.rejected; *)
	   ct = new_ct;
	   pattern = c.pattern;
	   state2lhs = c.state2lhs;
	   state2eq = c.state2eq;
	   alhs = c.alhs;
	   aeq = c.aeq;
	   svlhs = c.svlhs;
	   sveq = c.sveq},
	 [])  (* If there are no new transition at all (even with merging operations)
	       then patterns cannot be matched now if it was not the case before *)
    else begin
      (* 3) We search and produce new transitions *)
      (* We use transitions of the LAST completion step to obtain new
	 matching with lhs *)
(*       let _ = print_string "7 "; flush stdout in *)

      Verbose.println 3
        (fun _ -> "Before complete filter:\nnewaredex=" ^ (Automaton.to_string new_aredex) ^
           "\nc.new_trans=" ^ (Transitions.symextlist_to_string c.new_trans) ^
           "\nc.new_epsi=" ^ (Transitions.epsiextlist_to_string c.new_epsi) ^
           "\nnew_slhs=" ^ (SubstitutionTable.to_string new_slhs) ^
           "\nsvlhs=" ^ (Dumper.dump c.svlhs)
        );

(*         let _ = print_string "8 "; flush stdout in *)

      let _ = flush stdout in
      let sublhs =
        complete_filter c.alhs new_aredex c.new_trans c.new_epsi new_slhs c.svlhs false
      in
(*      let _ = print_string "9 "; flush stdout in *)
      let _ = flush stdout in
(*	let _ = print_string "9.5 "; flush stdout in *)
        let all_sublhs =
	  add_conditional_sub (Hashtbl2.elements new_ct) new_slhs sublhs
        in
(*	   let _ = print_string "10 "; flush stdout in *)
        (* We check if a pattern is filled (pattern or subpattern
	   depending on what was declared in the specification file *)
        let patsub =
	  if Hashtbl.length c.pattern = 0 then [] else
            let temp =
              apply_subst_pattern all_sublhs new_aredex.trans c.pattern []
            in
	      if !Common.subpattern then temp else
	        (* for patterns we check in final states q and predecessors of final states q'
                   w.r.t epsilon i.e. terms that comes from final states of the initial TA...
                   Of course, the corresponding formula is completed by the formula phi that may
                   label q'->q *)
	        filter_final
                  temp
                  (Transitions.get_all_predPhi new_aredex.fstates new_aredex.trans)
                  (Transitions.get_all_pred new_aredex.fstates new_aredex.trans)
        in
(*	let _ = print_string "11 "; flush stdout in *)
        (* We apply the substitution to obtain the new transitions. Note that
           they are added to delta! *)
        let obtained_trans = apply_subst_lhs all_sublhs new_aredex.trans c.rt new_ct c.state2lhs [] in
(*	let _ = print_string "12 "; flush stdout in *)
        (* We split transitions between trans and epsilon trans *)
        let (trans_not_covered, epsiR) = List.partition Transitions.not_epsi obtained_trans in
(*	  let _ = print_string "13 "; flush stdout in *)
          (* Epsi transitions are added and states to merge are gathered. epsiR transitions
             are checked, i.e. check if they are not already covered by new_aredex.
             EpsilonE transitions automatically added to the new transitions to consider,
             they are not filtered by multiple_add_epsilon... *)

          (* new_epsi_from_merge are used as an accumulator and are thus not added by epsi_not_covered!
             this is important since we add to those epsi the formula corresponding to the transition
             itself so that substitution can be propagated in the right way.
             I.e. the transition (coming from merge) q1->q2 is labelled by the formula q1->q2! *)

          let epsi_not_covered =
            multiple_add_epsilon epsiR new_aredex.trans new_epsi_from_merge
          in
            Verbose.println 3
              (fun _ -> "All epsilon to be reconsidered for next matching: " ^
                 (Transitions.epsiextlist_to_string epsi_not_covered)
              );
            (* We explicitly build the transitive Rclosure of the set of transitions *)
            (* Transitions.transitive_Rclosure new_aredex.trans; *)
(*	    let _ = print_string "14 "; flush stdout in  *)
            (* 5) *)
	    ( { subst_for_eq = new_subst_for_eq;
	        aredex = new_aredex;
	        sublhs = new_slhs;
	        subeq = new_seq;
	        new_trans = List.map Transitions.ext2trans trans_not_covered;
	        new_epsi = epsi_not_covered;
	        et = c.et;
	        eet = c.eet;
	        rt = c.rt;
	        ct = new_ct;
                (* rejected = c.rejected; *)
	        pattern = c.pattern;
	        state2lhs = c.state2lhs;
	        state2eq = c.state2eq;
	        alhs = c.alhs;
	        aeq = c.aeq;
	        svlhs = c.svlhs;
	        sveq = c.sveq
              }, patsub )
    end (* of else located before (3) *)



(* Interactive pattern checking. Rebuild a substitution table from a
   set of transitions and epsilon (epsiR and epsiE). Note that
   substitution are not propagated to predecessors of states unless if
   it is necessary. For instance if we find a substitution sigma on q1
   and q1 -> q2 sigma is only propagated to fathers of q2 and not to
   q2 *)
let check_intersection
    (a: automaton) (pattern: term list) (vl: variable list):
    (term * state * substitutionset) list =

  let state2term_table = Automaton.new_s2t () in
  let (autpat, svpat, fspat) =
    Automaton.term_to_automaton pattern vl state2term_table
      (*   let _= fill_pattern pattern fspat state2term_table in *)
      (*  print_string ("alhs= "^(Automaton.to_string autpat)^"\n");    *)
  and (trans, epsi) =
    List.partition Transitions.not_epsi (Transitions.to_list a.trans)
  and subpat =
    SubstitutionTable.new_substable ()
  in

  (* symbolic trans can be added at once *)
  let fil =
    complete_filter autpat a (List.map Transitions.ext2trans trans) [] subpat svpat false
  in
    (* epsilon transitions have to be added one by one *)
  let fil =
    List.fold_right
      (fun trans ls ->
         (complete_filter autpat a [] [trans] subpat svpat false) @ ls)
      (List.map Transitions.ext2epsi epsi) fil
  in
    apply_subst_pattern fil a.trans state2term_table []


let patternres_to_string (sol: (term * state * substitutionset) list) : string =
  List2.to_string ~sep:"\n" ~left:"" ~right:"\n"
    (fun (t, s, es) -> (Term.to_string t) ^ ": " ^ (State.to_string s) ^
       " -> " ^ (SubstitutionSet.to_string es)
    ) sol













    (* ------------------------------TESTS ----------------------------------------*)


    (* ------------------------------TESTS ----------------------------------------*)

IFDEF TEST THEN print_string "Completion.ml\n\n" ENDIF

IFDEF TEST THEN
let s = SubstitutionPhi.addBoth "x" (Term.State 1) Formula.mk_true SubstitutionPhi.empty
let set1 = SubstitutionSet.add s SubstitutionSet.empty
let t = SubstitutionTable.new_substable ()
let _ = SubstitutionTable.add s (1,1) t
let sv1 = new_statevar ()
let _ = Automaton.statevar_add sv1 1 "x"
let sv2 = new_statevar ()
let tl =
  List.map
    (function
       | Trans t -> t
       | Epsilon _ -> failwith "Completion:unitary testing")
    (Transitions.trans_to_list (Automaton.transtable Automaton.aut2))

let substsol = SubstitutionTable.new_substable ()
let substsolres = SubstitutionTable.new_substable ()
let substsol4 = SubstitutionTable.new_substable ()
let substsol7 = SubstitutionTable.new_substable ()
let substsol9 = SubstitutionTable.new_substable ()
let substsol2: substable = SubstitutionTable.new_substable ()
let subst1: substitutionPhi =
  SubstitutionPhi.addPhi
    Formula.mk_true
    (SubstitutionPhi.add "x" (State 2) (SubstitutionPhi.add "y" (State 1) SubstitutionPhi.empty))
let substens = SubstitutionSet.add subst1 SubstitutionSet.empty
ENDIF





IFDEF TEST THEN
open Debug

begin
  print_string "match_one\n";
  asserteq (match_one 1 1 (SubstitutionTable.new_substable()) sv1) set1 "match_one_1";
  asserteq (match_one 1 1 (SubstitutionTable.new_substable()) sv2) (SubstitutionSet.empty) "match_one_2";
  asserteq (match_one 1 1 t sv2) set1 "match_one_3";
  print_string "filter\n\n";
  print_string ("Before filter substsol=\n"^(SubstitutionTable.to_string substsol)^"\n")
end
END



IFDEF TEST THEN
(* aut1= f(a, g(x, y)) and tl contains f(a, g(f(a,b), b)) *)
let stt= Automaton.new_s2t ()
let sol= filter Automaton.aut1 tl [] substsol Automaton.sv false
let sol2= complete_filter Automaton.aut1 Automaton.aut2 tl [] substsol2 Automaton.sv false
let t3= Fsym("f", [Fsym("c",[])])
let tr2= ("f",[0],0)  (* Transition to add to aut4 such that f(a) -> q0 *)
let t4= Fsym("a",[])
let t5= Fsym("b",[])
let t6= Fsym("c",[])
let t7= Fsym("f", [Fsym("g",[Var("x")]);Fsym("g",[Var("y")])])
let t8= Fsym("f", [Fsym("g",[Fsym("a",[])]);Fsym("b",[])])

let (aut7,sv7,_)= Automaton.term_to_automaton [t7] ["x";"y"] stt
  (* aut8 and aut9 are identical but used for distinct tests *)
let (aut8,sv8,_)= Automaton.term_to_automaton [t8] [] stt
let (aut9,sv9,_)= Automaton.term_to_automaton [t8] [] stt

(* aut3 recognizes f(a) *)
let (aut3,sv3,_)= Automaton.term_to_automaton [t3] [] stt
  (* aut4 recognizes f(a), b and s(s(o)) *)
let (aut4,_,_)= Automaton.term_to_automaton [t4;t5;t6] [] stt

let tl4= List.map (function (Trans t) -> t | (Epsilon _) -> failwith "Completion:unitary testing") (Transitions.trans_to_list (Automaton.transtable aut4))

let s7= Transitions.make_state aut8.trans
let s8= Transitions.make_state aut8.trans
let _= Transitions.add' aut8.trans  ("g", [s7], s8)
let _= Transitions.add' aut8.trans  ("b", [], s7)

let s7b= Transitions.make_state aut9.trans
let s8b= Transitions.make_state aut9.trans
let _= Transitions.add' aut9.trans ("g", [s7b], s8b)
let _= Transitions.add' aut9.trans ("b", [], s7b)


let tl8 = List.map (function Trans t -> t | Epsilon _ -> failwith "Completion:unitary testing") (Transitions.trans_to_list (Automaton.transtable aut8))

(* Adding R-epsilon trans such that
   a                 b                c
   |      (3)        |       (2)      | (1)
   q0 <-(q2,q3) --- q1 <- (q3,q4) -- q2
*)

let sol4= complete_filter aut3 aut4 tl4 [] substsol4 sv3 false
let _= print_string ("(1) substsol4= "^(SubstitutionTable.to_string substsol4)^"\n")

let _= Transitions.addR aut4.trans (2, 1, Formula.mk_atom 3 4 Eq)
let sol5= complete_filter aut3 aut4 [] ((2,1,Formula.mk_atom 3 4 Eq)::[]) substsol4 sv3 false
let _= print_string ("(2) substsol4= "^(SubstitutionTable.to_string substsol4)^"\n")

let _= Transitions.addR aut4.trans (1,0,Formula.mk_atom 2 3 Eq)
let sol6= complete_filter aut3 aut4 [] ((1,0,Formula.mk_atom 2 3 Eq)::[]) substsol4 sv3 false
let _= print_string ("(3) substsol4= "^(SubstitutionTable.to_string substsol4)^"\n")

let _= Transitions.add' aut4.trans tr2
let sol6f= complete_filter aut3 aut4 [tr2] [] substsol4 sv3 false
let _= print_string ("(4) substsol4= "^(SubstitutionTable.to_string substsol4)^"\n")


let sol7= complete_filter aut7 aut8 tl8 [] substsol7 sv7 false
let _= complete_filter aut7 aut9 tl8 [] substsol9 sv7 false


(* for completion *)

let t1= Fsym("f", [Fsym("a",[]);Fsym("b",[])]) (* f(a,b) *)
  (* f(x, y) -> f(s(x), s(y)) *)
let (autb,_,_)= Automaton.term_to_automaton [t1] [] stt
let rule1= (Fsym("f", [Var("x");Var("y")]), Fsym("f", [Fsym("s", [Var("x")]);Fsym("s", [Var("y")])]), [])
  (* s(s(x)) = s(x) *)
let eq= Regular(Fsym("s", [Fsym("s", [Var("x")])]), Fsym("s", [Var("x")]), [])
let vl= ["x";"y"]


(* for completion *)

let t2= Fsym("a",[]) (* a *)
  (* a -> f(a)  *)
let (autc,_,_)= Automaton.term_to_automaton [t2] [] stt
let rule2= Fsym("a", []),Fsym("f", [Fsym("a", [])]),[]
let vl2= []

(* for completion *)

let t3= Fsym("f",[Fsym("a",[])]) (* f(a) *)
  (* f(x) -> f(s(x))  *)
let (autd,_,_)= Automaton.term_to_automaton [t3] [] stt
let rule3= (Fsym("f", [Var("x")]),Fsym("f", [Fsym("s", [Var("x")])]),[])
let vl3= ["x"]


(* for completion *)

let t4= Fsym("f",[Fsym("a",[])]) (* f(a) *)
let t5= Fsym("b",[])
let t6= Fsym("s", [Fsym("s", [Fsym("o",[])])])

(* f(b) -> f(c), a -> b  *)
let (aute,_,_)= Automaton.term_to_automaton [t4;t5] [] stt
let rule4= (Fsym("f", [Fsym("b",[])]),Fsym("f", [Fsym("c", [])]),[])
let rule5= (Fsym("a",[]),Fsym("b",[]),[])
let rule6= (Fsym("f", [Fsym("c",[])]),Fsym("d", []),[])
let vl4= []
let t12=Fsym("g",[Fsym("f",[Fsym("a",[])])])
let t13= Fsym("f",[Fsym("c",[])])

let stt1= Automaton.new_s2t ()
let (autf,_,_)= Automaton.term_to_automaton [t13;t5;t12] [] stt1

(* for completion *)

let rule3= (Fsym("f", [Var("x")]),Fsym("f", [Fsym("s", [Var("x")])]),[])
let eq2= Regular(Fsym("s", [Var("x")]), Var("x"), [])

(* for completion *)

let eq3= Regular(Fsym("f", [Var("_");Var("_")]),Fsym("f",[Var("_");Var("_")]),[])
let eq4= Regular(Fsym("s", [Fsym("o",[])]),Fsym("o",[]),[])
let eq5= Regular(Fsym("s", [Fsym("s", [Fsym("s", [Fsym("o",[])])])]),Fsym("s", [Fsym("s", [Fsym("o",[])])]),[])
let eq6= Regular(Fsym("a",[]),Fsym("b",[]),[])
let eq7= Regular(Fsym("f",[Fsym("b",[])]),Fsym("f",[Fsym("c",[])]),[])
END

IFDEF TEST THEN
begin
  print_string ("filter aut1="^(Automaton.to_string Automaton.aut1)^"\n on aut2="^(Automaton.to_string Automaton.aut2)^"\n is equal to "^(SubstitutionTable.list_to_string sol)^"\n");
  asserteq (SubstitutionTable.list_to_string sol)
    "(q3,q3) -> {{y -> q1, x -> q2}[Top]}\n(q0,q0) -> {{}[Top]}\n" "test filter";
  asserteq substsol substsolres "NEW substitutions should NOT be added to
     substable";
  print_string "complete_filter\n";
  asserteq (match_in_list [1;2;3;4;3;5] [2;2;2;2;5;4] 3 5) true "";
  asserteq (match_in_list [1;2;3;4;5;5] [2;2;2;2;5;4] 3 5) false "";
  print_string ("aut1 (lhs)="^(Automaton.to_string aut1)^"\n");
  print_string ("aut2 (redex)="^(Automaton.to_string aut2)^"\n");
  print_string ("Complete filter on aut1 + aut2 should fill substsol2="^(SubstitutionTable.to_string substsol2)^
		  "\n and substens="^(SubstitutionSet.to_string substens)^"\n");

  print_string ("\nsol2= "^(SubstitutionTable.list_to_string sol2)^"\n");
  asserteq (string_equal (SubstitutionTable.list_to_string sol2) (SubstitutionTable.list_to_string [((4,4),substens)]))
    true "Set of solution w.r.t. **final** set of automata should only contain new substitutions on final states";
  asserteq 	(string_equal "(q4,q4) -> {{y -> q1, x -> q2}[Top]}
(q3,q3) -> {{y -> q1, x -> q2}[Top]}
(q0,q0) -> {{}[Top]}" (SubstitutionTable.to_string substsol2)) true "substsol2 should have been filled by complete_filter";

    (* L(aut3,q0)={ c }, L(aut3,q1)= {f(c)}, q1 final*)

    print_string ("aut3 (lhs)="^(Automaton.to_string aut3)^"\n");

    (*
       a                 b                c
       |      (3)        |       (2)      | (1)
       q0 <-(q2,q3) --- q1 <- (q3,q4) -- q2

    *)

    print_string ("aut4 (redex)="^(Automaton.to_string_check aut4)^"\n");

    (* substsol4 should be
       (q0,q0) -> {{}[((q3 -> q4 /\ q2 -> q3))]}
       (q0,q1) -> {{}[((q3 -> q4))]}
       (q0,q2) -> {{}[(Top)]}

       sol4 should be empty (because no transition to final state of aut3(lhs)
       sol5 same
       sol6 same
       sol6f should propose one solution with the right formula: (q1,q0) -> {{}[((q3 -> q4 /\ q2 -> q3))]}
       f(a) (q2,q3)-> f(b) (q3,q4)-> f(c)
    *)

    asserteq sol4 [] "sol4 should be empty (because no transition to final state of aut3(lhs)";
    asserteq sol5 [] "sol5 should be empty (because no transition to final state of aut3(lhs)";
    asserteq sol6 [] "sol6 should be empty (because no transition to final state of aut3(lhs)";
    print_string ("sol6f= "^(SubstitutionTable.list_to_string sol6f)^"\n");

    asserteq (string_equal
                (SubstitutionTable.list_to_string sol6f)
                "(q1,q0) -> {{}[(Eq(q2,q3) /\\ Eq(q3,q4))]}") true
      "sol6 should propose one solution with the right formula: (q1,q0) -> {{}[(Eq(q2,q3) /\\ Eq(q3,q4))]}";

    print_string ("substsol4= "^(SubstitutionTable.to_string substsol4)^"\n");
    asserteq (string_equal
                (SubstitutionTable.to_string substsol4)
                "(q0,q2) -> {{}[Top]}
(q0,q1) -> {{}[Eq(q3,q4)]}
(q1,q0) -> {{}[(Eq(q2,q3) /\\ Eq(q3,q4))]}
(q0,q0) -> {{}[(Eq(q2,q3) /\\ Eq(q3,q4))]}") true "Substsol4 is finally filled";

  end
  END

  IFDEF TEST THEN
let d10=Automaton.aut10.trans
let t10=Fsym("f",[Fsym("g",[Var("_")]);Fsym("g",[Fsym("b",[])])])
let t11=Fsym("g",[Fsym("g",[Var("X")])])
let sl= [(Substitution.add "X" (Fsym("a",[])) Substitution.empty)]

  END

  IFDEF TEST THEN
  begin
    print_string "Test apply_subst_equationlist\n";
    asserteq (apply_subst_equationlist [t10,t11] sl d10 [0,6;4,6]) [0,6;4,6] "Application of subst gives the right state list"
  end
  END


  IFDEF TEST THEN
  (* aute= f(a) et b
     rule4= f(b) -> f(c) et rule5= a -> b *)
let dc= new_step aute [rule4;rule5] [] [] vl4 vl4 in
  print_string "Test completion one step\n";
  print_string ("dc= "^(to_string dc)^"\n");
  let (dc1,_)= do_one_step dc in
    print_string ("dc1 (a -> b)= "^(to_string dc1)^"\n");
    let (dc2,_)= do_one_step dc1  in
      print_string ("dc2 (f(b) -> f(c)= "^(to_string dc2)^"\n");
      let (dc3,_)= do_one_step dc2  in
	print_string ("dc3= "^(to_string dc3)^"\n");
	asserteq (Transitions.symextlist_to_string dc2.new_trans) "f(q3) -> q4\nc -> q3\n" "adding an epsilon trans should raise a new computation of substitution tables";
	asserteq (dc3.new_trans=[] && dc3.new_epsi=[]) true "The automaton is complete"
          END




          IFDEF TEST THEN
let cb= new_step autc [rule2] [] [] vl2 vl2 in
  print_string ("\n\n other one_step \n\n");
  print_string ("cb= "^(to_string cb)^"\n");
  let (c2b,_)= do_one_step cb  in
    print_string ("c2b= "^(to_string c2b)^"\n");
    let (c3b,_)= do_one_step c2b  in
      print_string ("c3b= "^(to_string c3b)^"\n");
      asserteq c3b.new_trans [] "After 2 completion steps the t.a. is complete";
      END



        IFDEF TEST THEN
        print_string ("Test fixpoint_merging\n")
        (* rule6= g(f(c)) -> d
           autf= g(f(a)), b, f(c)
           eq6= a=b
           eq7= f(b)=f(c)

           fixpoint merging should recursively add equation between a = b and then between f(a)=f(b)=f(c)
        *)

let cfm = new_step autf [rule6] [eq6; eq7] [] [] [] in
let autfCopy = Automaton.copy autf in
let cfmb = new_step autfCopy [rule6] [eq6;eq7] [] [] [] in
let _= print_string ("cfm before fixpoint merging= "^(to_string cfm)^"\n") in
let new_subst_for_eq = Hashtbl.copy cfm.subst_for_eq in
let new_seq = SubstitutionTable.copy cfm.subeq in
let new_epsiE =
  fixpoint_merging
    cfm.new_trans cfm.new_epsi cfm.aredex
    cfm.et cfm.eet cfm.aeq cfm.sveq
    new_subst_for_eq new_seq
    cfm.state2eq (* cfm.rejected *)
in
  begin
    print_string ("New epsilon due to equations: " ^ (Transitions.epsiextlist_to_string new_epsiE)^"\n");
    asserteq (string_equal (Transitions.epsiextlist_to_string new_epsiE) "q1 -> q4 || Eq(q1,q4)
q4 -> q1 || Eq(q4,q1)
q3 -> q2 || Eq(q3,q2)
q2 -> q3 || Eq(q2,q3)") true "Fixpoint merging successfully applied the 2 equations, one permitting the application of the other";
    let (cfmb2,_) = do_one_step cfmb in
      print_string ("cfmb2= "^(to_string cfmb2)^"\n");
      let (cfmb3,_) = do_one_step cfmb2 in
        print_string ("cfmb3= "^(to_string cfmb3)^"\n");
        asserteq (cfmb3.new_trans) [] "After 1 completion step the t.a. is complete sym";
        asserteq (cfmb3.new_epsi) [] "After 1 completion step the t.a. is complete epsi";
  end
    END

    IFDEF TEST THEN
    print_string ("\n\n other one_step \n\n");
  let cc= new_step autd [rule3] [eq] [] vl3 vl3 in
    print_string ("cc before completion= "^(to_string cc)^"\n");
    let (c2c,_)= do_one_step cc  in
      print_string ("cc after 1 completion= "^(to_string c2c)^"\n");
      let (c3c,_)= do_one_step c2c  in
        print_string ("cc after 2 completion= "^(to_string c3c)^"\n");
        let (c4c,_)= do_one_step c3c  in
          print_string ("cc after 3 completion= "^(to_string c4c)^"\n");
          let (c5c,_)= do_one_step c4c  in
            print_string ("cc after 4 completion= "^(to_string c5c)^"\n");
            let (c6c,_)= do_one_step c5c  in
              print_string ("cc after 5 completion= "^(to_string c6c)^"\n")
                END



                IFDEF TEST THEN
let c= new_step autb [rule1] [eq] [] vl vl in
let cstring= (to_string c) in
let (c2,_)= do_one_step c  in
  print_string "\n\n one_step\n\n";
  asserteq cstring (to_string c) "do_one_step should not change c";
  print_string ("c before completion="^(to_string c)^"\n");
  print_string ("c after completion="^(to_string c2)^"\n");
  let (c3,_)= do_one_step c2  in
    print_string ("c after 2 completion="^(to_string c3)^"\n");
    let (c4,_)= do_one_step c3  in
      print_string ("c after 3 completion="^(to_string c4)^"\n");
      let (c5,_)= do_one_step c4  in
        print_string ("c after 4 completion="^(to_string c5)^"\n");
        print_string "merge_one_layer\n\n";
        END


          IFDEF TEST THEN
let rule2= (Fsym("a", []),Fsym("f", [Fsym("a", [])]),[])
let rule10= (Fsym("f", [Fsym("a", [])]),Fsym("a", []),[])
let vl2= []
  END

  (* terme: a et regles a -> f(a), f(a) -> a  *)

  IFDEF TEST THEN
let ec= new_step autc [rule2;rule10] [] [] [] []  in
  print_string ("Epsilon cycles are no longer merged\n");
  print_string ("c = "^(to_string ec)^"\n");
  let (ec2,_)= do_one_step ec  in
    print_string ("c2= "^(to_string ec2)^"\n");
    let (ec3,_)= do_one_step ec2  in
      print_string ("c3= "^(to_string ec3)^"\n");
      let (ec4,_)= do_one_step ec3  in
        print_string ("c4= "^(to_string ec4)^"\n");
        print_string ("c4.aredex="^(Automaton.to_string ec4.aredex)^"\n");
        asserteq (string_equal (Automaton.to_string ec4.aredex)
                    "States q0 q1 \nFinal States q0 \nTransitions\na -> q0\nf(q0) -> q1\nEpsilonR=\nq1 -> q0\nq0 -> q1") true "adding epsilon rules no longer merge states q0 and q1";
        asserteq ec4.new_epsi [] "";
        asserteq ec4.new_trans [] "Automaton is complete 1";
        print_string "Verification pour equation s(x)=x\n"
          END

          (* autd={f(a), rule3=f(x) -> f(s(x)), eq2= s(x)=x *)
          IFDEF TEST THEN
let ed= new_step autd [rule3] [eq2] [] vl3 vl3
let (ed2,_)= do_one_step ed
let (ed3,_)= do_one_step ed2
let (ed4,_)= do_one_step ed3
  END


  IFDEF TEST THEN
  begin
    print_string ("ed = "^(to_string ed)^"\n");
    print_string ("ed2= "^(to_string ed2)^"\n");
    print_string ("ed3= "^(to_string ed3)^"\n");
    print_string ("ed4= "^(to_string ed4)^"\n");
    assertneq ed3.new_epsi [] "Automaton is not complete";
    asserteq ed4.new_epsi [] "Automaton is complete 2";
    asserteq ed4.new_trans [] "Automaton is complete 3";
    print_string ("Transition list= "^(Transitions.list_to_string (Transitions.to_list ed4.aredex.trans))^"\n")
  end
  END

  IFDEF TEST THEN
let pat= Fsym("f",[Fsym("s",[Fsym("s",[Fsym("s",[Var "z"])])])])
let pat2= Fsym("f",[Fsym("s",[Fsym("f",[Fsym("s",[Var "z"])])])])
  END


  IFDEF TEST THEN
  begin
    print_string "Test check_intersection on\n";
    print_string ("aredex= "^(Automaton.to_string ed4.aredex)^"\n");
    let res= check_intersection ed4.aredex [pat] ["z"] in
      print_string ("res="^(patternres_to_string res)^"\n");
      asserteq (patternres_to_string res) "f(s(s(s(z)))): q3 -> {{z -> q0}[Eq(q2,q0)]}\n"
        "only one solution. The fact that it is also a solution of q1 (because q3 -> q1) is detected by the
       pattern detection algo that searches all the predecessors of final states";
      let res2= check_intersection ed4.aredex [pat2] ["z"] in
        print_string ("res2="^(patternres_to_string res2)^"\n");
        asserteq res2 [] "No solution for this pattern";

        print_string "Completion step: pattern detection\n";
        (* autd={f(a), rule3=f(x) -> f(s(x)), eq2= s(x)=x *)
        let ee= new_step autd [rule3] [eq2] [pat;pat2] ["x";"z"] ["x";"z"]  in
        let _= print_string ("ee = "^(to_string ee)^"\n") in
        let (ee2,res1)= do_one_step ee  in
        let _= print_string ("ee2= "^(to_string ee2)^"\n") in
        let (ee3,res2)= do_one_step ee2  in
        let _= print_string ("ee3= "^(to_string ee3)^"\n") in
        let (ee4,res3)= do_one_step ee3  in
        let _= print_string ("ee4= "^(to_string ee4)^"\n") in
          print_string "Check intersection during completion\n";
          print_string ("Pattern res1= "^(patternres_to_string res1)^"\n");
          asserteq res1 [] "No pattern found";
          print_string ("Pattern res2= "^(patternres_to_string res2)^"\n");
          asserteq (patternres_to_string res2) "f(s(s(s(z)))): q3 -> {{z -> q0}[Eq(q2,q0)]}\n" "only one solution"
  end
  END


  IFDEF TEST THEN
  begin
    print_string "Check anonym variables in equations\n";
    let ef= new_step autb [rule1] [eq3] [] ["x";"y";"_"] ["x";"y";"_"]  in
      print_string (to_string ef);
      let (ef2,_)= do_one_step ef  in
        print_string (to_string ef2);
        let (ef3,_)= do_one_step ef2  in
	  print_string (to_string ef3);
	  print_string "\nTests for conditional rules\n"
  end
  END



  IFDEF TEST THEN
let t1= Fsym("f", [Fsym("a",[]);Fsym("b",[])]) (* f(a,b) *)
  (* f(x, y) -> f(s(x), s(y)) *)
let (autb,_,_)= Automaton.term_to_automaton [t1] [] stt
let rule11= (Fsym("f", [Var("X");Var("Y")]),Fsym("g",[Var("X")]),[Joinable(Var("X"),Var("Y"))])
let rule12= (Fsym("a", []),Fsym("b",[]),[])
let pat3= Fsym("g",[Var("_")])
let vl2= ["X";"Y"]
let ef= new_step autb [rule11] [] [pat3] vl2 vl2
let eg= new_step autb [rule11;rule12] [] [pat3] ("_"::vl2) ("_"::vl2)
  END

  IFDEF TEST THEN
  begin
    let (ef1,res1)= do_one_step ef  in
      print_string (to_string ef1);
      let (ef2,res2)= do_one_step ef1  in
        print_string (to_string ef2);
        asserteq res1 [] "g(_) is not found, rule does not apply";
        asserteq res2 [] "g(_) is not found, rule does not apply, even after 2 steps";
        let (eg1,res1)= do_one_step eg  in
        let (eg2,res2)= do_one_step eg1  in
        let (eg3,res3)= do_one_step eg2  in
        let (eg4,res4)= do_one_step eg3  in
          asserteq res1 [] "After first step conditional rule does not apply";
          asserteq res2 [] "After second step conditional rule applies but pattern is still not found";
          assertneq res3 [] "g(_) is found, condition rule has been applied";
          asserteq eg4.new_trans [] "Automaton is complete after 3 steps";
          asserteq eg4.new_epsi []  "Automaton is complete for epsi after 3 steps"
  end
  END






