(* Propositional formula built on and, or and implies connectives and propositional variables
   that are external epsilon, i.e. couples of states representing a transition between them.
   Formulas are built as BDDs *)

IFDEF TEST THEN
open Debug
ENDIF

open Definitions


let maxVar= ref 10000
module Buddy= Buddy.Make(struct let nodenum = 1000000 let cachesize = 100000 end)

(* This initial size is automatically increased if necessary *)
let _ = Buddy.setvarnum !maxVar

let gc_handler (flag_before_gc: bool) (gbs: Buddy.gbc_stat)=
  if flag_before_gc then Gc.compact() else print_string "Buddy: garbage collecting done\n"
  
let _= Buddy.gbc_hook gc_handler

let atom_to_string (q1, q2, k) =
  (match k with Eq -> "Eq(" | Rw -> "Rw(") ^
    (State.to_string q1) ^ "," ^ (State.to_string q2) ^ ")"

(* [vars2eq] associates a bdd var name to an atom *)
let vars2eq = Hashtbl.create !maxVar

(* [eq2vars] associates an atom to a bdd var name *)
let eq2vars = Hashtbl.create !maxVar



type formula = Buddy.t

type t = formula (* -> Formula.t *)



(* Numbering of variables in Bdd. Each atom will be associated to such
   variable numbers *)
let currentVar = ref 0

(* To void Bdd explosion, it is necessary to free Bdd variables when possible *)

let free_vars = ref []

let mk_true = Buddy.tt ()
let mk_false = Buddy.ff ()

(* bijection from atoms to Bdd variables *)
let atom_of_var (id: int) : atom =
  try Hashtbl.find vars2eq id with
    | Not_found ->
        failwith ("\nNo atom corresponds to "^(string_of_int id))

let var_of_atom (a: atom) : int =
  try Hashtbl.find eq2vars a with
    | Not_found -> 
        failwith ("\nNo variable corresponds to "^(atom_to_string a))


let mk_atom (q1: state) (q2: state) (k: kind) : formula =
  (* If the atom is known, we get back its variable number,
     otherwise we associate a new var number to it and add it in the
     tables *)
  let var =
    try Hashtbl.find eq2vars (q1, q2, k) with
      | Not_found ->
          let i =
            if !free_vars = [] then (
              let i = !currentVar in
                incr currentVar;
                i
            ) else (
              let i= List.hd !free_vars in
                free_vars := List.tl !free_vars;
                i
            )
          in
            Hashtbl.add vars2eq i (q1, q2, k);
            Hashtbl.add eq2vars (q1, q2, k) i;
            i
  in
    (* If we have reached the maximal variable number, we increase it
       before allocating the variable in Buddy *)
    if var >= !maxVar then (
      maxVar:= !maxVar*2;
      ignore (Buddy.setvarnum !maxVar)
    );
    Buddy.ithvar var


let kill_atom (q1: state) (q2: state) (k: kind) : unit =
  let idx = var_of_atom (q1, q2, k) in
    Hashtbl2.remove vars2eq idx;
    Hashtbl2.remove eq2vars (q1, q2, k);
    free_vars := idx :: !free_vars


let mk_and (f1: formula) (f2: formula) : formula =
  Buddy.apply f1 f2 Buddy.AND

let mk_or (f1: formula) (f2: formula) : formula =
  Buddy.apply f1 f2 Buddy.OR

let mk_not (f: formula) : formula =
  Buddy.not f

(** [remove f at] returns a formula that is equivalent to [f] where all
    occurences of [at] are removed *)
let rm_atom (f: formula) (atom: atom) : formula =
  let idx= var_of_atom atom in
  let bddidx= Buddy.ithvar idx in
    Buddy.forall f bddidx

let unsat (f1: formula) : bool =
  f1 = mk_false

let tautology (f1: formula) : bool =
  f1 = mk_true

let subsume (f1: formula) (f2: formula) : bool =
  tautology (Buddy.apply f1 f2 Buddy.IMP)


let equival f1 f2 : bool =
  tautology (Buddy.apply f1 f2 Buddy.BIIMP)

let is_sat (f: formula) : bool = f <> mk_false

let to_list (a: Buddy.value array): (atom * bool) list=
  let rec tolist i res =
    if i < 0 
    then res 
    else tolist (i - 1) (let value=Array.unsafe_get a i in
                           match value with
                               Buddy.F -> (atom_of_var i,false)::res
                             | Buddy.T -> (atom_of_var i,true)::res
                             | _ -> res)
  in
    tolist (Array.length a - 1) []

let allsat (f: formula) : (atom * bool) list list =
  Buddy.allsat (fun acc a -> (to_list a)::acc) [] f

(** [all_negatives alpha] returns all the list of valutations ([atom list])
    that satisfy [alpha]. But in each valuation, there is only negative
    atoms, positive ones are removed *)
let negatives (alpha: formula): atom list list =
  List.map
    (List.fold_left
       (fun atoms (a, p) ->
          if not p then a :: atoms else atoms
       ) []
    ) (allsat alpha)


let valuation_to_string (l: (atom * bool) list list) =
  List2.to_string ~sep:";"
    (List2.to_string ~sep:";"
       (fun (a, b) ->
          "(" ^ (atom_to_string a) ^ "," ^ (string_of_bool b) ^ ")"
       )
    ) l

(* Conversion of bdd representation of formulas to usual flat representation *)
let rec to_string (f: formula) : string =
  let litt_of v =
    try atom_to_string (Hashtbl.find vars2eq v) with
      | Not_found ->
          failwith "Formula.to_string ... should not happen 3"
  in
    if f = mk_true then
      "Top"
    else if f = mk_false then
      "Bot"
    else
      let lit = litt_of (Buddy.var f) in
      let h = Buddy.high f in
      let l = Buddy.low f in
        match l, h with
          | l1, h1 when l1 = mk_false && h1 = mk_true -> lit
          | l1, h1 when l1 = mk_true && h1 = mk_false -> "not(" ^ lit ^ ")"
          | l1, _ when l1 = mk_false -> "(" ^ lit ^ " /\\ " ^ (to_string h) ^ ")"
          | _, h1 when h1 = mk_true -> "(" ^ lit ^ " \\/ " ^ (to_string l) ^ ")"
          | l1, _ when l1 = mk_true -> "(not(" ^ lit ^ ") \\/ " ^ (to_string h) ^ ")"
          | _, h1 when h1 = mk_false -> "(not(" ^ lit ^ ") /\\ " ^ (to_string l) ^ ")"
          | _, _ ->
              "((" ^ lit ^ " /\\ " ^ (to_string h) ^ ") \\/ " ^
                (to_string l) ^ ")"




                (*
                (* Dot export with equation names *)
                  let dot_export (f: formula) (file: string) : unit =
                  try
                  Bdd.print_to_dot f
                  (fun i ->
                  atom_to_string (Hashtbl.find vars2eq i)
                  ) file
                  with
                  | Not_found ->
                  failwith "Formula.dot_export: unknown variable name in vars2eq"
                *)

                IFDEF TEST THEN
let f1 = mk_atom 1 0 Eq
let f2 = mk_atom 2 0 Eq
let f3 = mk_atom  3 0 Eq
let f4= mk_atom 4 0 Eq

let f1etf2 = mk_and f1 f2
let f1etf1 = mk_or f1 f1
let f3etf4 = mk_and f3 f4
let f1etf2ouf3etf4 = mk_or f1etf2 f3etf4

let f1etorf2 = mk_or f1etf1 f1etf2
let f12 = mk_and f1 f3
let f13 = mk_and f2 f3
let f4 = mk_or f1 f2
let f5 = mk_and f3 f4
let f6 = mk_or f1 f5
let f7 = mk_and f5 f5
let f8 = mk_and f3 (mk_or f1 f2)
let negf1etf2 = mk_not f1etf2

let f9 = mk_and (mk_and f1 f2) f3
let f10 = mk_and (mk_not f1) (mk_not f2)
let f11 = mk_or f9 f10
END



IFDEF TEST THEN
begin
  print_string "Test formula.ml\n";
(*  dot_export f1etf2 "formulef1etf2.dot";
  dot_export f12 "formulef1etf3.dot";
  dot_export f9 "f9.dot";
  dot_export f10 "f10.dot";
  dot_export f11 "f11.dot";
*)

  asserteq (rm_atom mk_true (1, 0, Eq)) mk_true "Test propagation negation - 1";
  print_string ("f1etf2= "^(to_string f1etf2)^"\n(rm_atom f1etf2 (1,0,Eq))= "^(to_string (rm_atom f1etf2 (1, 0, Eq)))^"\nmk_false= "^(to_string mk_false)^"\n");
  asserteq (rm_atom f1etf2 (1, 0, Eq)) mk_false "Test propagation negation -2";
  asserteq (rm_atom f8 (2, 0, Eq)) f12 "Test propagation negation -3";
  print_string ("f1etf2ouf3etf4= "^(to_string f1etf2ouf3etf4)^"\n(rm_atom f1etf2ouf3etf4 (1,0,Eq))= "^(to_string (rm_atom f1etf2ouf3etf4 (1, 0, Eq)))^"= f3etf4 = "^(to_string f3etf4)^"\n");
  asserteq (rm_atom f1etf2ouf3etf4 (1,0,Eq)) f3etf4 "Test rm_atom";

  print_string "Test all_sat\n";
  print_string (valuation_to_string (allsat f1etf2ouf3etf4));

(*  dot_export f8 "f8.dot";
  dot_export f5 "formula.dot"; 
*)

  asserteq (equival f8 f3) false "Test-1 mk_equ";
  asserteq (subsume f8 f3) true "Test-2 mk_equ";
  asserteq (subsume f3 f8) false "Test-3 mk_equ";
  asserteq (subsume f7 f2) false "Test1 subsume";
  asserteq (subsume f7 f3) true "Test2 subsume";
  asserteq (subsume mk_false f4) true "Test3 subsume";
  asserteq (subsume f4 mk_false) false "Test4 subsume";
  asserteq (subsume mk_true mk_true) true "Test 5 subsume";
  asserteq (subsume f4 mk_true) true "Test6 subsume";
end
END



