(** Module [List2] extends module [List] of the standard library with
    usefull stuffs.
*)

(** All function of [List] module can be invoked using by prefixing it
    with [List2] i.e. List2.fold_left instead of List.fold_left *)
include List


let cons x l = x::l



(** [rm_last l] returns [l] without its last element or fails if [l = []]. *)
let rec rm_last = function
  | [] -> failwith "List2.rm_last"
  | [_] -> []
  | x::rem -> x::(rm_last rem)


(** [ fmax f init l ] returns the greatest [f x] for all element [x]
    of [l], which is also greater than [init], or returns [init] else.
*)
let rec fmax f init =
  List.fold_left
    (fun cur x ->
       let y = f x in if y > cur then y else cur)
    init


(** [max init l] returns the greatest element [x] of [l], which is also
    greater than [init], or returns [init] else.
*)
let rec max init =
  List.fold_left
    (fun cur x -> if x > cur then x else cur)
    init


(** [fmap p f l] applies function f to [a1, ..., ak], elements of
    [l] that satisfy the predicate [p], and builds the list [f a1;
    ...; f ak] with the results returned by f. Not tail-recursive.
*)
let fmap p f =
  let rec aux_fmap = function
    | [] -> []
    | x::l -> if p x then f x::(aux_fmap l) else aux_fmap l
  in
    aux_fmap

(** [optmap f [ a1; ...; an ]] applies partial function f to [[a1, ..., an]],
    and builds the list [[f a1; ...; f ak]] with the results returned by f for
    whose are defined. Not tail-recursive.
*)
let optmap f =
  let rec aux_optmap = function
    | [] -> []
    | x::l ->
        match f x with
          | Some v -> v::(aux_optmap l)
          | None -> aux_optmap l
  in
    aux_optmap

(** Same as {!List.iter}, but for a two-argument function. Raise
    [Invalid_argument] if the two lists have different lengths. *)
let rec iter2 f l1 l2 =
  match (l1, l2) with
    | ([], []) -> ()
    | (a1::l1, a2::l2) -> f a1 a2; iter2 f l1 l2
    | (_, _) -> invalid_arg "List2.iter2"



(** Parametrizable formater to display lists. To use a standard display
    just invoke [to_string] using the appropiate formater corresponding to
    for elements of the list. For example, [to_string string_of_int [1;2;3]]
    returns the string ["[1; 2; 3]"]. Moreover, semi-colon, left and
    right marker may also be changed by specifying some of optional parameters:
    [to_string ~sep:" " ~left:"<" ~right:">" string_of_int [1;2;3]] builds the
    string ["<1 2 3>"].
*)
let to_string ?(sep = "; ") ?(left = "[") ?(right = "]") string_of = function
  | [] -> left ^ right
  | x::l ->
      left ^ (List.fold_left
                (fun str e -> str ^ sep ^ (string_of e))
                (string_of x) l
             ) ^ right

(** Set as List *)
let add x l =
  if List.mem x l then l else x::l

(** [l1] ⊆ [l2] *)
let subseteq l1 l2 =
  List.for_all (fun a -> List.mem a l2) l1

(** [l1] ⊆ [l2] and [l1] ⊇ [l2] *)
let equal l1 l2 =
  subseteq l1 l2 && subseteq l2 l1

(** Union of [l1] and [l2] without repetitions ([l2] supposed to be without
    several occurrence of the same element *)
let rec union l1 l2 =
  List.fold_left
    (fun l2 a ->if List.mem a l2 then l2 else a::l2)
    l2 l1

(* B: Aucune hypothese sur le nombre d'occurences des elements de l1 et l2....
   Est-ce normal ?
*)
(** Intersection of lists *)
let rec inter l1 l2 =
  List.fold_left
    (fun l a -> if List.mem a l2 then a::l else l)
    [] l2

(* retrieves repetitions in a list *)
let clean l = union l []

let add e l =
  if List.mem e l then l else e::l

(* B: A nouveau des fonctions sur les listes *)
let remove e l =
  List.fold_left
    (fun k x -> if x = e then k else x::k)
    [] l

let rec removetest e = function
  | [] -> raise Not_found
  | e2::rem ->
      if e2 = e then rem else e2::(removetest e rem)

(** [diff l1 l2] returns a list that contains elements of [l2] that are not in [l1]. *)
let diff l1 l2 =
  List.fold_left
    (fun k x -> if List.mem x l2 then k else x::k)
    [] l1

(* *)
let rec no_double = function
  | [] -> true
  | x::l -> not (List.mem x l) && (no_double l)

(** For the generation of new free indexes in transition tables *)
(* B: Pourquoi construire explicitement la liste des indexes ???? *)
let rec sequence i j =
  let rec aux_seq i j l =
    if i >= j then i::l else aux_seq i (j - 1) (j::l)
  in
    aux_seq i j []


(* B: le nom est hyper mal choisi: j'ai failli me planter!!!!
   Ca surcharge List.combine mais rien a voir!
   ca calcule un produit ici...
*)
(** Combine all elements of l1 with all elements of l2 *)
let prod k l =
  List.fold_right
    (fun x acc ->
       List.fold_right
         (fun y acc -> (x, y) :: acc) l acc
    ) k []





(* Assuming that [l] is a ordered list w.r.t. [(<)], [insert (<) x l]
   returns a list equivalent to [l] plus [x] inserted at the corresponding place
   w.r.t. [(<)]. Not tail-recursive.
*)
let insert cmp x =
  let rec aux = function
    | [] -> [x]
    | (y::l') as l -> if cmp x y then x::l else y::(aux l')
  in
    aux

