(* This module extends module Set of the standard library by adding
    a function "to_string" to dump the structure. Each element of the dumped
    set is convert to a string using a function [to_string] provided by
    a module of type [OrderedType], which is required by the functor [Make].
*)

(* Library str must be explicitly linked before using module [Str]
   which is required by function [to_string]
*)
module type OrderedType =
sig
  include Set.OrderedType

  val to_string: t -> string
end

module type S =
sig
  include Set.S

  val add_list: elt list -> t -> t
  val to_string :
    ?sep:string -> ?left:string -> ?right:string ->
    ?str_elt:(elt -> string) -> t -> string
  val replace : elt -> elt -> t -> t
end


module Make(Ord: OrderedType) =
struct

  include Set.Make(Ord)

  (* Insertion of several elements *)
  let add_list l s =
    List.fold_left (fun s x -> add x s) s l



  (** Parametrizable formater to display set content. To use the
      default format just invoke [to_string]. For example, [to_string]
      can be applied to any set of elements. As for {!List2.to_string},
      semi-colon, left and right markers, and format function for element
      can be specified by using optional parameters syntax:

      [to_string ~sep:" " ~left:"<< " ~right:" >>" ~str_elt:(fun _ -> "#")]
      returns a string like ["<< # # # ... # # # >>"].
  *)

  let to_string ?(sep = ",") ?(left = "{") ?(right = "}") ?(str_elt) set =
    let str_elt = match str_elt with
      | None -> Ord.to_string
      | Some f -> f
    in
    let str =
      fold (fun elt str -> str ^ (str_elt elt) ^ sep) set left
    in
      (if str = left then
         left
       else     (* Here, we know that str = "{e1, ... ,en,"
                   and the last [sep] must be removed    ^ *)
         String.sub str 0 (String.length str - (String.length sep))
      ) ^ right

  (* Replace an element by an other in a set *)

  let replace e1 e2 set=
    fold (fun elt res -> 
            if elt=e1 then add e2 res
            else add elt res) set empty

end
