(* ocamlbuild -cflags -dtypes -pp "camlp4o pa_macro.cmo -DTEST" main.byte *)
(* ocamlbuild -pp "camlp4o pa_macro.cmo -UTEST" main.native *)
open Term
open Equation
open Dataextracting
open Transitions
open Automaton

(* Property type: pattern or Automaton (automaton permits the automatic refinement) *)
type property = Pat | Aut of automaton

let delete_file (filename: string): unit=
  try (Sys.remove filename)
  with (Sys_error _) -> ()

let write_file
    (filename: string) (alpha: (string*int) list) (varset: string list)
    (trs: Rewriting.trs) (ainit: Automaton.automaton) (pl: term list)
    (el: equational_theory) (acomp: Automaton.automaton) : unit =
  let resstring =
    "% Completion time: "^ (string_of_float (Sys.time() -. !Common.start_time)) ^
      " seconds\n\n" ^ "Ops " ^
      (List.fold_right
         (fun (s, i) rem -> s ^ ":" ^ (string_of_int i) ^ " " ^ rem)
         alpha
         ("\n\nVars "^
	    (List.fold_right
               (fun x rem -> x ^ " " ^ rem)
	       varset
	       ("\n\nTRS R1\n" ^ (Rewriting.to_string trs) ^
	          "\n\nAutomaton Ainit\n" ^ (Automaton.to_timbuk ainit) ^
	          "\nAutomaton Acomp\n" ^ (Automaton.to_timbuk acomp) ^
	          (if List.length pl = 0 then
                     ""
                   else if !Common.subpattern then
                     "\nSubPatterns\n"
                   else
                     "\nPatterns\n"
                  ) ^ (Term.list_to_string pl) ^ "\n"
               ) ^
	       if List.length el = 0 then "" else
	         "Equations Approx\nRules\n" ^ (Equation.to_string el)^"\n"
            )
         )
      )
  in
  let outchan = open_out filename in
    try
      Printf.fprintf outchan "%s" resstring;
      close_out outchan
    with exc -> close_out outchan; raise exc



(* Check only the intersection, no completion step *)
let check_only ta pl trs_and_pl_vars al =
  match !Common.property_type with
    | Common.Patterns ->
        let allres = Completion.check_intersection ta pl trs_and_pl_vars in
          (* If not SubPatterns are considered we focus on final states and
	     predecessors w.r.t. epsilon trans of final states *)
        let allfiltered =
          if !Common.subpattern then allres else
            Completion.filter_final
              allres
              (Transitions.get_all_predPhi ta.fstates ta.trans)
              (Transitions.get_all_pred ta.fstates ta.trans)
        in
          (* no completion, only pattern check *)
	  print_string ("\n\nPattern check:\n--------------\n\n" ^
		          (Completion.patternres_to_string allfiltered) ^ "\n");
	  (false, ta, 0, [], false) (*Formula.mk_false*)
             (* don't care of the result in that case *)

    | Common.Automata propname ->
        let prop_ta =
          try Dataextracting.autassoc propname al with
            |  Dataextracting.Automaton_not_found ->
                 failwith ("Main:first_glance: No automaton of this name "^propname)
        in
        let phi = Automaton.inter_of' ta prop_ta in
	  if not(Formula.is_sat phi) then (
	    print_string "empty intersection = no forbidden term is reachable\n";
            (false, ta, 0, [], false (* Formula.mk_false*)) (* don't care of the result in that case *)
	  ) else (
	    print_string "At least one forbidden term is reachable\n";
	    (false, ta, 0, [], (* Formula.mk_false*) false)
          )

(* Main function call for completion: filename and maximal number of
   completion steps. Results into a bool saying if the automaton is
   complete, the completed automaton, the number of the last performed
   completion step, a list of found patterns and a bool saying if a counterexample has been found
*)

let first_glance (filename: string) (autname: string) (propname: string) (nbstep: int):
    (bool * Automaton.automaton * int * ((Term.term * Definitions.state * Substitution.substitutionset) list) * bool) =

  let (trsl, al, pl, dpl, el) = Dataextracting.extraction filename in
  let alpha = !Global.alpha and varset = !Global.varset in
    match trsl, al with
      | [], _ -> failwith "Main.first_glance: Missing Term Rewriting System\n"
      | _, [] -> failwith "Main.first_glance: Missing Tree Automaton\n"
      | (_, trs)::_, (_, tmpta, _)::_ ->
	  let ta =
            if autname = "" then tmpta
            else
              try Dataextracting.autassoc autname al with
	        | Dataextracting.Automaton_not_found ->
                    failwith "Main.first_glance: No automaton of this name"
	  in
	  let equ = try snd (List.hd el) with _ -> [] in
	    (* get all variables in patterns + trs on one side and equations *)
	  let trs_and_pl_vars =
            List.fold_right (fun t -> List2.union (Term.vars t))
	      pl (Rewriting.vars_trs trs)
          in
	  let eqvars = Equation.vars_eq equ in
	    if !Common.check_only then
              check_only ta pl trs_and_pl_vars al
	    else (* full completion *)
              let c = (* B: OK, l'algo commence seulement ici *)
                ref (Completion.new_step ta trs equ pl trs_and_pl_vars eqvars)
              in
              let prop =
                if propname = "" then
                  Pat
                else
                  try Aut (Dataextracting.autassoc propname al) with
	            | Dataextracting.Automaton_not_found ->
                        failwith ("Main.first_glance: No automaton of this name " ^ propname )
              in
	      let i = ref 1 and pat = ref [] and failed = ref false in
                (* and inter = ref Formula.mk_false in *)
	      let olddelta = ref (Transitions.make ()) in
	        (* begin *)
                Verbose.print 3 (fun _ -> "init cs=" ^ (Completion.to_string !c) ^"\n\n");
		while
                  (* tant que des transitions sont produites *)
                  not (!c.Completion.new_trans = [] && !c.Completion.new_epsi = []) &&
                    (* Tant qu'il n'y a pas de contre-exmples *)
                    not !failed && (* (Formula.tautology !inter) && *)
                    (* Tant que la limite n'est pas depassee *)
                    !i <= nbstep &&
                    (* tant qu'aucun pattern interdit n'a pas ete trouve... *)
                    !pat = []
                do
                  Verbose.println 3 (fun _ -> "Before computation: " ^ (Automaton.to_string !c.Completion.aredex));
                  let t1 = Sys.time () in
		  let (c2, pat2) = Completion.do_one_step !c in
		  let t2 = Sys.time () in
                    Verbose.println 3 (fun _ -> "Right after computation: " ^ (Automaton.to_string c2.Completion.aredex));
		    print_string ("\nCompletion step done in: "^(string_of_float (t2 -. t1))^" seconds\n");
                    print_string ("\n\nCompletion step: "^(string_of_int !i)^
				    "\n---------------------------------------------------------\n");
                    Verbose.print 3 (fun _ -> "Completion step at: " ^ (string_of_int !i) ^
				       ":\n" ^ (Completion.to_string !c) ^ "\n\n");
                    Verbose.println 2 (fun _ -> "Automaton at step: " ^ (string_of_int !i) ^
                                         "\n" ^ (Automaton.to_string !c.Completion.aredex));
		    print_string ("New transitions: "^(string_of_int (List.length c2.Completion.new_trans))^"\n");
		    print_string ("New epsilon: "^(string_of_int (List.length c2.Completion.new_epsi))^"\n\n");
		    flush stdout;
		    Verbose.print 0
                      (fun _ -> "Transitions to add: " ^
                         (Transitions.symextlist_to_string c2.Completion.new_trans)^"\n\n");
		    Verbose.println 0
                      (fun _ -> Transitions.epsiextlist_to_string c2.Completion.new_epsi);
		    pat := pat2;
                    Verbose.print 0
                      (fun _ -> Transitions.to_string c2.Completion.aredex.trans);
                    if !Common.interactive then ignore(Unix.system "read");
                    (*  0 => aredex raffin;
                     *  1 => pas de collision avec a_prop;
                     * -1 => il existe un contre-example *)
                    let ret =  (* if not(!i mod 6 = 0) then 1 else *)
                      match prop with
                        | Pat -> 1 (* AH si y a pas d'automate y a pas de raffinement...
                                           donc on suppose que l'intersection est vide *)
                        | Aut a_prop ->
                            Verbose.println 3
                              (fun _ -> "Automaton aredex for intersection : "^
                                 (Automaton.to_string c2.Completion.aredex));
                            Verbose.println 3
                              (fun _ -> "Automaton prop for intersection : "^
                                 (Automaton.to_string a_prop));
                            let t1 = Sys.time () in
                            let ret = Automaton.refine c2.Completion.aredex a_prop in
                            let t2 = Sys.time () in
                              if ret = 0 then
                                Format.printf "Refinement done in %f\n@?" (t2 -. t1)
                              else
                                Format.printf "Intersection done in %f\n@?" (t2 -. t1);
                              ret
                    in
                      if !Common.interactive then ignore (Unix.system "read");
                      failed := ret = -1;
                      if ret = -1 || ret = 1 then ( (* si on n'a pas raffin *)
                        (* if not (Formula.is_sat !inter) || (Formula.tautology !inter) then ( *)
                        olddelta := !c.Completion.aredex.Automaton.trans;
                        c := c2;
                        incr i
                          (* B: counter [i] is only incremented if verification does not fail!!! *)
                      ) else (
                        c := Completion.reset_step c2;
                        Verbose.exec 3 (
                          fun _ ->
                            print_string ("c after reset= "^(Completion.to_string !c)^"\n");
                            print_string ("aredex after reset "^(Automaton.to_string !c.Completion.aredex)^"\n")
                        )
                      )
                done;
                if !failed then
                  Format.printf "Found a counterexample!\nIntersection with automaton %s not empty!\n@?" propname;
                let fnameres= (Filename.chop_extension filename) in
                  Verbose.exec 2 
                    (function _ -> 
                       print_string ("Done\nExport "^fnameres^".full file...");
                       (write_file (fnameres^".full") alpha varset trs ta pl equ !c.Completion.aredex));
                  print_string ("Done\nExport "^fnameres^".res.epsi file...");
                  (try 
                     (write_file (fnameres^".res.epsi") alpha varset trs ta pl equ 
                        (Automaton.export_epsi !c.Completion.aredex))
                   with e -> print_string "\nCannot write .res.epsi file!\n";
                     raise e);
                  print_string ("Done\n");
                  if !Common.export then
                    begin
                      print_string ("Export "^fnameres^".res file...");
                      (try 
                         (write_file (fnameres^".res") alpha varset trs ta pl equ
                            (Automaton.export !c.Completion.aredex))
                       with e -> print_string "\nCannot write .res file!\n";
                         raise e);
                      print_string "Done\n"
                    end
                  else
                    delete_file (fnameres^".res");
                  print_string ("Export "^fnameres^".dot file...");
                  (try
                     Dot.write_dot (fnameres^".dot") !c.Completion.aredex dpl
                   with e -> print_string "\nCannot write .dot file!\n";
                     raise e);
                  print_string "Done\n";
                  ((!c.Completion.new_trans = [] && !c.Completion.new_epsi = []),
                   !c.Completion.aredex, !i - 1,  !pat, !failed)
                    (* end *)

                    (* Main program *)
                    (* Main program can be found after unit testing for clarity reasons (outputs of
                       unit testing must appear before normal outputs of the program *)




                    (* for debug with -g

                       first_glance "nspk2.txt" 5
                    *)


                    (* ------------------------------TESTS ----------------------------------------*)


IFDEF TEST THEN
open Debug

let b,_,i,p,_= first_glance "tests/TU-ex1.txt" "" "" 6 in
  print_string "tests/TU-ex1.txt\n";
  asserteq b true "ex1 Automaton completed";
  asserteq i 5 "With exactly 5 steps";
  asserteq p [] "No pattern is found";

let _= print_string "tests/TU-ex2.txt\n" in
let b,a,i,_,_= first_glance "tests/TU-ex2.txt" "" "" 6 in
  print_string "tests/TU-ex2.txt\n";
  asserteq b true "ex2 Automaton completed";
  asserteq i 2 "With exactly 2 steps";

let _= print_string "tests/TU-ex3.txt\n" in
let b,_,i,p,_= first_glance "tests/TU-ex3.txt" "" "" 6 in
  print_string "tests/TU-ex3.txt\n";
  asserteq b false "Not completed";
  asserteq i 4 "Pattern found on 4th step";
  asserteq (List.length p) 2 "The two patterns are found";

let _= print_string "tests/TU-ex4.txt\n" in
let b,_,i,p,_= first_glance "tests/TU-ex4.txt" "" "" 10 in
  print_string "tests/TU-ex4.txt\n";
  asserteq b true "Completed";
  asserteq i 6 "Pattern not found on 6th step";

let _= print_string "tests/TU-ex5.txt\n" in
let b,_,i,p,_= first_glance "tests/TU-ex5.txt"  "" "" 10 in
  print_string "tests/TU-ex5.txt\n";
  asserteq b false "Not completed";
  asserteq i 2 "Pattern found on 3rd step. TU-ex5.txt";
  asserteq (List.length p) 2 "The solutions are found";

let _= print_string "tests/TU-ex5.txt... without approx\n" in
let _= Common.approx_type:= Common.None in (* unset the approx flag *)
let b,_,i,p,_= first_glance "tests/TU-ex5.txt"  "" "" 10 in
  print_string "tests/TU-ex5.txt\n... again";
  asserteq b false "Not completed";
  asserteq i 10 "Run until the 10th completion step";
  asserteq p [] "No pattern found";
  Common.approx_type:= Common.Equationnal (* set the approx flag *)



let _= print_string "tests/TU-ex6.txt\n" in
let b,_,i,_,_= first_glance "tests/TU-ex6.txt" "" "" 4 in
  print_string "tests/TU-ex6.txt\n";
  asserteq b true "Completed!";
  asserteq i 3 "Completed on the 2nd step. TU-ex6.txt";

let _= print_string "tests/TU-ex7.txt\n" in
let b,_,i,_,_= first_glance "tests/TU-ex7.txt" ""  "" 4 in
  print_string "tests/TU-ex7.txt\n";
  asserteq b true "Completed!";
  asserteq i 3 "Completed on the 2nd step. TU-ex7.txt"

let _= print_string "tests/TU-ex8.txt\n" in
let b,_,i,_,_= first_glance "tests/TU-ex8.txt" "" "" 6 in
  print_string "tests/TU-ex8.txt\n";
  asserteq b false "Not completed!";
  asserteq i 6 "Still not completed on the 6th step"

let _= print_string "tests/TU-ex9.txt\n" in
let b,_,i,_,_= first_glance "tests/TU-ex9.txt" "" "" 6 in
  print_string "tests/TU-ex9.txt\n";
  asserteq b true "Completed! and Pattern not Found (look only for final states)";
  asserteq i 4 "Completed on the 4th step"

let _= print_string "tests/TU-ex10.txt\n" in
let b,c,i,p,_= first_glance "tests/TU-ex10.txt" "" "" 6 in
  print_string "tests/TU-ex10.txt\n";
  Transitions.print_delta c.trans;
  asserteq b true "Completed! and SubPattern found (look in all states)";
  asserteq i 4 "Completed on the 4th step";
  asserteq (List.length p) 2 "Two solutions found"

let _= print_string "tests/TU-ex11.txt\n" in
let b,_,i,p,_= first_glance "tests/TU-ex11.txt" "" "" 2 in
  print_string "tests/TU-ex11.txt\n";
  asserteq b true "Completed! and Pattern found (look predecessors of final states)";
  asserteq i 2 "Completed on the 2nd step";
  asserteq (List.length p) 1 "One solution found"

let _= print_string "tests/TU-ex12.txt\n" in
let b,_,i,p,_= first_glance "tests/TU-ex12.txt" "" "" 3 in
  print_string "tests/TU-ex12.txt\n";
  asserteq b true "Completed! and Pattern not found!";
  asserteq i 2 "Completed on the 2nd step";
  asserteq (List.length p) 0 "Zero solution found"

let _= print_string "tests/TU-ex13.txt\n" in
let b,_,i,p,_= first_glance "tests/TU-ex13.txt" "" "" 3 in
  print_string "tests/TU-ex13.txt\n";
  asserteq b true "Completed! and Pattern found!";
  asserteq i 2 "Completed on the 2nd step";
  asserteq (List.length p) 1 "One solution found";
  let (_,q,es)=List.hd p in
    asserteq q 0 "Found on a final state";
    asserteq (Substitution.SubstitutionSet.to_string es) "{{}[Eq(q2,q1)]}" "with a formula"

let _= print_string "tests/TU-ex14.txt\n" in
let b,_,i,p,_= first_glance "tests/TU-ex14.txt" "" "" 3 in
  asserteq b false "Not completed! and Pattern found!";
  asserteq i 2 "Pattern found on the 2nd step";
  asserteq (List.length p) 1 "One solution found";
  let (_,q,es)=List.hd p in
    asserteq q 1 "Found on a final state";
    asserteq (Substitution.SubstitutionSet.to_string es) "{{}[Eq(q3,q2)]}" "with a formula"

let _= print_string "tests/TU-ex14.txt\n... without approx\n" in
let _= Common.approx_type:= Common.None in (* unset the approx flag *)
let b,_,i,p,_= first_glance "tests/TU-ex14.txt" "" "" 10 in
  asserteq b false "Not completed!";
  asserteq i 10 "Run until the 10th completion step";
  asserteq p [] "No pattern found";
  Common.approx_type:= Common.Equationnal (* set the approx flag *)

(* Completion behave the same when no refinement is necessary *)
let _= print_string "tests/TU-ex17.txt\n" in (* Same as TU-ex14.txt but with the right equation *)
let b,_,i,p,_= first_glance "tests/TU-ex17.txt" "" "" 10 in
  asserteq b true "Complete completed!";
  asserteq i 3 "Completed at the 3rd step";
  asserteq p [] "No pattern found"

(* Automatic refinement *)
(*
let _= print_string "tests/TU-ex15.txt\n" in (* Same as TU-ex14.txt but with the right equation *)
let b,_,i,p,contrex= first_glance "tests/TU-ex15.txt" "" "A1" 10 in
  asserteq b true "Completed!";
 asserteq i 3 "Completed at the 3rd step";
  asserteq contrex false "No contrex found"
*)
(* Counter example detection *)

let _= print_string "tests/TU-ex18.txt\n...\n" in
let b,_,i,p,counterEx= first_glance "tests/TU-ex18.txt" "" "A1" 10 in
  asserteq b false "Not completed!";
  asserteq i 1 "Counter-example found at 1st completion step";
  asserteq counterEx true "Counter ex found"


let _= print_string "tests/TU-ex18.txt\n... without approx\n" in
let _= Common.approx_type:= Common.None in (* unset the approx flag *)
let b,_,i,p,counterEx= first_glance "tests/TU-ex18.txt" "" "A1" 10 in
  asserteq b false "Not completed!";
  asserteq i 1 "Counter-example found at 1st completion step";
  asserteq counterEx true "Counter ex found";
  Common.approx_type:= Common.Equationnal (* set the approx flag *)

    

let _= print_string "tests/TU-ex19.txt\n... parsing\n" in
  try(
    let _,_,_,_,_= first_glance "tests/TU-ex19.txt" "" "A1" 10 in
      ()
  )
  with 
      Dataextracting.SpecialStateDeclaration -> 
	begin 
	  asserteq true true  "Detection of the bad use of a special state";
	  ();
	end
    | _ -> 
	
	begin
	  asserteq true false "No detection of the use of a special state";
	  ()
	end




let _= print_string "tests/TU-ex20.txt\n... parsing\n" in
  try(
    let _,_,_,_,_= first_glance "tests/TU-ex20.txt" "" "A1" 10 in
      ()
  )
  with 
      Dataextracting.SpecialStateAsRHS -> 
	begin 
	  asserteq true true  "Detection of a qTF in the RHS of a transition";
	  ();
	end
    | _ -> 
	
	begin
	  asserteq true false "Problem in the detection of a qTF in the RHS of a transition";
	  ()
	end



let _= print_string "tests/TU-ex21.txt\n... term reachable\n" in
let b,_,i,p,counterEx= first_glance "tests/TU-ex21.txt" "" "ToBeFound" 10 in
  asserteq b false "Not completed!";
  asserteq i 3 "Counter-example found at 3rd completion step";
  asserteq counterEx true "Pattern found"


let _= print_string "tests/TU-ex21.txt\n... term reachable\n" in
let b,_,i,p,counterEx= first_glance "tests/TU-ex21.txt" "" "NotFound" 10 in
  asserteq b true "Completed!";
  asserteq i 7 "After the 7th completion step";
  asserteq counterEx false "No Pattern found"

let _= print_string "tests/TU-ex22.txt\n" in
let b,a,i,p,counterEx= first_glance "tests/TU-ex22.txt" "" "Bad" 10 in
  asserteq b true "Completed!";
  asserteq i 2 "After the 2nd completion step";
  asserteq (Automaton.all_run a.trans (Fsym ("f",[Fsym ("s", [Fsym ("a",[])])]))) [] "f(s(a)) n'est pas reconnu"

END

(* ------------------------------ Main program ------------------------------ *)



let rec apply_option_list : string list -> unit = function
  | [] -> ()
  | "--new"::rem -> Common.completion_strategy:=Common.New;
      apply_option_list rem
  | "--prior"::rem -> Common.completion_strategy:=Common.Prior;
      apply_option_list rem
  | "--epsi"::rem -> Common.completion_type:=Common.Epsi;
      apply_option_list rem
  | "--noepsi"::rem -> Common.completion_type:=Common.Noepsi;
      (* In noepsi mode we also merge loops *)
      Common.merge_loop:=true;
      apply_option_list rem
  | "--interactive"::rem -> Common.interactive:=true;
      apply_option_list rem
  | "--equation"::rem -> Common.approx_type:=Common.Equationnal;
      apply_option_list rem
  | "--noapprox"::rem -> Common.approx_type:=Common.None;
      apply_option_list rem
  | "--check"::rem -> Common.check_only:= true;
      apply_option_list rem
  | "--export"::rem -> Common.export:= true;
      apply_option_list rem
  | "--autname"::s::rem -> Common.autname:=s;
      apply_option_list rem
  | "--propname"::s::rem -> Common.property_type:=Common.Automata s;
      apply_option_list rem

  | "--verbose"::s::rem ->
      (try
	 let vals = int_of_string s in
	   if vals >= 0 && vals <= 4 then (
	     Common.verbose:=vals;
	     apply_option_list rem
	   ) else
	     failwith ("--verbose option expects 0, 1, 2, 3 or 4 as parameter, given: "^s)
       with Failure _ ->
         failwith ("--verbose option expects 0, 1, 2, 3 or 4 as parameter, given: "^s)
      )
  | "--dotson"::rem -> Common.dotson:=true;
      apply_option_list rem
  | "--nodotson"::rem -> Common.dotson:=false;
      apply_option_list rem
  | "--exact"::rem -> Common.exact_completion:=true;
      apply_option_list rem
  | ["--autname"] -> failwith "--autname option should be followed by an automaton name"
  | ["--propname"] -> failwith "--propname option should be followed by an automaton name"
  | ["--verbose"] -> failwith "--verbose option should be followed by an integer value between 0 and 4"
  | s::_ -> failwith ("Unknown timbuk option: "^s)

let online_help =
"Usage= timbuk [option] <max number of completion steps> <timbuk spec file>
with options=
             --exact                        : uses exact normalization strategy
             --noapprox                     : does not use approximations defined
                                              in the file spec
             --noepsi                       : uses old style completion and also
                                              sets --mergeloop to true
             --autname <name in spec file>  : name of automaton to use as initial
             --propname <name in spec file> : name of automaton to use as property
             --export                       : triggers exportation without epsilon 
                                              transitions 
             --check                        : does not perform completion but
                                              only search for patterns in initial
                                              tree automaton
             --new                          : uses only new states to normalize
             --verbose <value>              : verbose mode 0 (default),1,2,3,4
             --dotson                       : direct subterm edges are exported
                                              the comp.dot file in red

             --epsi(default)                : uses new style completion with
                                              epsilon transitions
             --equation(default)            : uses approximation equations
             --prior(default)               : uses existing states/transitions
                                              when possible
             --nodotson(default)            : subterm edges are not exported
"

let arg_number = Array.length Sys.argv in
  if arg_number = 1 then (
    print_string online_help;
    exit 0
  );
  let l = ref [] in
    (* Ca ne serait pas plus precis de commencer les tests apres le parsing et l'extraction ??? *)
    (* Common.start_time := Sys.time(); *)
    (* C'est quoi l'interet de convertir un tableau en liste ???? *)
    if arg_number > 2 then (
      for i = arg_number - 3 downto 1 do
        l:= Sys.argv.(i)::!l
      done;
      apply_option_list !l
    );

    let nb =
      try int_of_string Sys.argv.(arg_number - 2) with
        | Failure "int_of_string" ->
            print_string ("Bad number of completion steps: " ^
                            (Sys.argv.(arg_number - 2))^"\n");
            exit 0 (* Valeur retournee dans la version precedente *)
    in
    let file = Sys.argv.(arg_number - 1) in
      if not (Sys.file_exists file) then (
        print_string "File does not exist!\n";
        exit 0
      );
      let propname = match !Common.property_type with
	| Common.Patterns -> ""
	| Common.Automata yoh -> yoh (* :-) *)
      in

	try (
	  let (completed, auto, i, patres, phi) =
            first_glance file !Common.autname propname nb
	  in
	    if not !Common.check_only then (
              (* On ne peut pas afficher l'automate quand les formules sont trop grosses *)
	      (*	  print_string (Automaton.to_string auto); (* Automaton.unepsilon a; *) *)
	      print_string ("\nStep: "^(string_of_int i)^"\n");
	      if completed then
		print_string "\n\nAutomaton complete!\n------------------\n";
	      if patres <> [] then (
		print_string "\n\nPattern found!\n--------------\n";
		print_string (Completion.patternres_to_string patres);
		if !Common.property_type= Common.Patterns then print_string "NO REFINEMENT!\nFor automatic refinement use the --propname option\n"
	      ) else if (*Formula.tautology phi*) phi then
		begin
		  print_string ("\nFormula:"^(string_of_bool phi)^"\n");
		  print_string "\nCounterexample found!\n---------------------\n\n";
		  if !Common.property_type= Common.Patterns then print_string "No property automaton was given, automatic refinement was not possible!\n"
		end
              else
		begin
		  if completed
		  then print_string "\nProof done!\n-----------\n\n";
		  print_string ("\nFormula:"^(string_of_bool phi)^"\n");
		end;
              let fnameres= Filename.chop_extension file in
		print_string "\nExported ";
                if !Common.export then print_string (fnameres^".res, ");
                print_string (fnameres^".res.epsi, "^fnameres^".dot files\n");

		print_string ("Completion time: " ^ (string_of_float (Sys.time() -. !Common.start_time))^" seconds\n");
		print_string ("Time in Rclosure computations: " ^ (string_of_float !Common.rclosure_time)^" seconds\n");
		print_string ("Time in approximation closure computations: " ^ (string_of_float !Common.approx_time)^" seconds\n")
	    )
	) 
	with 
	  | Dataextracting.SpecialStateAsRHS -> 
	      begin 
		print_string "\nThe state qTF must not appear in the RHS of a transition\n";
		exit(-1);
	      end
	  | Dataextracting.SpecialStateInEpsilon -> 
	      begin 
		print_string "\nThe state qTF must not appear in a rewriting transition\n";
		exit(-1);
	      end
	      
	  | Dataextracting.SpecialStateDeclaration -> 
	      begin 
		print_string "\nThe state qTF is a special state that can not be declared\n";
		exit(-1);
	      end

