(** The goal of this file is to extract data from a parsed
    specification. The parser ensures several properties:

    -- arity of functionnal symbols (in automata, sets, trs,
       approximations) fits with the signature

    -- an initial set of terms is ground over the set of functionnal
       symbols

    -- there is no state in the TRS.

    -- there is no variable in transitions of a tree automaton. 

    -- each functionnal symbol or variable has been declared in the Ops
       or Variables section.
   

    Main function: [extraction]. 

    It returns [(trs, automatalist, approximationlist)] where: 

    -- [trs] if the term rewriting system parsed and checked;

    -- [automatalist] is a list of tree automata defined from a
       sequence of both automata specifications or a terms sets
       specifications.

    -- [approximationlist] is a list of approximations. (* B: pas tres informatif :-p *)

*)

open Global
open Definitions
open Rewriting
open Automaton

exception BadAutomaton of string
exception ApproxFormatNotHandled of string
exception Automaton_not_found
exception SpecialStateDeclaration
exception SpecialStateInEpsilon
exception SpecialStateAsRHS

(* Counter for relabelling of transitions for the --exact normalization *)
let counter = ref 0

(* Initialisation of correspondance table for "new" states used for
   --exact completion. Those states should be of smaller
   index... thus we begin the numbering by those ones... *)
let rec init_corres (sl: string list) (d: Transitions.t) (corres: (string * state) list): (string * state) list=
  match sl with
    | [] -> corres
    | s1::rem ->
	let ns= Transitions.make_state d in
	  init_corres rem d ((s1,ns)::corres)

(*
  The following function must be called before the 
  update_correspondence_table. 
  It returns a set of final states and a correspondence table which must be 
  updated with the list of states of the parsed automaton.

*)

let rec initialize_final_states_set (fsl: string list) (sl:string list) (d:Transitions.t) (res: Stateset.t) (corres: (string * state) list): (Stateset.t * ((string * state) list)) = 
  match fsl with 
    | [] -> (res, corres)
    | cs::rem -> 
      if List.mem cs sl then 
	if List.mem cs rem then 
	   initialize_final_states_set rem sl d res corres 
	else
	  let ns = Transitions.make_state d in 
	  let nres = Stateset.add ns res in 
	  let ncorres = (cs,ns)::corres in 
	    initialize_final_states_set rem sl d nres ncorres 
      else
	raise (BadAutomaton "A final state is not declared in the states section.")


(* The following function updates the correspondence table. 
   The arguments corres is the result of the function 
   initialize_final_states_set called with the following parameters: 
   fls sl d Stateset.empty []
*)
let rec update_correspondence_table
    (sl: string list) (fsl: string list) (corres: (string * state) list) (d: Transitions.t) : (string * state) list = 
  match sl with 
    | [] -> corres 
    | cs::rem -> 
	if List.mem cs rem then 
	  update_correspondence_table rem fsl corres d 
	else if List.mem cs fsl then 
	  update_correspondence_table rem fsl corres d 
	else
	  let ns = Transitions.make_state d in 
	  let ncorres = (cs, ns)::corres in 
	    update_correspondence_table rem fsl ncorres d 





(* The following function builds the set of transitions recognizing
   T(F) into the state qTF *)
let rec extended_set_of_transitions_TF (delta:Transitions.t) (corres:(string*state) list): unit =
  let alpha = !Global.alpha in
  let qTFF = List.assoc Global.specialState corres in
  let rec construction_of_n_list = function
    | 0 -> []
    | n -> qTFF::(construction_of_n_list (n - 1))
  in
  let rec for_each_symbol = function
    | [] -> ()
    | (f, a)::rem ->
	let args = construction_of_n_list a in
	  Transitions.add delta (f,args,qTFF);
	  for_each_symbol rem
  in
    for_each_symbol alpha

(* The following function builds a set of transitions from the one
   parsed. *)
let transitions_set_building
    (transitions: (pTerm * pTerm) list) (lst: pTerm list)
    (corres: (string * state) list) (delta: Transitions.t) : unit =
  let state_of_string x =
    List.assoc (Global.PTerm.to_string x) corres
  in
    List.iter
      (fun (lhs, rhs) ->
	 if List.mem rhs lst then
	   match lhs with
	     | PFsym (f, args) ->
		 if List.for_all (fun x -> List.mem x lst) args then
                   let qargs = List.map state_of_string args
                   and qdest = state_of_string rhs in
                     Transitions.add delta (f, qargs, qdest)
		 else
		   failwith
                     "transitions_set_building: Some of states used are not declared."
	     | PState q ->
		 if List.mem lhs lst then
                   Transitions.addR delta
                     (List.assoc q corres, state_of_string rhs, Formula.mk_true)
                 else
		   failwith ("The state " ^ (Global.PTerm.to_string lhs) ^ " is not declared.")
	     | _ ->
                 failwith "There is a variable in the lhs of a transition."
	 else
	   failwith ("The state " ^ (Global.PTerm.to_string rhs) ^
                       " is not declared in the states section.")
      ) transitions


let rec build_a_set_of_transitions_for_each_term
    (id: string) (l: pTerm list) (d:Transitions.t)
    (finalstates: Stateset.t) : Automaton.automaton =
  match l with
    | [] -> { trans = d; fstates = finalstates }
    | t::rem ->
        try
	  let (qf, _) =
            Automaton.normalize_term d (Term.pTerm_to_term t) false
          in
          let nfinalstates = Stateset.add qf  finalstates in
	    build_a_set_of_transitions_for_each_term id rem d nfinalstates
        with
	  | _ ->
              failwith ("Check the initial term set " ^ id ^
                          ": states or variables may occur within.")



(* Those function produce a copy of the initial transition function so that
   for each transition t -> q we add t -> ###iq where i is incremented each time
   so that the corresponding transitions fulfill the --exact requirements.
   Since states of the form ###iq will also be the first to be numbered, those
   renamed transitions will be the first to be used for --prior normalization...
   This encode the --exact principle.
*)
let dupplicate_transition (l, r: pTerm * pTerm) : string * (pTerm * pTerm) =
  match r with
    | PState x ->
	let news = "###" ^ (string_of_int !counter) ^ x in
	  incr counter;
	  news, (l, PState news)
    | _ ->
        failwith "Dataextracting:dupplicate_transitions: unexpected term!"

let dupplicate_trans_list (l: (pTerm * pTerm) list) : string list * (pTerm * pTerm) list =
  List.fold_right
    (fun tr acc ->
       let id, tr' = dupplicate_transition tr in
         id::(fst acc), tr'::(snd acc)
    )
    l ([], [])


(** This function builds a 3-uplet [(name, automaton ,l)] where [name]
    is the identifier of the parsed automaton, automaton is the result
    after filling 

    and what is l???
*)
let automaton_construction (aut: pAutomaton):
    (string * Automaton.automaton * (string * state) list) =
  match aut with 
    | PAut (id, templst, lfst, prior, temptrans) -> 
        (* let _= print_string "temptrans="; List.iter (function l,r
           -> print_string ((PTerm.to_string l)^" -> "^(PTerm.to_string
           r)^"\n")) temptrans in *)
	let d = Transitions.make () in
 	let empty_states = Stateset.empty in
	let st = List.map Global.PTerm.to_string templst in
	let completeTF = List.mem Global.specialState st in
	let fst = List.map Global.PTerm.to_string lfst in
	let (corres, lst, trans) =
	  if !Common.exact_completion then
	    let (news, newt) = dupplicate_trans_list temptrans in
	      (init_corres news d [],
	       (List.map (fun s -> PState s) news) @ templst,
	       newt @ temptrans
              )
          else
            ([], templst, temptrans)
	in
          (* let _= print_string "trans="; List.iter (function l,r ->
             print_string ((PTerm.to_string l)^" -> "^(PTerm.to_string
             r)^"\n")) trans in *)
	let (finalstates, corres2) = 
          initialize_final_states_set fst st d empty_states corres in 

	let correspondence_table = 
          update_correspondence_table st fst corres2 d in 
          (* let _= print_string "corres=" ; List.iter (function
             (s,s1)-> print_string (s^"="^(string_of_int s1)^", "))
             correspondence_table in *)
	  begin
	    transitions_set_building (prior @ trans) lst correspondence_table d;
	    if completeTF then
	      extended_set_of_transitions_TF d  correspondence_table;
	    (id, {trans = d; fstates = finalstates}, correspondence_table)
	  end

    | PSet (id, termslist) ->
	let d = Transitions.make () in
	let empty_states = Stateset.empty in
	let baut =
          build_a_set_of_transitions_for_each_term id termslist d empty_states
        in 
	  (id, baut, [])


let rec autassoc s l : automaton =
  match l with 
    | [] -> raise Automaton_not_found
    | (n, a, _)::rem -> if s = n then a else autassoc s rem


let approx_of_parser_format = function
  | PApp (id, dec, rules) -> 
      raise (ApproxFormatNotHandled "Normalization rules are not handled for now.")
  | PSEquation (id, eql) ->
      id, List.map Equation.pEquation_to_equation eql

(* Timbuk Specification Files parsing *)

(** L'[extraction] consiste a ouvrir le fichier dont le nom [nomFichier] est passe en argument.
    Lorsque [nomFichier] est une chaine vide, la fonction lit puis parser l'entree standard.

    Dans l'ordre, la fonction passe le parser sur le fichier [channel] puis construit une 
    une representation intermediaire definit dans le module [Global], aussitot convertit
    dans la representation utilisee par la completion.
*)
let extraction nomFichier =
  let lexbuf =
    Lexing.from_channel (if nomFichier = "" then stdin else open_in nomFichier)
  in
  try
    let (trss, automata, patterns, dotpatterns, approxs) =
      Parser.timbukspec Lexer.token lexbuf
    in
      (List.map (fun (id, trs) -> id, Rewriting.trs_from_parser_format trs) trss,
       List.map automaton_construction automata,
       List.map Term.pTerm_to_term patterns,
       List.map Term.pTerm_to_term dotpatterns,
       List.map approx_of_parser_format approxs
      )
  with
    | BadAutomaton s
    | NotHandledYet s

    | Rewriting.InclusionTestFailed s
    | Rewriting.RecapError s

    | Equation.EquationMember s
    | Equation.ProblemWithConditions s ->
        print_string (s^"\n");
        exit (-1)

    | Equation.ContextualEquationNotHandledYet ->
	print_string ("Sorry, contextual equations are coming next.\n");
	exit (-1)
    | Global.SpecialStateDeclaration -> raise SpecialStateDeclaration
    | Global.SpecialStateInEpsilon -> raise SpecialStateInEpsilon
    | Global.SpecialStateAsRHS -> raise SpecialStateAsRHS
    | _ -> ErrorHandler.handleParseError lexbuf


IFDEF TEST THEN
(* let resultat_extraction =  extraction "TU-ex0.txt" *)

let pretty_string_for_parsed_automata l =
  List.fold_left
    (fun str (name, aut, _) ->
       str ^ "Automaton " ^ name ^ ":\n" ^
         (Transitions.to_string aut.trans) ^ "\n"
    )
    "" l

let pretty_string_for_parsed_trss =
  List.fold_left
      (fun str (name, rewr) ->
	 str ^ "TRS " ^ name ^ ":\n" ^
           (Rewriting.to_string rewr) ^"\n"
      )
    ""

let pretty_string_for_parsed_data (trss, automata, _) =
  (pretty_string_for_parsed_trss trss) ^ "\n" ^ 
    (pretty_string_for_parsed_automata automata)

ENDIF




(************************************ TESTS ***********************************)

IFDEF TEST THEN
open Debug

let d_yoh = Transitions.make ()
let state_spec = ["qa"; "qb"; "qc"; "qd"; "qf"]
let final_state_spec = ["qf"]
let set_yoh = Stateset.empty

let (set_yoh1, corres) =
  initialize_final_states_set final_state_spec state_spec d_yoh set_yoh []

let _ =
  print_string "dataextracting.ml\n\n";
  print_string "Test  initialize_final_states_set \n\n";
  asserteq corres [("qf",0)] ""
ENDIF


IFDEF TEST THEN
let _ =
  print_string "Test  update_correspondence_table \n\n"

let final_corres =
  update_correspondence_table state_spec final_state_spec corres d_yoh

let _ = asserteq final_corres
  [("qd", 4); ("qc", 3); ("qb", 2); ("qa", 1); ("qf", 0)] ""
ENDIF


IFDEF TEST THEN
let _ =
  print_string "Test  transitions_set_building \n\n"

let parsedStates = List.map (fun x -> PState x) state_spec
let trans_1_yoh = PFsym("a", []), PState("qa")
and trans_2_yoh = PFsym ("f", [PState "qa";PState "qb"; PState "qc"]), PState "qa"
and trans_3_yoh = PState ("qa"), PState("qf")
and trans_4_yoh = PFsym ("b", []), PState("qb")
and trans_5_yoh = PState ("qb"), PState("qc")
and trans_6_yoh = PState ("qp"), PState("qc")
and trans_7_yoh = PFsym ("f", [PState "qa"; PState "qp"; PState "qc"]), PState "qa"

let trans_set = [trans_1_yoh; trans_2_yoh; trans_3_yoh;
                 trans_4_yoh; trans_5_yoh]
and bad_trans_set_1 = [trans_1_yoh; trans_2_yoh; trans_3_yoh;
                       trans_4_yoh; trans_5_yoh; trans_6_yoh]

and bad_trans_set_2 = [trans_1_yoh; trans_2_yoh; trans_3_yoh;
                       trans_4_yoh; trans_5_yoh; trans_7_yoh]
let test_errors transitions =
  try
    transitions_set_building transitions parsedStates final_corres d_yoh;
    "ok"
  with
    | Failure res -> res
let _ =
  transitions_set_building trans_set parsedStates final_corres d_yoh;
  asserteq (Transitions.to_string d_yoh) 
    "b -> q2\na -> q1\nf(q1,q2,q3) -> q1\n\nEpsilonR=\nq2 -> q3\nq1 -> q0\n" "";
  asserteq (test_errors trans_set) "ok" "";
  asserteq (test_errors bad_trans_set_2)
    "transitions_set_building: Some of states used are not declared." "";
  asserteq (test_errors bad_trans_set_1)
    "The state qp is not declared." "";
  asserteq (test_errors ((PVar "X", PState "qa")::trans_set))
    "There is a variable in the lhs of a transition." ""
ENDIF



IFDEF TEST THEN 
let _ = print_string "Test  automaton_construction \n\n"

let terme1_yoh = PFsym ("f", [PFsym ("h", [PFsym ("b", []);
                                           PFsym ("g", [PFsym ("a", [])]) ]);
                              PFsym ("b", []) ])
and terme2_yoh = PFsym ("f", [PFsym ("b", []); PFsym ("b", [])] )
and terme3_yoh = PFsym ("f", [PFsym ("g", [PFsym ("a", [])]);
                              PFsym ("g", [PFsym ("a", [])]) ])

let spec_Set = PSet ("Init", [terme1_yoh; terme2_yoh; terme3_yoh])

let (nom, aut, correspondence) = automaton_construction spec_Set

let _ =
  asserteq (Transitions.to_string aut.trans)
    "f(q2,q2) -> q6\nf(q0,q0) -> q5\nf(q3,q0) -> q4\nh(q0,q2) -> q3\n\
g(q1) -> q2\na -> q1\nb -> q0\n" "" ;
  asserteq (Stateset.to_string aut.fstates) "{q4,q5,q6}" "" ;
ENDIF

IFDEF TEST THEN
let rec do_for_each_trs f = function
  | [] -> "\n"
  | (x, trs)::tail -> x ^ "\n" ^ (f trs) ^ (do_for_each_trs f tail)


let rec do_for_each_autom f = function
  | [] -> "\n"
  | (x, trs, _)::tail -> x ^"\n" ^ (f trs) ^ (do_for_each_autom f tail)


let rec do_for_each_equation = function
  | [] -> "\n"
  | (id, eqs)::tail -> 
      id ^ "\n" ^ (Equation.to_string eqs) ^ (do_for_each_equation tail)

let rec string_list f s = function
  | [] -> "\n"
  | x::[] -> (f x) ^ "\n"
  | x::tail -> (f x) ^ s ^ (string_list f s tail)


let _ = 
  print_string "benchs/TU-ex0.txt\n"

let (trs, autom, patterns, dotpatterns, approx) =
  extraction "benchs/TU-ex0.txt"
in 
  print_string (do_for_each_trs Rewriting.to_string trs);
  print_string (do_for_each_autom Automaton.to_string autom);
  print_string ("Patterns\n" ^ (string_list Term.to_string "," patterns));
  print_string ("DotPatterns\n" ^ (string_list Term.to_string "," dotpatterns));
  print_string (do_for_each_equation approx);
ENDIF
