open Definitions
open Alphabet
open State
open Term
open Map
(* open Set2 *)


(* A substitution associates a variable to an element of the range *)

module Substitution =
struct

  module Subtemp = Map.Make(Variable)
  type substitution = term Subtemp.t

  type t = substitution

  let empty = Subtemp.empty
  let add = Subtemp.add
  let equal = Subtemp.equal
  let find = Subtemp.find
  let fold = Subtemp.fold
  let map = Subtemp.map
  let compare (s1: substitution) (s2: substitution) =
    (Subtemp.compare Variable.compare s1 s2)

  (* Cardinal of a substitution *)
  (* B: Mouais c'est pas tres mathematique comme def... *)
  let cardinal s =
    Subtemp.fold (fun _ _ i -> i + 1) s 0


  (* s1 and s2 are NO LONGER supposed to be two substitutions with disjoint domains *)
  let union s1 s2 =
    Subtemp.fold Subtemp.add
      (* fun x t subst -> if Subtemp.mem x subst then subst else
	 Subtemp.add x t subst *)
      s1 s2

  let to_string (s: substitution) =
    let str =
      Subtemp.fold
        (fun x t str ->
           ", " ^ (Variable.to_string x) ^ " -> " ^ (Term.to_string t) ^ str
        ) s ""
    in
      "{" ^ (if str = "" then "" else Str.string_after str 2) ^ "}"
        (*
          let rec aux l = function
          | [] -> ""
          | [(x, v)] -> (Variable.to_string x) ^ " -> " ^ (Term.to_string v)
          | (x, v)::rem -> (Variable.to_string x) ^ " -> "^ (Term.to_string v) ^ ", " ^ (aux rem)
          in
          let res = ref [] in
          let _ = Subtemp.iter (fun x v -> res:= (x,v)::!res) s in
          "{"^(aux !res)^"}"
        *)


  exception Inclusion

  (* covers : s1 covers by s2 if value for all keys is the same *)
  let covers (s1: substitution) (s2: substitution) : bool =
    try
      Subtemp.fold
        (fun key t b ->
	   if key = "_" then
             failwith "No _ in substitution at that level"
	   else if t = Subtemp.find key s2 then
             b
           else
	     raise Inclusion
        )
	s1 true
    with Not_found | Inclusion -> false

  (* Substitution inclusion *)

  (* fold with exception, in the case where an association is not included
     in s2, in order not to follow all s1... due to unfriendly interface
     of maps: no for_all for instance... *)
  (* Note that variable "_" plays here a specific role.
     if _ is associated to v1 on one side it may be associated (or not)
     on the other side *)

  let is_included (s1: substitution) (s2: substitution) : bool =
    try
      Subtemp.fold
        (fun key t b ->
	   if key = "_" then
             b
	   else if t = Subtemp.find key s2 then
             b
           else
	     raise Inclusion
        )
	s1 true
    with Not_found | Inclusion -> false

  (* Extended substitution equivalence such that {X -> q1, Z -> q3} is equivalent in
     {Y -> q2, Z-> q3}, i.e. substitution are always equivalent on variables they do not
     share.  *)

  (* Used for matching lhs and rhs of equations. Those equations only produce Trivial
     formulas in substitution, no check has to be performed on formulas *)
  let extended_equivalence (s1: substitution) (s2: substitution) : bool =
    try
      Subtemp.fold
        (fun key t b ->
	   if key = "_" then
             b
	   else
             try
	       if t = Subtemp.find key s2 then
                 b
               else
                 raise Inclusion
	     with Not_found -> b
        )
	s1 true
    with Inclusion -> false

end (* Substitution *)


type substitution = Substitution.t


(* A substitutionPhi contains a substitution and a formula *)
module SubstitutionPhi = struct

  type t = { value : substitution;
             phi: Formula.t
           }

  exception Inclusion


  let empty =
    { value = Substitution.empty; phi = Formula.mk_false }

  let add x t s =
    { s with
       value = Substitution.add x t s.value
    }

  let define sigma phi =
    { value = sigma; phi = phi }

  let addPhi p s =
    { s with
        phi = Formula.mk_or s.phi p
    }

  let addBoth x t f s =  (* AddPhi f (add x t s) *)
    { value = Substitution.add x t s.value;
      phi = Formula.mk_or s.phi f
    }

  let compare (s1: t) (s2: t) =
    Substitution.compare s1.value s2.value

  (* Cardinal of a substitution *)
  let cardinal s = Substitution.cardinal s.value

  (* s1 and s2 are NO LONGER supposed to be two substitutions with disjoint domains *)
  (* Union of maps and AND of formulas *)
  let union s1 s2 =
    { value = Substitution.union s1.value s2.value;
      phi = Formula.mk_and s1.phi s2.phi
    }

  let to_string (s: t) =
    (Substitution.to_string s.value) ^ "[" ^ (Formula.to_string s.phi) ^ "]"

  (* coversValue : s1 covers by s2 if value is the same *)
  let coversValue (s1: t) (s2: t) : bool =
    Substitution.covers s1.value s2.value


  (* Covered substitution: s1 covers by s2 if value is the same AND
     !!!formula of s2 implies s1!!!

     Indeed if s2=sigma||A
               s1=sigma||A\/B

     s1 covers s2 because it is more general (more application cases)
  *)

  let covers (s1: t) (s2: t) : bool =
    coversValue s1 s2 && (Formula.subsume s2.phi s1.phi)

  let equal (s1: t) (s2: t) : bool =
    covers s1 s2 && (covers s2 s1)

  (* Substitution inclusion *)

  (* fold with exception, in the case where an association is not included
     in s2, in order not to follow all s1... due to unfriendly interface
     of maps: no for_all for instance... *)
  (* Note that variable "_" plays here a specific role.
     if _ is associated to v1 on one side it may be associated (or not)
     on the other side *)

  (* Does not take formulas into account! *)
  let is_included (s1: t) (s2: t) : bool =
    Substitution.is_included s1.value s2.value


  (* Extended substitution equivalence such that {X -> q1, Z -> q3} is equivalent in
     {Y -> q2, Z-> q3}, i.e. substitution are always equivalent on variables they do not
     share.  *)

  (* Used for matching lhs and rhs of equations. Those equations only produce Trivial
     formulas in substitution, no check has to be performed on formulas *)

  let extended_equivalence (s1: t) (s2: t) : bool =
    Substitution.extended_equivalence s1.value s2.value
end

type substitutionPhi = SubstitutionPhi.t


(* A substitution set is a map associating a substitution (Subtemp.t) to a formula *)
module SubstitutionSet =
struct
  module Substsettemp = Map.Make(Substitution)

  open SubstitutionPhi

  (* open SubstitutionPhi *)
  (* exception Not_found= Substsettemp.Not_found *)
  type substitutionset = Formula.t Substsettemp.t

  type t = substitutionset
  let empty = Substsettemp.empty

  let to_list es =
    Substsettemp.fold
      (fun s phi l -> (SubstitutionPhi.define s phi)::l) es []


  let to_string es =
    "{" ^ (String.concat "," (List.map SubstitutionPhi.to_string (to_list es)) ) ^ "}"

  let list_to_string les=
    "[" ^ (String.concat ";" (List.map to_string les)) ^"]"


  (* Membership in set of substitution is replaced by coverage *)
  (* A substitution [s1] is in [es] if [s1.value] is in [es] and the formula associated
     in to [s1.value] in [es] is IMPLIED by [s1.phi] *)
  let mem s1 es =
    try
      let phi2 = Substsettemp.find s1.value es in
	Formula.subsume s1.phi phi2
    with
      | Not_found -> false


  let replace s phi es =
    Substsettemp.add s phi (Substsettemp.remove s es)

  (* Addition of substitution to a set of substs, checks if there is a similar subst
     in the set. In case, adds a disjonction to the formula *)
  let add (s: substitutionPhi) es =
    try
      let phi2 = Substsettemp.find s.value es in
	if Formula.subsume phi2 s.phi then
          es
	else
          replace s.value (Formula.mk_or phi2 s.phi) es
    with
      | Not_found -> Substsettemp.add s.value s.phi es


  let fold = Substsettemp.fold

  let addPhi_to_all (f: Formula.t) (es: substitutionset) : substitutionset =
    fold (fun s f2 es -> Substsettemp.add s (Formula.mk_and f2 f) es) es empty

  let from_list l = List.fold_right (fun s es -> add s es) l empty


  (* Substitution set difference
     in diff A B, we remove a substitution s associated to formula fB in B
     from substitutions of A if s occurs in A and is associated to fA such
     that fA => fB.
  *)
  let diff (esA: substitutionset) (esB: substitutionset) =
    fold
      (fun s fA es ->
         try
           let fB = Substsettemp.find s esB in
             if Formula.subsume fA fB then
               es
             else
               Substsettemp.add s fA es
         with
           | Not_found -> Substsettemp.add s fA es
      ) esA empty


  let is_empty =  Substsettemp.is_empty

  (* B: Assez cher : 3 * |es|
  *)
  let filter f es =
     from_list (List.filter f (to_list es))
       (* B: Autre proposition (ne type pas plus bas... )
         fold
         (fun sigma phi acc ->
         if f sigma then Substsettemp.add sigma phi es else es)
         es Substsettemp.empty
       *)



  (* Semantic inclusion of substitution set. A set es1 is included in es2 if
     all substitutions of es1 are present (syntactically) in es2 and with EQUIVALENT formulas.
     Only a criterion if Formula.subsume is a criterion.
  *)

  let included es1 es2 =
    Substsettemp.fold
      (fun s p b ->
	 try
           Formula.equival p (Substsettemp.find s es2) && b
	 with Not_found -> false
      ) es1 true


  (* Semantic equivalence of substitution sets, see included  *)
  let equal es1 es2 = included es1 es2 && (included es2 es1)

  let singleton s = add s empty

  let
elements s : substitutionPhi list = to_list s
    (*
       Substsettemp.fold (fun e phi l -> {value=e;phi=phi}::l) s []
    *)
  let union es1 es2=
    Substsettemp.fold
      (fun e phi es -> add {value=e;phi=phi} es) es1 es2

  (* Combination of substitionset list *)
  (* combine of [{{x -> q1},{x -> q2}};{{y->q3,z->q4}}] =
     {{x -> q1,y->q3,z->q4},{x -> q2,y->q3,z->q4}}
     if one set of susbt is empty, so is the combination.
     Combination of {({x->q1},Phi1), ({x->q2},Phi2)} with
  {({y -> q3},Phi3), ({y -> q4},Phi4)} results in
  {({x -> q1, y -> q3},Phi1 et Phi3), ({x->q1,y->q4},Phi1 et Phi4) etc. *)

  let rec combine (sl: substitutionset list): substitutionset=
    let rec one_subst_combine (s1: substitutionPhi) (s2: substitutionPhi list): substitutionset=
      match s2 with
	  [] -> Substsettemp.empty
	| s::rem ->
	    add (SubstitutionPhi.union s s1)(one_subst_combine s1 rem)
    in
    let rec substens_combine (s1: substitutionPhi list) (s2: substitutionPhi list)=
      match s1 with
	  [] -> Substsettemp.empty
	| e::rem ->
	    let combine_first= (one_subst_combine e s2) in
	      union combine_first (substens_combine rem s2)
    in
      match sl with
	  [] -> Substsettemp.empty
	| [s] -> s
	| s::rem ->
	    if Substsettemp.is_empty s then Substsettemp.empty
	    else
	      let res_rem= combine rem in
		if Substsettemp.is_empty res_rem
		then Substsettemp.empty
		else substens_combine (to_list s) (to_list res_rem)


(* Matching substitution checks if at least one substitution of es1 is included in
   one substitution of es2 (or one of es2 is included in one of es1)
   Note also that substitutions are not compared on variables not occurring in
   both es1 and es2. For instance, {{X -> q1}} and {{Y -> q2}} is matching. This
   is due to the fact that it is necessary for equations of the form s(X)=s(Y) where
   any substitution on X and Y is valid.
*)

(* Matching substitution w.r.t. extended_equivalence, no check on formulas since
   it is used for equations (trivial formulas on subst produced for equations) *)

let matching_substitutionset (es1: substitutionset) (es2: substitutionset): bool=
  if is_empty es1 then false
  else
    if is_empty es2 then false
    else
      List.exists
	(function elt -> List.exists
	   (function elt2 -> SubstitutionPhi.extended_equivalence elt elt2) (to_list es2)) (to_list es1)

(* Intersection of substitution sets where identity of substitutions is replaced by
   extended_equivalence. For extended equations -> no need to check formulas associated
   to substitution they are all trivial. *)

  let extended_intersection (es1: substitutionset) (es2: substitutionset): substitutionset=
    let rec aux sl1 sl2 sres=
      match sl1 with
	  [] -> sres
	| s1::rem -> (* Solutions are union of substitutions of es1 and es2 such that
			they are identitcal w.r.t. extended_equivalence *)
	    let newsres= List.fold_right
			   (function s2-> function es ->
			      if SubstitutionPhi.extended_equivalence s1 s2
			      then (add (SubstitutionPhi.union s1 s2) es)
			      else es)
			   sl2 sres
	    in
	      aux rem sl2 newsres
    in
      aux (to_list es1) (to_list es2) empty

end

type substitutionset= SubstitutionSet.t


module SubstitutionTable=
struct
  (* in implementation of substable set of substitutions are ordered by redex
     state then by rule state. But presentation (with to_string for instance) is
     just the opposite. The reason for ordering table by redex state firt is
     that merge applied on redex states merges substitutions for rule states. It
     is thus important to be abble to find all subtables for 2 distinct redex
     states in order to merge them *)
  (* to 2 states we associate a s

     qredex -> qlhs -> substitution set where each substitution is associated to
     a formula Phi. Formula Phi tells which are the q1->q2 necessary for the
     substitution to exists.  *)

  type substable= (state, (state, substitutionset) Hashtbl.t) Hashtbl.t
  type t = substable
  let new_substable () : substable = Hashtbl.create Common.max_state_symb
  exception Couple_not_found

  (* Finding a substitution s in a table t for a given couple of states cq.
     Throws Couple_not_found if cq is not in the table  *)

  let find ((qlhs,qred): state * state) (t: substable): substitutionset=
    try Hashtbl.find (Hashtbl.find t qred) qlhs
    with Not_found -> raise Couple_not_found

  (* Checks if a substitution is COVERED by a table *)
  let mem (cq: state * state) (s: substitutionPhi) (t: substable): bool=
    try (let es= (find cq t) in SubstitutionSet.mem s es)
    with Couple_not_found -> false


(* Adding of a substitution s in a table t for a given couple of states cq.
   Returns true if the substitution was not *COVERED* by the table *)

  let add (s: substitutionPhi) ((qlhs,qred): state * state) (t: substable): bool=
    try
      (let sub_by_qlhs= Hashtbl.find t qred in
	 try
	   (let ens_subst= Hashtbl.find sub_by_qlhs qlhs in
	      if SubstitutionSet.mem s ens_subst then false
	      else
		begin
		  Hashtbl.replace sub_by_qlhs qlhs (SubstitutionSet.add s ens_subst);
		  true
		end)
	 with Not_found ->
	   begin
	     Hashtbl.add sub_by_qlhs qlhs (SubstitutionSet.singleton s);
	     true
	   end)
    with Not_found ->
      let newsub_by_qlhs= Hashtbl.create Common.max_rule_node in
	begin
	  Hashtbl.add newsub_by_qlhs qlhs (SubstitutionSet.singleton s);
	  Hashtbl.add t qred newsub_by_qlhs;
	true
      end


  (* Adding a set of substitution es in a table t for a given couple of
     states cq. Returns the set of substitution that were not already *COVERED* by
     the table, and the empty set otherwise (in the case where they are
     all in the table. The first state is for lhs, the second one for redex *)

  let add_set (es: substitutionset) (cq: state * state) (t: substable): substitutionset=
    SubstitutionSet.filter (function s -> add s cq t) es


let to_string (t: substable): string=
  let all_sub1= Hashtbl2.elements t in
  let all_sub= List.map (function (q,symtable) -> (q, (Hashtbl2.elements symtable))) all_sub1 in
  let rec to_string_aux1 l q=
    match l with
	[] -> ""
      | (q2,es)::rem -> "("^(State.to_string q2)^","^(State.to_string q)^") -> "^
			(SubstitutionSet.to_string es)^"\n"^(to_string_aux1 rem q) in
  let rec to_string_aux l=
    match l with
	[] -> ""
      | (q,l)::rem -> (to_string_aux1 l q)^(to_string_aux rem)
  in
    to_string_aux all_sub


let list_to_string (l: ((state * state) * substitutionset) list): string=
  List.fold_right (function ((q1,q2),es) -> function s ->
		     ("("^(State.to_string q1)^","^(State.to_string q2)^") -> "^
			(SubstitutionSet.to_string es)^"\n"^s))
    l ""


  let copy (t: substable): substable=
    Hashtbl2.copy_level2 t

end

type substable= SubstitutionTable.t




(* ------------------------------TESTS ----------------------------------------*)

IFDEF TEST THEN print_string "Substitution.ml\n\n" END

IFDEF TEST THEN
open Debug
open Substitution
open SubstitutionPhi
open SubstitutionSet
open SubstitutionTable
END

IFDEF TEST THEN
(* sub1= {y->q3,z->q4}
   sub2= {y->q3}
   sub3= {y -> q3, z -> q5}
   sub4= {z -> q1}
   sub5= {_ -> q2}
   sub6= {_ -> q1, y -> q3, z-> q4}
   sub8= {_ -> q2, y-> q3, z-> q5}
   s=s4={}
   setsubempty={{}}
   stest= {{x -> q1}}
   set2b= {{x -> q2}}
   set2= {{x -> q1},{x -> q2}}
   set3= {{y->q3, z->q4}}
   set4= {{x -> q1,y->q3,z->q4},{x -> q2,y->q3,z->q4}}
*)

  let sub1= (SubstitutionPhi.add "y" (Term.State 3)
		 (SubstitutionPhi.add "z" (Term.State 4)
		    SubstitutionPhi.empty))
  let sub2= (SubstitutionPhi.add "y" (Term.State 3) SubstitutionPhi.empty)
  let sub3= (SubstitutionPhi.add "y" (Term.State 3)
		 (SubstitutionPhi.add "z" (Term.State 5)
		    SubstitutionPhi.empty))
  let sub4= SubstitutionPhi.add "z" (Term.State 1) SubstitutionPhi.empty
  let sub5= SubstitutionPhi.add "_" (Term.State 2) SubstitutionPhi.empty
  let sub6= SubstitutionPhi.add "_" (Term.State 1) sub1
  let sub8= SubstitutionPhi.add "_" (Term.State 2) sub3

  let s=SubstitutionPhi.empty
  let setsubstempty=
    SubstitutionSet.add (SubstitutionPhi.addPhi Formula.mk_true s) SubstitutionSet.empty
  let s4=SubstitutionPhi.empty
  let stest= (SubstitutionSet.add
		 (SubstitutionPhi.add "x" (Term.State 1) SubstitutionPhi.empty)
		 SubstitutionSet.empty)

	   (* set2b= {{x -> q1}[(1,0)]} *)
  let set2b= (SubstitutionSet.add
		(SubstitutionPhi.addPhi Formula.f1
		   (SubstitutionPhi.add "x" (Term.State 1) SubstitutionPhi.empty))
		SubstitutionSet.empty)

	   (* set2= {{x -> q1}[(1,0)\/(2,0)],{x -> q2}[(2,0)]} *)
  let set2= SubstitutionSet.add
	      (SubstitutionPhi.addPhi Formula.f4
		 (SubstitutionPhi.add "x" (Term.State 1) SubstitutionPhi.empty))
	      (SubstitutionSet.add
		 (SubstitutionPhi.addPhi Formula.f2
		    (SubstitutionPhi.add "x" (Term.State 2) SubstitutionPhi.empty))
		 SubstitutionSet.empty)

	   (* set2e= {{x -> q1}[(1,0)],{x -> q2}[(2,0)]} *)
  let set2e= SubstitutionSet.add
	      (SubstitutionPhi.addPhi Formula.f1
		 (SubstitutionPhi.add "x" (Term.State 1) SubstitutionPhi.empty))
	      (SubstitutionSet.add
		 (SubstitutionPhi.addPhi Formula.f2
		    (SubstitutionPhi.add "x" (Term.State 2) SubstitutionPhi.empty))
		 SubstitutionSet.empty)


	   (* set2f= {{x -> q1}[(1,0)/\(3,0)],{x -> q2}[(2,0)/\(3,0)]} *)
  let set2f= SubstitutionSet.add
	      (SubstitutionPhi.addPhi Formula.f12
		 (SubstitutionPhi.add "x" (Term.State 1) SubstitutionPhi.empty))
	      (SubstitutionSet.add
		 (SubstitutionPhi.addPhi Formula.f13
		    (SubstitutionPhi.add "x" (Term.State 2) SubstitutionPhi.empty))
		 SubstitutionSet.empty)


	   (* set2c= {{x -> q1}[(1,0)\/(2,0)]} *)
  let set2c= SubstitutionSet.add
	      (SubstitutionPhi.addPhi Formula.f4
		 (SubstitutionPhi.add "x" (Term.State 1) SubstitutionPhi.empty))
	       SubstitutionSet.empty

	      (* set3= {{y->q3,z->q4}[(1,0)]} *)
  let set3= SubstitutionSet.add
	      (SubstitutionPhi.addPhi Formula.f1
		 (SubstitutionPhi.add "y" (Term.State 3)
		    (SubstitutionPhi.add "z" (Term.State 4)
		       SubstitutionPhi.empty)))
	      SubstitutionSet.empty

  (* set4= {{x -> q1,y->q3,z->q4}[((1,0)/\(1,0))\/((1,0)/\(2,0)],
     {x -> q2,y->q3,z->q4}[(2,0)/\(1,0)]} *)

  let set4=
    SubstitutionSet.add
      (SubstitutionPhi.addPhi Formula.f1etorf2
	 (SubstitutionPhi.add "x" (Term.State 1)
	    (SubstitutionPhi.add "y" (Term.State 3)
	       (SubstitutionPhi.add "z" (Term.State 4)
		  SubstitutionPhi.empty))))
      (SubstitutionSet.add
	 (SubstitutionPhi.addPhi Formula.f1etf2
	    (SubstitutionPhi.add "x" (Term.State 2)
	       (SubstitutionPhi.add "y" (Term.State 3)
		  (SubstitutionPhi.add "z" (Term.State 4)
		     SubstitutionPhi.empty))))
	 SubstitutionSet.empty)
END

IFDEF TEST THEN
  let _= print_string ((SubstitutionPhi.to_string s)^"\n") in
  let s2= SubstitutionPhi.add "x" Term.tn s in
  let s3= SubstitutionPhi.add "y" Term.tm s2 in
  let s5= SubstitutionPhi.add "y" Term.tm s4 in
  let s6= SubstitutionPhi.add "x" Term.tn s5 in
  let set1= SubstitutionSet.add s6 SubstitutionSet.empty in
  let setemptysub= SubstitutionSet.add SubstitutionPhi.empty SubstitutionSet.empty in
  let t1= SubstitutionTable.new_substable () in
    begin
      print_string ((SubstitutionPhi.to_string s3)^"\n" );
      (* order of associations have no effect on subst values *)
      asserteq (compare s3 s6) 0 "order of assocications have no effect on subst values";
      asserteq (SubstitutionSet.add s3 set1) set1 "adding equivalent substitutions to a set has no effect";
      (* Combination of substitionset list *)
      (* combine of [{{x -> q1},{x -> q2}};{{y->q3,z->q4}}] =
	 {{x -> q1,y->q3,z->q4},{x -> q2,y->q3,z->q4}} *)

      print_string "Test SubstitutionSet.addPhi_to_all\n";
      print_string ((SubstitutionSet.to_string set2e)^"\n");
      print_string ((SubstitutionSet.to_string (SubstitutionSet.addPhi_to_all Formula.f3 set2e))^"\n");
      asserteq (SubstitutionSet.equal (SubstitutionSet.addPhi_to_all Formula.f3 set2e) set2f) true "Test 1 addPhi_to_all";
      print_string "Test SubstitutionSet.diff\n";
      print_string (SubstitutionSet.to_string (SubstitutionSet.diff set2b set2));
      asserteq (SubstitutionSet.diff set2b set2) SubstitutionSet.empty "Test1 diff";
      asserteq (SubstitutionSet.equal (SubstitutionSet.diff set2 set2b) set2) true "Test2 diff";

      print_string "Test SubstitutionSet.included\n";

      asserteq (SubstitutionSet.included set2 set2) true "Test1 included";
      asserteq (SubstitutionSet.included SubstitutionSet.empty set2) true "Test1b included";
      asserteq (SubstitutionSet.included set2b set2) false "Test2 included";
      asserteq (SubstitutionSet.included set2c set2) true "Test3 included";
      asserteq (SubstitutionSet.included set2 set2c) false "Test4 included";

      print_string "Test SubstitutionSet.equal\n";

      asserteq (SubstitutionSet.equal set2 set2) true "Test1 equal";
      asserteq (SubstitutionSet.equal set2b set2) false "Test2 equal";
      asserteq (SubstitutionSet.equal set2c set2) false "Test3 equal";


      print_string ("\n\nCombine set2="^(SubstitutionSet.to_string set2)^" with set3="^(SubstitutionSet.to_string set3)^" is "^(SubstitutionSet.to_string (SubstitutionSet.combine [set2;set3]))^" and should be equal to "^(SubstitutionSet.to_string set4)^"\n\n");
      asserteq (equal (SubstitutionSet.combine [set2;set3]) set4) true "combine of [{{x -> q1},{x -> q2}};{{y->q3,z->q4}}] = {{x -> q1,y->q3,z->q4},{x -> q2,y->q3,z->q4}}";
      print_string ("\n\nCombine set3="^(SubstitutionSet.to_string set3)^" with set2="^(SubstitutionSet.to_string set2)^" is "^(SubstitutionSet.to_string (SubstitutionSet.combine [set3;set2]))^" and should be equal to "^(SubstitutionSet.to_string set4)^"\n\n");
      asserteq (equal (SubstitutionSet.combine [set3;set2]) set4) true "combine of [{{y->q3,z->q4}};{{x -> q1},{x -> q2}}] = {{x -> q1,y->q3,z->q4},{x -> q2,y->q3,z->q4}}";
      asserteq (equal (SubstitutionSet.combine [set3;setsubstempty]) set3) true "combine of [{{x -> q1},{x -> q2}};{{}}] should be equal to [{{x -> q1},{x -> q2}}]";


(*
t1:
(1,2) -> {{}}
(2,3) -> {{}}
(4,5) -> {{x -> q1}[(1,0)}

*)
      let _= SubstitutionTable.add_set setemptysub (1,2) t1 in
      let _= SubstitutionTable.add_set setemptysub (2,3) t1 in
      let _= SubstitutionTable.add_set set2b (4,5) t1 in
	begin
	  print_string "Test SubstitutionTable.find\n";
	  asserteq (SubstitutionTable.find (1,2) t1) setemptysub "addition in substitution table";
	  asserteq (SubstitutionTable.find (2,3) t1) setemptysub "addition in substitution table";
	  asserteq (try (let _ = SubstitutionTable.find (5,5) t1 in false) with Couple_not_found -> true) true "(5,5) should not be already in the table";
	  print_string "Test SubstitutionTable.add_set\n";
	  asserteq (SubstitutionSet.equal (SubstitutionTable.add_set set2b (4,5) t1) SubstitutionSet.empty) true "Adding a set of substitution covered results in empty set";
	  asserteq (SubstitutionSet.equal (SubstitutionTable.add_set set2 (4,5) t1) set2) true "Adding a new set of substitution covering the first one should give the set of new substitutions";

	  print_string "Test SubstitutionTable.add\n";
  	  asserteq (SubstitutionTable.add (SubstitutionPhi.add "x" (Term.State 1) SubstitutionPhi.empty) (4,5) t1) false "Adding a subst already present returns false";
    	  asserteq
	    ((SubstitutionTable.add (SubstitutionPhi.add "x" (Term.State 1) SubstitutionPhi.empty) (6,6) t1) &&
	     ((SubstitutionTable.find (6,6) t1)=stest)) true "Adding a subst not present returns true and adds it in the table";
	  print_string "cardinal\n";
	  asserteq (cardinal s4) 0 "";
	  asserteq (cardinal sub1) 2 "";
	  print_string "is_included\n";
	  asserteq (is_included SubstitutionPhi.empty sub1) true "";
	  asserteq (is_included sub1 sub1) true "";
	  asserteq (is_included sub2 sub1) true "";
	  asserteq (is_included sub1 sub2) false "";
	  asserteq (is_included sub1 sub3) false "";
	  asserteq (is_included sub2 sub4) false "";
	  asserteq (is_included sub4 sub2) false "";
	  asserteq (is_included sub5 sub5) true "";
	  asserteq (is_included sub5 sub6) true "";
	  asserteq (is_included sub5 sub6) true "";
	  asserteq (is_included sub6 sub8) false "";
	  print_string "extended_equivalence\n";
	  asserteq (extended_equivalence sub5 sub6) true "";
	  asserteq (extended_equivalence sub1 sub2) true "on both directions";
	  asserteq (extended_equivalence sub2 sub1) true "on both directions";
	  asserteq (extended_equivalence sub1 sub3) false "";
	  asserteq (extended_equivalence sub4 sub2) true "No variables in common";
	  print_string "extended_intersection\n";
	  print_string (SubstitutionSet.to_string (extended_intersection set3 set4));
	  asserteq (SubstitutionSet.equal (extended_intersection set3 set4) set4) true ""
	end
    end
END
