(* We assume that all hashtbl have unique occurrences of keys
   Otherwise, table copies for transtable, satable etc. are not correct.
   older keys may be copied such that they became the more recent keys *)

open Definitions
open Pervasives2

let average_nb_symb_by_state = 3 (* basic size of hashtbl for a given symbol *)
let average_nb_trans_by_state = 3 (* basic size of hashtbl for a given symbol *)

(* Sets of ground transitions are represented by 3 tables:
   transtables, satables and ftables.

   * transtable: destination state -> symbol -> set of lists of states
   e.g.: f(q1,q2) -> q0 can be found as follows  q0 -> f -> {(q1,q2)}

   * satable (argument state access table):
   argument state -> symbol -> destination state -> set of lists of states
   e.g.: f(q1,q2) -> q0 has 2 entries in the satable:
   q1 -> f -> q0 -> {(q1,q2)}  and
   q2 -> f -> q0 -> {(q1,q2)}

   * ftables (symbol access table):
   symbol -> destination state set
   e.g.: f(q1,q2) -> q0 has the following entry in ftable:
   f -> {q0}
*)

(* A table is a set of lists of states (set of vectors) *)
(* We have one table by destination state (unique symbol), by symbol
   that gives a set of state list.*)
module Statetable = Set2.Make(Statelist)

type table = Statetable.t
(* a symboltable maps a symbol to its table *)
type symtable = (symbol, table) Hashtbl.t
(* a TRANSTABLE maps a final state (symbol) to its symbol table *)
type transtable = (state, symtable) Hashtbl.t
(* a SATABLE (state access table) maps a state qi, to a symbol f, to a dest
   state q to a statetable such that f(...qi...) -> q belongs to this table.
   Constant transitions do not occur in this table *)
type satable = (state, (symbol, (state , table) Hashtbl.t) Hashtbl.t) Hashtbl.t
(* a FTABLE gives for a symbol, all destination states where there are
   transitions for this symbol *)
type ftable = (symbol, Stateset.t) Hashtbl.t

(* Matrix representation for epsilon transitions *)
(* infered est voue a remplacer le type closure elt: *)
(* Ce type est utilise pour evaluer l'etat d'une transition:

   Away         : jamais ajoutee

   Alive x      : x indique le degree d'inference
                    x = 1 transition existante et ajoutee par completion
                    x > 1 transition existante et inferee lors de l'elagage dans ce cas
   Ghost        : supprimee mais on a besoin de savoir qu'elle a existe a posteriori

*)
type infered = Away | Alive of int | Ghost

type imatrix = infered array array
type fmatrix = Formula.t array array
type bmatrix = bool array array

let loop = true
  (* pour les fonctions epsilon[E/R]matrix_to_string :
     permet de considerer les self_loop ou non pour construire la chaine *)


let fmatrix_to_string etable : string =
  let res = ref "\n" and t = etable#fclosure
  and last = etable#last and line_feed = ref false in
    for i = 0 to last do
      for j = 0 to last do
	if (loop || i <> j) && (Formula.is_sat t.(j).(i)) then (
          line_feed := true;
          res := !res ^ (State.to_string i) ^ "----->" ^ (State.to_string j) ^
            " | " ^ (Formula.to_string t.(j).(i)) ^ "\n"
        )
      done;
      if !line_feed then (
        line_feed := false;
        res := !res ^ "\n"
      )
    done;
    !res

let imatrix_to_string ?(ghost = false) ?(timbuk = false) etable : string =
  let str = ref "" and t = etable#degrees
                   and maxind = etable#last in
    for i = 0 to maxind do
      for j = 0 to maxind do
	match t.(i).(j) with
	  | Alive x ->
              str := !str ^ (State.to_string j) ^
                (if timbuk then
                   " -> "
                 else
                   " --" ^ (string_of_int x) ^ "--> "
                ) ^ (State.to_string i) ^ "\n"
          | Ghost ->
              if ghost && not timbuk then
	        str := !str ^ (State.to_string j) ^ " --G--> " ^
                  (State.to_string i) ^ "\n"
	  | Away -> ()
      done
    done;
    !str

(* Dans la table fclosure : Maj de la formule t.(q2).(q1) de la
   transition q1 -> q2 avec phi, puis maintient de la cloture si
   necessaire: on suppose que [etable#fclosure] est deja close par
   transitivite avant la MaJ.  *)
(* Pour appeler la fonction:
   [update_transitive_closure etable q1 q2 phi
   - etable = delta.epsilonR ou delta.epsilonE
   - la transition q1 -f-> q2 dont on veut enrichir la formule f avec la formule phi
*)
let update_transitive_closure etable (q1: state)(q2: state)(phi: Formula.t) =
(*  let _= Gc.compact () in *)
  let t = etable#fclosure and maxind = etable#last in
    (* Format.printf "phi = %s\n" (Formula.to_string phi); *)
  (* let t_start = Sys.time () in *)
    if not (Formula.subsume phi t.(q2).(q1)) then
      (* If phi contains more information than t.(q2).(q1) => update *)
      begin
        t.(q2).(q1) <- Formula.mk_or phi t.(q2).(q1);
        (* add to predecessors of q2 all predecessors of q1 *)
        for i = 0 to maxind do
          t.(q2).(i) <- Formula.mk_or t.(q2).(i) (Formula.mk_and t.(q2).(q1) t.(q1).(i))
        done;
        (* add to predecessors of successors of q2 all predecessors of q1 *)
        for i = 0 to maxind do
          if Formula.is_sat t.(i).(q2) then (* i is a successor of q2 *)
	    for j = 0 to maxind do (* we add all pred of q2 to i *)
	      t.(i).(j) <- Formula.mk_or t.(i).(j) (Formula.mk_and t.(i).(q2) t.(q2).(j))
	    done
        done
      end(*;
    let t_elapsed = Sys.time () -. t_start in
      Common.approx_time +.= t_elapsed;
      print_string ("\nupdate_transitive_closure time: " ^
                      (string_of_float t_elapsed) ^ " seconds\n");
      flush stdout *)

(* Maintenant on definit une table de transition epsilon : *)
(* J'utilise le paradigme objet uniquement pour definir le soustypage tel que
   - epsilontableE < epsilontable
   - epsilontableR < epislontable

   L'avantage de l'approche par rapport aux types records, c'est que
   comme il n'y a pas de types "classe" en Caml, on ne peut pas avoir
   de sous-typage: donc les types record epsilonRtable et
   epsilonEtable sont incomparables: on se retrouve avec une
   duplication du code, plus une nomenclature aberente car 2 types
   record differents ne peuvent pas partager les noms des champs. Tous
   ces problemes sont resolus avec le sous-typage des objets.
*)

(* ---------------------------------------- Tests ---------------------------------------- *)
(* IFDEF TEST THEN *)
let print_fclosure t =
  let phi = t#fclosure in
    print_string "\n fclosure=\n";
    for q1 = 0 to t#last do
      for q2 = 0 to t#last do
        let phi = phi.(q1).(q2) in
	  if (loop || q1 <> q2) && (Formula.is_sat phi) &&
            not (Formula.tautology phi)
          then
            Format.printf "q%i -> q%i | Φ = %s\n@?" q2 q1 (Formula.to_string phi)
      done
    done;
    if !Common.interactive then ignore (Unix.system "read")
(* ENDIF *)


(* maxstate = Common.max_state_symb *)
class virtual epsilontbl max_state = object (me)

  val mutable succs : (state, Stateset.t) Hashtbl.t =
    Hashtbl.create max_state

  val mutable preds : (state, Stateset.t) Hashtbl.t =
    Hashtbl.create max_state

  val mutable fclosure : fmatrix = (* Formules pour (RE)^* ou E^* *)
    let t = Array.make_matrix max_state max_state Formula.mk_false in
      for i = 0 to max_state - 1 do
        t.(i).(i) <- Formula.mk_true (*  i ---->^* i est toujours vrai *)
      done;
      t

  (* Indique la taille de la fmatrix *)
  val mutable last = -1

  (* Pour acceder aux attributs *)
  method fclosure = fclosure
  method succs = succs
  method preds = preds
  method last = last

  method incr_last = last <- last + 1

  method succs_of q : Stateset.t =
    try Hashtbl.find succs q with
      | Not_found -> Stateset.empty

  method preds_of q : Stateset.t =
    try Hashtbl.find preds q with
      | Not_found -> Stateset.empty

  (* Ajout de la transition q1 -> q2 *)
  method private add q1 q2 : unit =
    Hashtbl.replace succs q1 (Stateset.add q2 (me#succs_of q1));
    Hashtbl.replace preds q2 (Stateset.add q1 (me#preds_of q2))

  (* Suppression de la transition q1 -> q2 *)
  method private remove q1 q2 : unit =
    Hashtbl.replace succs q1 (Stateset.remove q2 (me#succs_of q1));
    Hashtbl.replace preds q2 (Stateset.remove q1 (me#preds_of q2))

  (* Nettoyage de la cloture : retire de la cloture tous les chemins
     utilisant la transition q1 -> q2, et retire toutes les transitions
     qui doivent etre retirees a ce moment-la...
  *)
  method private virtual rm_atom : state -> state -> unit

  (* Redimensionne [fclosure] pour la nouvelle taille definie par [size] *)
  method resize size =
    (* On redimentionne fclosure a max_state en recopiant les elements
       de [fclosure] dans [fclosure'], un tableau de la nouvelle dimension,
       qui remplace [fclosure] puis on recopie les elements *)
    let fclosure' = Array.make_matrix size size Formula.mk_false
    and max = min (Array.length fclosure) size - 1 in
      for i = 0 to max do
        for j = 0 to max do
          fclosure'.(i).(j) <- fclosure.(i).(j)
        done
      done;
      (* states on the diagonal are cycles to merge *)
      for i = last + 1 to size - 1 do
        fclosure'.(i).(i) <- Formula.mk_true
      done;
      fclosure <- fclosure';
      last <- min last (size - 1)
        (* pour rester coherent, on s'assure que [last] est au max le
           plus grand index possible de [fclosure] *)

  (* Apres Oo.copy, [clone] et [me] partagent les memes structures de donnees...
     Pour realiser une copie un peu plus profonde, et comme les attributs sont prives,
     on est obliges de remplacer les attributs de [me] par une copie de ceux de [clone]
     qui ne peuvent etre modifies, a moins de definir Les Methodes [set] en plus des
     [get], ce qui devient un peu lourding...
  *)
  method copy =
    let clone = Oo.copy me in
      fclosure <- Array.map Array.copy clone#fclosure;
      succs <- Hashtbl.copy clone#succs;
      preds <- Hashtbl.copy clone#preds;
      clone
(*
  method equal_to (you: epsilontbl) =
    me#last = you#last &&
    try
      for i = 0 to me#last do
        for j = 0 to me#last do
          if me#fclosure.(i).(j) <> you#fclosure.(i).(j) then
            raise Break
        done
      done;
      true
    with Break -> false
*)
end



class epsilonRtbl max_state = object(me)

  inherit epsilontbl max_state as super

  (* val mutable fbody : Formula.t ref StateMap.t ref StateMap.t =
    StateMap.empty

  method private set_body_of q1 q2 phi =
    try
      let sm = StateMap.find q1 fbody in
        try
          let phi' = StateMap.find q2 !sm in
            phi' := Formula.mk_or !phi' phi
        with Not_found ->
          sm := StateMap.add q2 (ref phi) !sm
    with Not_found ->
      fbody <- StateMap.add q1 (ref (StateMap.singleton q2 (ref phi))) fbody
  *)
  method addR q1 q2 phi =
    super#add q1 q2;
    (* set_body_of q1 q2 phi; *)
    update_transitive_closure me q1 q2 phi (*Formula.mk_atom q1 q2 Rw*)

  method addE q1 q2 =
    update_transitive_closure me q1 q2 (Formula.mk_atom q1 q2 Eq);

  method private rm_atom q1 q2 =
    (* let visited = ref StatesSet.empty in
    let to_remove := [] in
    let rec rm_atom q1 q2 =
      let clean_fbody q1 q2 =
        StateMap.iter
          (fun q sm ->
             let all_false := true in
               StateMap.iter
                 (fun q' phi ->
                    phi := Formula.rm_atom !phi (q1, q2, Rw);
                    if not (Formula.is_sat !phi) then (
                      if not StatesSet.mem (q1, q2) !visited then
                        to_remove := (q, q') :: !to_remove
                    ) else
                      all_false := false
                 ) !sm
                 if !all_false then sm := StatesSet.empty
          ) fbodies
      in *)
        for i = 0 to last do
          for j = 0 to last do
	    fclosure.(i).(j) <- Formula.rm_atom fclosure.(i).(j) (q1, q2, Eq);
            (* on retire l'atome Eq(q1, q2) de phi *)
            if not (Formula.is_sat fclosure.(i).(j)) then (* = False *)
              super#remove q1 q2
                (* si la formule est unsat, on retire la transition
                   q1 -> q2 de epsilonR si elle existe reellement sinon
                   on a juste retirer un chemin issu de la cloture RE^*
                *)
          done
        done
(*    in
      rm_atom q1 q2 *)


  method removeE q1 q2 =
    me#rm_atom q1 q2

end

class epsilonEtbl max_state = object (me)

  inherit epsilontbl max_state as super

  val mutable degrees : imatrix =
    Array.make_matrix max_state max_state Away

  method degrees = degrees

  (* retire tous les chemins de fclosure qui passe la transition q1 -> q2:
     pour chaque chemin qi —Φ—>^* qj, on retire toute les conjonctions
     de Φ qui contiennent l'atome Eq(q1, q2) *)
  method private rm_atom q1 q2 =
    for i = 0 to last do
      for j = 0 to last do
	fclosure.(i).(j) <- Formula.rm_atom fclosure.(i).(j) (q1, q2, Eq);
        (* on retire l'atome (i.e. la cunjonction) Eq(q1, q2) de Φ *)
      done
    done

  method addE ?(degree = 0) q1 q2 =
    match degrees.(q2).(q1) with
      | Alive _ | Ghost -> ()
      | Away -> (* q1 -> q2 jamais consideree *)
          (* Format.printf "epsilonE#addE: q%i -> q%i\n@?" q1 q2; *)
          super#add q1 q2;
          update_transitive_closure me q1 q2 (Formula.mk_atom q1 q2 Eq);
          degrees.(q2).(q1) <- Alive degree
          (* ; print_fclosure me *)


  method removeE q1 q2 =
    super#remove q1 q2;
    me#rm_atom q1 q2;
    degrees.(q2).(q1) <-
      (match degrees.(q2).(q1) with
         | Alive x -> Ghost
         | Ghost | Away ->
             failwith "RemoveE: remove a non-existent transition"
      )
    (* ; Format.printf "epsilonE#RemoveE: q%i -> q%i\n@?" q1 q2;
       Format.printf "infered:\n%s\n@?" (imatrix_to_string ~ghost:false me);
       print_fclosure me *)



  method resize size =
    let degrees' = Array.make_matrix size size Away
    and last = min (Array.length degrees) size - 1 in
      for i = 0 to last do
        for j = 0 to last do
          degrees'.(i).(j) <- degrees.(i).(j)
        done
      done;
      degrees <- degrees';
      super#resize size (* to extend fclosure *)

  method copy =
    let clone = super#copy in
      degrees <- Array.map Array.copy clone#degrees;
      clone

end


(* table.(q2).(q1) is true if q1 -> q2.  This is encoded in this way
   in order to get with table.(q2) all the predecessors of q2... *)

(* A state is 'free' when it occurs in no transition/substitution
   solution... otherwise it is not free. *)



(* external trans where the first state is the destination state *)
(* B: demander a Thomas d'ou ca vient ???? *)
type external_symtrans = symbol * state list * state

type external_epsilon = state * state * Formula.t

type external_trans =
  | Trans of external_symtrans
  | Epsilon of external_epsilon

let not_epsi : external_trans -> bool =
  function Trans _ -> true | _ -> false

let ext2trans = function
  | Trans trans -> trans
  | _ -> failwith "Transitions:ext2trans:cannot be an epsilon"

let ext2epsi = function
  | Epsilon trans -> trans
  | _ -> failwith "Transitions:ext2epsi:cannot be an regular trans"

(* Partition list of external_trans into a list of external_symtrams
   and external_epsilon *)
let partition =
  let rec partition ressym resepsi = function
    | [] -> ressym, resepsi
    | Trans t::lt  -> partition (t::ressym) resepsi lt
    | Epsilon t::lt -> partition ressym (t::resepsi) lt
  in
    partition [] []

type t = {
  transitions : transtable;
  satable : satable;
  ftable : ftable;
  epsilonR : epsilonRtbl;
  epsilonE : epsilonEtbl;
  mutable next_state : state
    (* [next_state] Definit le domaine des transitions : un nouvel etat prend la
       valeur de next qui doit etre incrementee... il conviendra de verifier
       apres increment que la valeur n'est pas [Sys.max_array_length], la
       taille limite des tableaux en caml, les etats servant d'index dans
       les tableaux.
    *)
}


let make () =
  { transitions = Hashtbl.create average_nb_symb_by_state;
    satable  = Hashtbl.create Common.max_state_symb;
    ftable   = Hashtbl.create Common.max_symb;
    (* transitions epsilon: E ∪ R *)
    epsilonR = new epsilonRtbl Common.max_state_symb;
    epsilonE = new epsilonEtbl Common.max_state_symb;
    next_state = 0
  }

(* All Constructors for delta *)
(* Table creation *)
(* for constant symbols, state vector is empty, dispo and filled are empty *)

(* Transition table creation *)
(* A transition table maps deststate->symb->state vector *)
(* for constant symbols, state vector is empty *)
let new_symtable () : symtable =
  Hashtbl.create average_nb_symb_by_state


(* State labels are integers. In the current implementation, new states
   are always new labels. Even if states are destroyed (by merging for instance)
   their label is not re-concerned. The reason is that substitution are
   not updated systematically and the correct value for a state is the value
   obtained by recursive applications of the state_merge table *)

(* Obtain a new state label *)

let make_state (delta: t) : state =
  let state_num = Array2.length delta.epsilonR#fclosure in
    if delta.next_state >= state_num then (
      (* next_state does not fit in index array *)
      delta.epsilonR#resize (2 * state_num);
      delta.epsilonE#resize (2 * state_num)
    );
    let state = delta.next_state in
      delta.next_state <- delta.next_state + 1;
      delta.epsilonE#incr_last;
      delta.epsilonR#incr_last;
      state

(* Set the maximum value of state counter. Used for building a TA
   from a list of external_symtrans (no epsi) *)
let set_state (delta: t) (s: state): unit= delta.next_state <- s

(* Check if a state is valid *)
let valid_state (delta: t) (s: state) = delta.next_state > s


(* Get the set of defined states *)
let get_all_states (delta: t) : state list =
  List2.sequence 0 (delta.next_state - 1)


let copy (delta: t) : t =
  { transitions = Hashtbl2.copy_level2 delta.transitions;
    satable = Hashtbl2.copy_level3 delta.satable;
    ftable = Hashtbl2.copy delta.ftable;
    epsilonR = delta.epsilonR#copy;
    epsilonE = delta.epsilonE#copy;
    next_state = delta.next_state
  }



IFDEF TEST THEN
(* Here we suppose d1.used ⊇ d1.used *)
let degrees_equal (d1: t) (d2: t) : bool =
  let t1 = d1.epsilonE#degrees and t2 = d2.epsilonE#degrees in
    d1.epsilonR#last = d2.epsilonR#last &&
      let last = d1.next_state - 1 in
        try
          for i = 0 to last do
            for j = 0 to last do
              if t1.(i).(j) <> t2.(i).(j) then
                raise Break
            done
          done;
          true
        with Break -> false


let fclosure_equal (d1: t) (d2: t) : bool =
  let t1 = d1.epsilonR#fclosure and t2 = d2.epsilonR#fclosure in
    d1.epsilonR#last = d2.epsilonR#last &&
      let last = d1.epsilonR#last in
        try
          for i = 0 to last do
            for j = 0 to last do
              if t1.(i).(j) <> t2.(i).(j) then
                raise Break
            done
          done;
          true
        with Break -> false


let delta_equal (d1: t) (d2: t) : bool =
  let trans = Hashtbl2.equal_level2 Statetable.equal d1.transitions d2.transitions in
  let satable = Hashtbl2.equal_level3 Statetable.equal d1.satable d2.satable in
  let ftable = Hashtbl2.equal_level1 Stateset.equal d1.ftable d2.ftable in
  let epsiR = fclosure_equal d1 d2 in
  let epsiE = degrees_equal d1 d2 in
  let next_state = d1.next_state = d2.next_state in
    print_string "\nDisequality:\n";
    if not trans then print_string "trans\n";
    if not satable then print_string "satable\n";
    if not ftable then print_string "ftable\n";
    if not epsiR then print_string "epsiR\n";
    if not epsiE then print_string "epsiE\n";
    if not next_state then print_string "used\n";
    trans && satable && ftable && epsiR && epsiE && next_state
END



(* Is a table representing a transition whose left-hand side is constant? *)
let constant_table (t: table) : bool = Statetable.mem [] t


let table_to_symlist (s: symbol) (dest: state) (t: table) : external_symtrans list =
  if constant_table t then
    [ (s, [], dest) ]
  else
    List.map (fun sl -> s, sl, dest) (Statetable.elements t)



(* Epsilon transition added by completion, we only print epsilon transitions
   added and not the transitive closure... We display only states occuring in efilled.
   If flag timbuk is true then formulas are not printed... so as to be readable by timbuk
   as an input *)
let epsilonR_to_string ?(timbuk=false) (d: t)  : string =
  let phi = d.epsilonR#fclosure in
    Hashtbl2.fold
      (fun s1 succs acc ->
         Stateset.fold
           (fun s2 acc ->
              State.to_string s1 ^ " -> " ^ (State.to_string s2) ^
              (if timbuk then "" else
                 (if (Formula.tautology phi.(s2).(s1)) then
                    ""
                  else
                    "||" ^ (Formula.to_string phi.(s2).(s1))
                 )) ^ "\n" ^ acc
           ) succs acc
      ) d.epsilonR#succs ""


let external_to_string : external_trans -> string = function
  | Trans (sym, sl, dest) ->
      (Symbol.to_string sym) ^ "(" ^ (Statelist.to_string sl) ^ ") -> " ^ (State.to_string dest)
  | Epsilon (q1, q2, f) ->
      (State.to_string q1) ^ " -> " ^ (State.to_string q2) ^ " || " ^ (Formula.to_string f)

let symext_to_string ((symb, l, dest): external_symtrans) : string =
  (Symbol.to_string symb) ^
    (if l = [] then "" else "(" ^ (Statelist.to_string l) ^ ")" ) ^
    " -> " ^ (State.to_string dest)

let symlist_to_string (l: external_symtrans list) : string =
  String.concat "\n" (List.map symext_to_string l) ^ "\n"

let table_to_string (s: symbol) (dest: state) (t: table) : string =
  if constant_table t then
    (Symbol.to_string s) ^ " -> " ^ (State.to_string dest) ^ "\n"
  else
    let str_start = Symbol.to_string s ^ "("
    and str_end = ") -> " ^ (State.to_string dest) ^ "\n" in
      List.fold_left
        (fun str sl -> str ^ str_start ^ (Statelist.to_string sl) ^ str_end)
        "" (Statetable.elements t)

(* B: Est-ce vraiment utile de transformer la table de hashage avant
   Il y a une fonction [fold] sur les tables de hash.
*)
let trans_to_string (t: transtable) : string =
  List.fold_left
    (fun str (q, symtable) ->
       List.fold_left
         (fun str (f, table) -> str ^ (table_to_string f q table))
         str
         symtable
    )
    ""
    (List.map
       (fun (q, symtable) -> q, Hashtbl2.elements symtable)
       (Hashtbl2.elements t)
    )


(* trans_to_list only produce regular transitions and no epsilon transition *)
let trans_to_list (t: transtable) : external_trans list =
  let table_to_list (s: symbol) (dest: state) (t: table) : external_trans list =
  if constant_table t then
    [Trans (s, [], dest)]
  else
    List.map (fun sl -> Trans (s, sl, dest)) (Statetable.elements t)
  in
  let all_trans =
    List.map
      (fun (q, symtable) -> q, Hashtbl2.elements symtable)
      (Hashtbl2.elements t)
  in
  let to_list_symboltable q =
    List.fold_left
      (fun l (f, table) -> List.append l (table_to_list f q table))
      []
  in
    List.fold_left
      (fun l (q, symtable) ->
         List.append l (to_list_symboltable q symtable)
      ) [] all_trans

let epsilonR_to_list (d: t) : external_trans list =
  let phi = d.epsilonR#fclosure in
    Hashtbl2.fold
      (fun s1 succs acc ->
         Stateset.fold
           (fun s2 acc ->
              if phi.(s2).(s1) <> Formula.mk_false then
              Epsilon (s1, s2, phi.(s2).(s1)) :: acc
              else acc
           ) succs acc
      ) d.epsilonR#succs []

let epsilonE_to_list (d: t) : external_trans list =
  let l = ref [] and degrees = d.epsilonE#degrees in
  let last = d.epsilonE#last in
    for i = 0 to last do
      for j = 0 to last do
        match degrees.(i).(j) with
          | Alive x ->
              l := Epsilon (j, i, Formula.mk_atom j i Eq) :: !l
          | _ -> ()
      done
    done;
    !l

let to_list (d: t) : external_trans list =
  (trans_to_list d.transitions) @ (epsilonR_to_list d) @ (epsilonE_to_list d)

let to_list_noE (d: t) : external_trans list =
  (trans_to_list d.transitions) @ (epsilonR_to_list d)



(* trans_to_transitionsBySymbols produces a hashtable
   symbol -> state -> state list
   For instance, the transitions f(q1,q2) -> q3 and
   f(q2,q5) -> q3 are stored as follows: f -> q3 -> [[q1; q2]; [q2; q5]] *)
type transitionsBySymbols = (symbol, (Statelist.t * state) list) Hashtbl.t
let trans_to_transitionsBySymbols (t1: transtable) : transitionsBySymbols =
  let ht = Hashtbl.create Common.max_symb in
  let update_table s f ls =
    if Hashtbl.mem ht f then
      let trans_list = Hashtbl.find ht f in
      let new_trans_list = (ls, s)::trans_list in
	Hashtbl.replace ht f new_trans_list
    else
      Hashtbl.add ht f [(ls, s)]
  in
    List.iter
      (function
         | Trans (f, sl, s) -> update_table s f sl
         | _ -> failwith "This case should never appear"
      )
      (trans_to_list t1);
    ht


let list_to_string (l: external_trans list) =
  String.concat "\n" (List.map external_to_string l)

IFDEF TEST THEN
let satable_to_string s =
  Hashtbl2.to_string_level3  "" s

let ftable_to_string s =
  Hashtbl2.to_string_level1 "" s

let print_delta (d: t) : unit =
  Format.printf "delta:\nnext_state: %i\n" d.next_state;
  Format.printf "transitions:\n%s\n" (trans_to_string d.transitions);
  Format.printf "epsilonR#last %i\n" d.epsilonR#last;
  Format.printf "epsilonE#last %i\n" d.epsilonE#last;
  Format.printf "epsilonR#fclosure: \n%s\n" (fmatrix_to_string d.epsilonR);
  Format.printf "epsilonE#fclosure: \n%s\n" (fmatrix_to_string d.epsilonE);
  Format.printf "epsilonE#degrees: \n%s\n" (imatrix_to_string d.epsilonE);
  Format.printf "\n\nSATABLE\n\n%s" (satable_to_string d.satable);
  Format.printf "\n\nFTABLE\n\n%s\n\n" (satable_to_string d.satable)
END

let to_string (delta: t) : string =
  let er = epsilonR_to_string delta in
  let ee = imatrix_to_string delta.epsilonE in
    (trans_to_string delta.transitions) ^
      (if er = "" then "" else "\nEpsilonR=\n" ^ er) ^
      if ee = "" then "" else "\nEpsilonE=\n" ^ ee


let to_timbuk (delta: t) : string =
  let er = epsilonR_to_string delta ~timbuk:true in
  let ee = imatrix_to_string delta.epsilonE ~timbuk:true in
    (trans_to_string delta.transitions) ^ er ^ ee



let symext_to_string ((symb, l, dest): external_symtrans) : string =
  (Symbol.to_string symb) ^
    (if l = [] then "" else "(" ^ (Statelist.to_string l) ^ ")") ^
    " -> " ^ (State.to_string dest)

let epsiext_to_string ((q1, q2, f): external_epsilon) : string =
  (State.to_string q1) ^ " -> " ^ (State.to_string q2) ^
    " || " ^ (Formula.to_string f)

let symextlist_to_string (l: external_symtrans list) : string =
  String.concat "\n" (List.map symext_to_string l) ^ "\n"

let epsiextlist_to_string (l: external_epsilon list) : string=
  String.concat "\n" (List.map epsiext_to_string l) ^ "\n"

let epsihash_to_string (t: (state,Stateset.t) Hashtbl.t) : string =
  (String.concat "\n"
     (List.map
        (fun (s, es) ->
	   "(" ^ (State.to_string s) ^ "," ^(Stateset.to_string es) ^ ")")
        (Hashtbl2.elements t)
     )
  ) ^ "\n"

(* s1 is the parameter state for which we want to add a table *)
let add_state_satable (sym: symbol) (dest: state) (sl:state list) (sa: satable) (s1:state) : unit =
  try
    let forstate = Hashtbl.find sa s1 in
      try
	let forsym = Hashtbl.find forstate sym in
	  try
	    let fordest = Hashtbl.find forsym dest in
	      Hashtbl.replace forsym dest (Statetable.add sl fordest)
	  with
            | Not_found -> (* destination does not exist already *)
	        Hashtbl.add forsym dest (Statetable.singleton sl)
      with
        | Not_found -> (* symbol -> dest does not exist already *)
	    let forsym = Hashtbl.create average_nb_symb_by_state in
	      Hashtbl.add forsym dest (Statetable.singleton sl);
	      Hashtbl.add forstate sym forsym
  with
    | Not_found -> (* parameter -> symbol -> dest does not exist already *)
        let forsym = Hashtbl.create average_nb_symb_by_state in
        let forstate = Hashtbl.create average_nb_symb_by_state in
  	  Hashtbl.add forsym dest (Statetable.singleton sl);
	  Hashtbl.add forstate sym forsym;
	  Hashtbl.add sa s1 forstate


(* Remove a state vector in a satable *)
let remove_state_satable (sym: symbol) (dest: state) (sl:state list) (sa: satable) (s1:state) : unit =
  try
    let forstate = Hashtbl.find sa s1 in
    let forsym = Hashtbl.find forstate sym in
    let fordest = Hashtbl.find forsym dest in
    let new_fordest = Statetable.remove sl fordest in
      if Statetable.is_empty new_fordest then (
	(* The set of transitions is empty for dest *)
	Hashtbl.remove forsym dest;
	if Hashtbl.length forsym = 0 then
          begin
	      (* The set of dest -> transitions is empty for sym *)
	    Hashtbl.remove forstate sym;
	    if Hashtbl.length forstate = 0 then
	      (* The set of sym -> dest -> transitions is empty for sym *)
	      Hashtbl.remove sa s1
          end
      ) else
        Hashtbl.replace forsym dest new_fordest
  with
    | Not_found -> ()


(* Add a transition in a satable w.r.t. to all its state parameters. *)
(* Constant transitions are not added in this table! *)
let add_satable ((sym, sl, dest): external_symtrans) (sa: satable) : unit =
  List.iter (fun x -> add_state_satable sym dest sl sa x) sl


(* Remove a transition in a satable w.r.t. to all its state parameters. *)
(* Constant transitions are not part of this table! *)
let remove_satable ((sym, sl, dest): external_symtrans) (sa: satable) : unit =
  List.iter (fun x -> remove_state_satable sym dest sl sa x) sl


(* Direct addition of a transition to a ftable *)
let add_ftable ((sym,sl,dest): external_symtrans) (fa: ftable) : unit =
  try
    let set = Hashtbl.find fa sym in
      if not (Stateset.mem dest set) then
	Hashtbl.replace fa sym (Stateset.add dest set)
  with
    | Not_found ->
        Hashtbl.add fa sym (Stateset.singleton dest)


(* low level adding of a transition to a transtable, we do not check if
   transition is already in the table. For constant transitions, a check is
   made *)

(* Only update the transtable, no symtable and satable update *)
let add_fast ((symb, sl, dest): external_symtrans) (t: transtable) : unit =
  try
    let symtable = Hashtbl.find t dest in
      try
        let table = Hashtbl.find symtable symb in
	  if sl <> [] then
	    Hashtbl.replace symtable symb (Statetable.add sl table)
              (* dest state exists but not for that symbol yet *)
       (* else if sl = [] then
            () --> add a constant transition, already present
       *)
      with
        | Not_found ->
	    Hashtbl.add symtable symb (Statetable.add sl Statetable.empty)
              (* dest state and symbol does not exist *)
  with
    | Not_found ->
        let new_table = Statetable.add sl Statetable.empty in
        let new_sym = new_symtable () in
	  Hashtbl.add new_sym symb new_table;
	  Hashtbl.add t dest new_sym


(* add with copy checking, returns true if transition is not present
   in the transtable, and false if the transition is present. *)
(* Only update the transtable, no symtable and satable update *)

let add_fast_trans ((symb,sl,dest): external_symtrans) (t: transtable) : bool =
  try
    let symtable = Hashtbl.find t dest in
    let table = Hashtbl.find symtable symb in
      if sl = [] || (Statetable.mem sl table) then
        false (* The transition is already in the table *)
      else (
	Hashtbl.replace symtable symb (Statetable.add sl table);
	true
      )
  with
    | Not_found ->
        add_fast (symb, sl, dest) t;
        true


(* general adding of a f(q1..qn)->q transition Returns false if the transition
   is already present in the table.
   The corresponding states are supposed to be already present in delta.epsilon *)

let add_regular_trans (t: external_symtrans) (delta: t) : bool =
  if add_fast_trans t delta.transitions then (
    add_satable t delta.satable;
    add_ftable t delta.ftable;
    true
  ) else
    false (* Transition is already present in the table *)


IFDEF TEST THEN
  Common.merge_loop := false;
ENDIF


(* infer_transitions :  pour la transition q1 -> q2 :
   Ici il n'y a pas besoin de mettre a jour degrees la table
   des degrees car dans closure chemin i --->^* j
   dans lequel on suprime Eq(q1, q2) n'est pas elementaire -- c'est
   n'est pas une transition mais un chemin -- et donc ne possede pas de degre!
   C'est le meme argument pour orderedTransitions et dsuccE et dpredE mais il faut
   quand meme supprimer la transition q1 -> q2, ainsi que mettre a jour les epsilonE.succs
   et epsilonE.preds mais ces cas sont traites dans la fonction
   remove_an_approximation_transition qui appelle new_infered_transitions
*)

(* q1 ---> q2 is supposed to be alive in degrees *)
let infer_transitions (delta: t) (q1: int) (q2: int) : unit =
  let degrees = delta.epsilonE#degrees in
  let addE (q1: state) (q2: state) (degree: int) =
    if q1 <> q2 then
      match degrees.(q2).(q1) with
        | Alive _ | Ghost ->
            Verbose.println 5 (fun _-> "infer_transitions: "^(State.to_string q1)^" -> "^(State.to_string q2)^" rejected\n");
        | Away ->
            Verbose.println 5 (fun _-> "infer_transitions: "^(State.to_string q1)^" -> "^(State.to_string q2)^" accepted\n");
	    delta.epsilonE#addE ~degree:degree q1 q2;
            delta.epsilonR#addE q1 q2
  in
  let deg = match degrees.(q2).(q1) with
    | Alive x -> x (* normalement x ≥ 0 *)
    | Ghost | Away -> assert false
  in
    (* Un predecesseur de q1 devient un predecesseur de q2 *)
    for q = 0 to delta.epsilonE#last do
      (match degrees.(q1).(q) with
	 | Alive deg' -> addE q q2 (max deg deg' + 1)
	 | Ghost -> addE q q2 (deg + 1)
         | Away -> ()
      );
      (* Un sucesseur de q2 devient un sucesseur de q1 *)
      match degrees.(q).(q2) with
	| Alive deg' -> addE q1 q (max deg deg' + 1)
        | Ghost -> addE q1 q (deg + 1)
        | Away -> ()
    done



(* a state is a predecessor of another if the formula associated to
   the edge is not the formula false *)
let is_succRE (delta: t)(q1: state) (q2: state) : bool =
  Formula.is_sat delta.epsilonR#fclosure.(q2).(q1)

let is_succRE_formula (delta:t)(q1:state) (q2: state): Formula.t=
  delta.epsilonR#fclosure.(q2).(q1)


let is_succE (delta: t)(q1: state) (q2: state) : bool =
  Formula.is_sat delta.epsilonE#fclosure.(q2).(q1)

(* get all pred of a state s by R (and E!). It never contains s itself
   (even if reachable) *)
let get_predRE (delta: t) (s: state) : (state * Formula.t) list =
  let t = delta.epsilonR#fclosure in
  let res = ref [] in
    try
      for i = 0 to delta.epsilonR#last do
	if i <> s && (Formula.is_sat t.(s).(i)) then
	  res := (i, t.(s).(i)) :: !res
      done;
      !res
    with Invalid_argument _ ->
      failwith "Transitions:get_pred: a state maybe not occuring in epsilon table!"


let get_succRE (delta: t)(s: state) : (state * Formula.t) list =
  let phi = delta.epsilonR#fclosure in
  let res = ref [] in
    try
      for i = 0 to delta.epsilonR#last do
	if i <> s && (Formula.is_sat phi.(i).(s)) then
	  res := (i, phi.(i).(s)) :: !res;
      done;
      !res
    with Not_found ->
          failwith "Transitions:get_succ: a state maybe not occuring in epsilon table!"

(* get all succ of a state s by E, only. It never contains s itself *)
let get_succE (delta: t) (q: state) : state list =
  let phi = delta.epsilonE#fclosure in
  let res = ref [] in
    for q' = 0 to delta.epsilonE#last do
      if q' <> q && (Formula.is_sat phi.(q').(q)) then
	res := q' :: !res
    done;
    !res



(* get all pred of a state s by E, only. It never contains s itself
   (even if reachable) *)
let get_predE (delta: t) (q: state) : state list =
  let phi = delta.epsilonE#fclosure in
  let res = ref [] in
    for q' = 0 to delta.epsilonE#last do
      if q' <> q  && (Formula.is_sat phi.(q).(q')) then
        res := q' :: !res
    done;
    !res



(* Retourne les composantes fortement connexes a partir de [closure],
   la fermeture transitive d'une relation donnee sous forme d'une
   liste de composantes non elementaires, i.e. au moins 2 etats: une
   composante est une liste d'etats.
*)
let get_strongly (rel: state -> state -> bool) (delta: t) : state list list =
  let res = ref [] in
    (* la classe d'équivalence (q_n, [q₀, ..., q_n]) ∈ !res avec
       q_n est un représentant arbitraire de la classe, et
       ∀ q q' on a q <--->^*_E q' *)
    for q = 0 to delta.epsilonE#last do
      try
        List.iter
          (fun (q', cycle) ->
             if rel q q' then (
               Verbose.println 5 (fun _ -> State.to_string q ^ "=" ^ (State.to_string q'));
               cycle := q :: !cycle;
               raise Break
             )
          ) !res;
        res := (q, ref [q]) :: !res
      with Break -> () (* on a ajoute q à la bonne classe *)
    done;
    (* on ne garde que les classes qui forment des cycles i.e. plus de 2 elts *)
    List.fold_left (fun acc (_, cycle) ->
                      match !cycle with
                        | _::_::_ -> !cycle::acc
                        | _ -> acc
                   ) [] !res


(* Get all pred of a set of states (pred by R and E) contains s itself! *)
let rec get_all_pred (set: Stateset.t) (delta: t) : Stateset.t =
   Stateset.fold
     (fun q acc ->
        Stateset.add_list (List.map fst (get_predRE delta q)) acc
     ) set set


(* Get all pred of a set of states (pred by R and E), with
   corresponding formula *)
(* Contains the set s itself with formula true, we do not take loops
   on final states into account *)

let rec get_all_predPhi (set: Stateset.t) (delta: t) : (state * (state * Formula.t)) list =
  Stateset.fold
    (fun s l ->
       List.fold_left
         (fun acc (q, phi) -> (q, (s, phi))::acc)
         l
         (get_predRE delta s)
    ) set
    (* We add (s,Formula.mk_true) for all state in ss *)
    (Stateset.fold (fun q l -> (q, (q, Formula.mk_true))::l) set [])


(* get the table for a given destination state, symbol
   (without modulo epsilon) *)
let simple_get_table (dest: state) (symb: symbol) (delta: t) : table =
  Hashtbl.find (Hashtbl.find delta.transitions dest) symb


(* get all transitions for a given destination state, symbol (without
   modulo epsilon) *)
let simple_get_trans (dest: state) (symb: symbol) (delta: t) : external_symtrans list =
  try
    let table = simple_get_table dest symb delta in
      table_to_symlist symb dest table
  with
    | Not_found -> []

(* get all transitions for a given symbol (without modulo epsilon) *)
let simple_get_trans_f (symb: symbol) (delta: t): external_symtrans list =
  try
    let stateset = Hashtbl.find delta.ftable symb in
      List.flatten
        (List.map
           (fun dest -> simple_get_trans dest symb delta)
	   (Stateset.elements stateset)
        )
  with
    | Not_found -> []

(* get all transitions for a given destination state, symbol, **modulo RE** *)
let get_trans (dest: state) (symb: symbol) (delta: t) : external_symtrans list =
  let phi = delta.epsilonR#fclosure in
  let acc = ref [] in
    for pred = 0 to delta.epsilonR#last do
      if Formula.is_sat phi.(dest).(pred) then (* we have pred ---->^*RE dest *)
        try
          let table = simple_get_table pred symb delta in
            if constant_table table then
              acc := (symb, [], pred)::!acc
            else
              acc := Statetable.fold (fun sl acc -> (symb, sl, pred)::acc) table !acc
        with Not_found -> ()
    done;
    !acc

(* Get all transitions by parameter state: i.e. get all f(...q2...) ->q  given q2 *)
let get_trans_parameter (par: state) (delta: t) init : external_symtrans list =
  try
    Hashtbl2.fold
      (fun symb t acc ->
         Hashtbl2.fold
           (fun dest vect acc ->
              Statetable.fold
                (fun sl acc -> (symb, sl, dest)::acc)
                vect acc
           ) t acc
      ) (Hashtbl.find delta.satable par) init
  with
    | Not_found -> [] (* states corresponding to constants are not in satable *)

let get_trans_parameter_list (pars: state list) (delta: t) : external_symtrans list =
  List.fold_left
    (fun acc par -> get_trans_parameter par delta acc)
    [] pars


(* get all transitions by top symbol and parameter state (no epsilon resolving) *)
(* symb is supposed not to be a constant symbol *)

let get_trans_sym_parameter (symb: symbol) (par: state) (delta: t) : external_symtrans list =
  try
    let all_trans =
      Hashtbl2.find (Hashtbl2.find delta.satable par) symb
    in
      Hashtbl2.fold
        (fun dest ts acc ->
           Statetable.fold
             (fun sl acc -> (symb, sl, dest)::acc)
             ts acc
        ) all_trans []
  with
    | Not_found -> []



(* get all transitions by top symbol for constant symbols (no epsilon resolving) *)
let get_trans_sym_constant (symb: symbol) (delta: t): external_symtrans list=
  try
    Stateset.fold
      (fun dest acc -> simple_get_trans dest symb delta @ acc)
      (Hashtbl.find delta.ftable symb) []
  with Not_found -> []


(* Get all symbolic transitions whose destination is in dest *)
let simple_get_all_trans (dest: state list) (delta: t) : external_symtrans list =
  let rec simple_get_all_trans acc = function
    | [] -> acc
    | st::rem ->
        simple_get_all_trans
	  (try
	     Hashtbl2.fold
               (fun sym t acc -> table_to_symlist sym st t @ acc)
               (Hashtbl.find delta.transitions st)
               acc
	   with Not_found -> acc
             (* If state in dest is not present in delta we ignore it *)
          ) rem
  in
    simple_get_all_trans [] dest




(* Cette fonction permet de supprimer une transition approximation
   tout en inférant toutes les transitions.  exemple, si nous avons
   les transitions qb -> q1, qa -> q1, q1 -> q2, q2 -> qc et q2 -> qd.
   Quand nous supprimons la transitions q1 -> q2, on infère les
   transitions q1 -> qc, q1 -> qd, qa -> q2 et qb -> q2.  Les
   transitions sont automatiquement ajoutées dans la matrice ainsi que
   dans la liste des transitions approximations triée.
*)
let removeE (delta: t)(q1: state)(q2: state) : unit =
  infer_transitions delta q1 q2;
  delta.epsilonE#removeE q1 q2;
  delta.epsilonR#removeE q1 q2


let addE (delta: t)(q1: state)(q2: state) : unit =
  match delta.epsilonE#degrees.(q2).(q1) with
    | Away when q1 <> q2 ->
        delta.epsilonE#addE q1 q2;
        delta.epsilonR#addE q1 q2
    | Away | Alive _ | Ghost ->
        Verbose.println 2 (fun _ -> "Transitions.addE: "^(Epsilon.to_string (q1, q2))^" rejected\n")

let addE' (delta: t)(q1: state)(q2: state) : bool =
  match delta.epsilonE#degrees.(q2).(q1) with
    | Away when q1 <> q2 ->
        delta.epsilonE#addE q1 q2;
        delta.epsilonR#addE q1 q2;
        true
    | Away | Alive _ | Ghost ->
        Verbose.println 2 (fun _ -> "Transitions.addE': "^(Epsilon.to_string (q1, q2))^" rejected\n");
        false



let add (delta: t) ((sym, sl, dest): external_symtrans) : unit =
  let _ =
    (* TESTS: States of t have to be already defined in d and formula should not be false *)
    IFDEF TEST THEN
      Debug.asserteq
      ((List.for_all (valid_state delta) sl) && (valid_state delta dest))
      true
      "States of trans should belong to delta!"
    ELSE () END
  in
    ignore (add_regular_trans (sym, sl, dest) delta)

let add' (delta: t) ((sym, sl, dest): external_symtrans) : bool =
  let _ =
    (* TESTS: States of t have to be already defined in d and formula should not be false *)
    IFDEF TEST THEN
      Debug.asserteq
      ((List.for_all (valid_state delta) sl) && (valid_state delta dest))
      true
      "States of trans should belong to delta!"
    ELSE () END
  in
    add_regular_trans (sym, sl, dest) delta


let addR (delta: t)((q1, q2, f): external_epsilon) : unit =
  let _ =
    (* TESTS: States of t have to be already defined in d and formula should not be false *)
    IFDEF TEST THEN
      Debug.asserteq ((valid_state delta q1) && (valid_state delta q2)) true
      "States of trans should belong to delta!";
     Debug.assertneq f Formula.mk_false
      "addR: No need to trigger add for mk_false formula" 
      ELSE () END
  in
    (* we have q1->q2 || fcurrent and we add q1->q2||fnew *)
    if q1 <> q2 && not(f=Formula.mk_false) then
      let phi = delta.epsilonR#fclosure.(q2).(q1) in
        if not (Formula.subsume f phi) then (
          Verbose.print 3 (fun _ -> "Adding symbolic epsilon transition " ^
                             (epsiext_to_string (q1, q2, f)) ^ "\n");
          (* let phi' = Formula.mk_or phi f in *)
          delta.epsilonR#addR q1 q2 f;
          Verbose.print 3
            (fun _ -> "Dsucc=" ^ (epsihash_to_string delta.epsilonR#succs) ^ "\n");
          Verbose.print 3
            (fun _ -> "Dpred=" ^ (epsihash_to_string delta.epsilonR#preds) ^ "\n")
        )

let addR' (delta: t)((q1, q2, f): external_epsilon) : bool =
  let _ =
    (* TESTS: States of t have to be already defined in d and formula should not be false *)
    IFDEF TEST THEN
      Debug.asserteq ((valid_state delta q1) && (valid_state delta q2)) true
      "States of trans should belong to delta!";
     Debug.assertneq f Formula.mk_false
      "addR': No need to trigger add for mk_false formula" 
      ELSE () END
  in
    if q1 <> q2 && not(f=Formula.mk_false) then
      (* we have q1->q2 || fcurrent and we add q1->q2||fnew *)
      let phi = delta.epsilonR#fclosure.(q2).(q1) in
        if not (Formula.subsume f phi) then (
          Verbose.print 3 (fun _ -> "Adding symbolic epsilon transition " ^
                             (epsiext_to_string (q1, q2, f)) ^ "\n");
          (* let phi' = Formula.mk_or phi f in *)
          delta.epsilonR#addR q1 q2 f;
          Verbose.print 3
            (fun _ -> "Dsucc=" ^ (epsihash_to_string delta.epsilonR#succs) ^ "\n");
          Verbose.print 3
            (fun _ -> "Dpred=" ^ (epsihash_to_string delta.epsilonR#preds) ^ "\n");
          true
        ) else false
    else false




(* Heuristic_one : the key idea is to order the cunjunctions of a DNF
   formula from the smallest one to the greatest one.  Thus, we remove
   the first transition occuring in each cunjunctions from the formula
   f until finding the smallest set making the whole formula be false.
   On our working example, it is sufficient for ensuring termination. *)
let rec heuristic delta sol : float =
  let degrees = delta.epsilonE#degrees in
  let res, len =
    List.fold_left
      (fun (sum, len) (q1, q2, _) ->
         match degrees.(q2).(q1) with
	   | Ghost -> (sum, len + 1)
           | Alive d -> (d + sum, len + 1)
           | Away ->
	       failwith ("Transitions.heuristic: " ^ (string_of_int q1) ^ " -> "
                         ^ (string_of_int q2) ^ " was already removed")
      ) (0, 0) sol
  in
    (float_of_int res) /. (float_of_int len)


(* the best renvoie "la meilleure" solution guidee par la fonction heuristique
   a partir d'un ensemble de solutions donnees sols *)
let find_best delta =
  let rec find_best selected min = function
    | [] -> selected
    | sol::tl ->
	let cur = heuristic delta sol in
	  if cur < min then (* i.e. sol is better than selected *)
	    find_best sol cur tl
	  else
	    find_best selected min tl
  in
    function
      | [] ->
          failwith "The best: No solution; quite hard to find the best one :-)"
      | s1::sols ->
          find_best s1 (heuristic delta s1) sols


(* [prune Δ Φ] remove transitions epsE and epsR of Δ. Thanks to the heusitic
   function applied to Φ, we select transitions of epsE that have to be removed.
   It impacts also that all transition q —Φ—> q' ∈ epsR is labeled by Φ' = Φ ^ ¬ (Eq(q₁, q₂)
   ^ ... ^ ¬ Eq(q_{n-1}, q_n). If Φ' |- ⊥, q—Φ'—>q' is removed from epsR.
*)
let prune (delta: t) (alpha: Formula.t) : unit =
  let elect =
      let all_solutions = Formula.negatives (Formula.mk_not alpha) in
        find_best delta all_solutions
  in
    Format.printf "pruning... \n@?";
    List.iter
      (fun (q1, q2, k) ->
         infer_transitions delta q1 q2;
         delta.epsilonE#removeE q1 q2;
         delta.epsilonR#removeE q1 q2;
         Formula.kill_atom q1 q2 k
      ) elect;
    Format.printf "pruning... done!\n@?"


(* Returns a list of all combinations for rewriting of any occurence
   of [e] by [x] in [l] *)
(** @UNUSED *)
let all_rewritings l a x =
  let rec all_rewritings = function
    | [] -> [[]]
    | b::tl ->
        let all_rew_tl = all_rewritings tl in
          List.fold_right
            (fun l -> List2.cons (List2.cons b l))
            all_rew_tl
            (if a = b then
               List.map (List2.cons x) all_rew_tl
             else
               []
            )
  in
    List.tl (all_rewritings l)

(* This function is used to propagate epsilon on regular list of
   transitions so as to produce a t.a. without epsilon

   Simply use the epsilonE and R tables... this is not optimal

   Use automaton.export to use state merging on <-E*-> cycles.
*)

(* Translation of a R/E automaton to a regular automaton without epsilon transitions.
   For each transition c -> q of les, we add c -> q' if q -> q' in epsilonRE closure.
   We do this only for states that are not in erased list (states removed by merging).
*)

let propagate_epsilon (les: external_symtrans list) (erased: state list)(delta: t) : external_symtrans list =
  let phi = delta.epsilonR#fclosure in
  let last = delta.epsilonR#last in
    List.fold_left
      (fun acc (f, sl, dest) ->
         let acc = ref acc in
           for s = 0 to last do
             if Formula.is_sat phi.(s).(dest) && (not (List.mem s erased)) then (
               acc := (f, sl, s)::!acc;
             )
           done;
           (* We add all the basic transitions of les to acc (Some of
              them may be related to some states not occuring in the
              closures of epsilonRE) *)
           !acc
      ) [] les




(* We propagate epsilon transitions in upward direction, i.e. if f(...q1...) -> q0 and q2 -> q1,
   we add f(...q2...) -> q0. Terminates only if there is no cycle of epsilon transitions in le. *)

let rec propagate_up_one_layer (le: external_epsilon list) ((f, ls, q): external_symtrans)
    (acc: external_symtrans list) : external_symtrans list =
  match le with
    | [] -> acc
    | (q1,q2,_)::rem ->
        (* We have epsilon transition q1 -> q2 and f(ls) -> q. If q2 occurs in ls then we produce all possible
           rewritings of f(ls)->q by q2 ->R q1 *)
        propagate_up_one_layer rem (f, ls, q)
          (List.fold_left
             (fun acc sl -> (f, sl, q) :: acc)
             acc
             (all_rewritings ls q2 q1)
          )

module ExtTrans =
struct
  type t = external_symtrans

  let compare = compare
  let to_string = symext_to_string
end

module ExttransSet = Set2.Make(ExtTrans)


let propagate_epsilon_up (le: external_epsilon list) (ls: external_symtrans list) : external_symtrans list =
  let rec prop_aux ls (acc: ExttransSet.t) : external_symtrans list =
  match ls with
    | [] -> ExttransSet.elements acc
    | (f, ls, q)::rem ->
        (* If the transitions has already been treated we skip it *)
        if ExttransSet.mem (f, ls, q) acc then
          prop_aux rem acc
        else
          (* We propagate on one layer (res) and resulting transitions are
             themselves used for propagation *)
          let res = propagate_up_one_layer le (f, ls, q) [] in
            prop_aux (List.rev_append res rem) (ExttransSet.add (f, ls, q) acc)
  in
    prop_aux ls ExttransSet.empty






      IFDEF TEST THEN
open Debug
      ENDIF





      IFDEF TEST THEN
      print_string "Transitions.ml\n\n"
      END

      IFDEF TEST THEN
      begin
        let dtest = make () in
          for i = 0 to 4 do
            dtest.epsilonE#incr_last
          done;
          Format.printf "Test des fonctions addE, removeR\n@?";
          let t = dtest.epsilonE#degrees in
            addE dtest 1 2;
            asserteq t.(2).(1) (Alive 0) "Transition 1->2 correctly added";
            removeE dtest 1 2;
            asserteq t.(2).(1) Ghost "Transition 1->2 correctly removed";
            addE dtest 2 3;
            removeE dtest 2 3;
            asserteq t.(3).(2) Ghost "Transition 2->3 correctly added and removed";
            asserteq t.(3).(1) (Alive 1) "Transition 1->3 correctly infered";
      end
      END



      IFDEF TEST THEN
      print_string "Test trans_to_transitionsBySymbols\n\n"
      END

      IFDEF TEST THEN
let dyoh = make ()
let s0 = make_state dyoh
let s1 = make_state dyoh
let s2 = make_state dyoh
let s3 = make_state dyoh
let s4 = make_state dyoh
let s5 = make_state dyoh

let t1 = ("f", [s1; s2], s0)  (* f(q1, q2) -> q0 *)
let t2 = ("g", [s1; s0], s2)  (* g(q1, q0) -> q2 *)
let t2b= ("g", [s2; s2], s2)  (* g(q2, q2) -> q2 *)
let t5 = ("f", [s2; s0], s1)  (* f(q2, q0) -> q1 *)
let t6 = ("a", [], s0)        (* a -> q0 *)
let t6b= ("a", [], s2)        (* a -> q2 *)


let tR1 = (s0, s1, Formula.mk_true) (* q0 -> q1 || T *)
let tR2 = (s3, s0, Formula.mk_true) (* q3 -> q0 || T *)
let tR3 = (s4, s5, Formula.mk_atom s2 s1 Eq) (* q4 -> q5 || Eq(q2,q1) *)
let tE1 = (s1, s2, Formula.mk_true) (* q1 -> q2 *)
let tE2 = (s2, s1, Formula.mk_true) (* q2 -> q1 *)
END

IFDEF TEST THEN
begin
  add dyoh t1;  Format.printf "add transition t1 to d\n@?";
  add dyoh t2;  Format.printf "add transition t2 to d\n@?";
  add dyoh t2b; Format.printf "add transition t2b to d\n@?";
  add dyoh t6;  Format.printf "add transition t6 to d\n@?";
  add dyoh t6b; Format.printf "add transition t6b to d\n@?";
  add dyoh t5;  Format.printf "add transition t5 to d\n@?";
  addR dyoh tR1; Format.printf "add transition tR1 to d\n@?";
  addR dyoh tR2; Format.printf "add transition tR2 to d\n@?";
  addR dyoh tR3; Format.printf "add transition tR3 to d\n@?";
  addE dyoh s1 s2; Format.printf "add approximation to d\n@?";
  addE dyoh s2 s1; Format.printf "add approximation to d\n@?";
  let ttbs = trans_to_transitionsBySymbols dyoh.transitions in
    asserteq (Hashtbl.mem ttbs "f") true
      "check that the the symbol f is indeed a key of the hashtable";
    asserteq (Hashtbl.mem ttbs "g") true
      "check that the the symbol g is indeed a key of the hashtable";
    asserteq (Hashtbl.mem ttbs "a") true
      "check that the the symbol a is indeed a key of the hashtable";
    asserteq  (Hashtbl.find ttbs "f") [([s1; s2], s0); ([s2; s0], s1)] "ok";
end
END





IFDEF TEST THEN
    print_string "Test add\n\n"
END

IFDEF TEST THEN
let d = make ()

let state_nb = 6
let s =
  let s = Array.make state_nb 0 in
    for i = 0 to state_nb - 1 do
      s.(i) <- make_state d
    done;
    s

let t1 = ("f", [s.(1); s.(2)], s.(0))     (* f(q1, q2) -> q0 *)
let t2 = ("g", [s.(1); s.(0)], s.(2))     (* g(q1, q0) -> q2 *)
let t2b= ("g", [s.(2); s.(2)], s.(2))     (* g(q2, q2) -> q2 *)
let t5 = ("f", [s.(2); s.(0)], s.(1))     (* f(q2, q0) -> q1 *)
let t6 = ("a", [], s.(0))           (* a -> q0 *)
let t6b= ("a", [], s.(2))           (* a -> q2 *)


let tR1 = (s.(0), s.(1), Formula.mk_true) (* q0 -> q1 || T *)
let tR2 = (s.(3), s.(0), Formula.mk_true) (* q3 -> q0 || T *)
let tR3 = (s.(4), s.(5), Formula.mk_atom s.(2) s.(1) Eq) (* q4 -> q5 || Eq(q2,q1) *)
let tE1 = (s.(1), s.(2), Formula.mk_true) (* q1 -> q2 *)
let tE2 = (s.(2), s.(1), Formula.mk_true) (* q2 -> q1 *)

(* added later *)
let tE3 = (s.(2), s.(3), Formula.mk_true) (* q2 -> q3 *)
  END

  IFDEF TEST THEN
  begin
    add d t1; Format.printf "add transition t1 to d";
    add d t2; Format.printf "add transition t2 to d";
    add d t6; Format.printf "add transition t6 to d";
    addR d tR1; Format.printf "add transition tR1 to d";
    addR d tR2; Format.printf "add transition tR2 to d";
    addR d tR3; Format.printf "add transition tR3 to d";
    addE d s.(1) s.(2); Format.printf "add approximation to d";
    addE d s.(2) s.(1); Format.printf "add approximation to d";
    print_string (to_string d);

    (* should be equivalent (modulo ordering) to

       g(q1,q0) -> q2
       a -> q0
       f(q1,q2) -> q0

       EpsilonR=
       q0 -> q1
       q3 -> q0

       EpsilonE=
       q1 -> q2
       q2 -> q1

    *)
    Format.printf "To string check trans d= \n";
    print_delta d;
  end
  ENDIF

  IFDEF TEST THEN
  begin
    print_string "Test get_predRE, get_succRE, get_predE, get_succE, is_succE\n\n";
    asserteq (fst (List.split (get_succRE d s0))) [s2; s1] "Successors of q0 by R/E";
    asserteq (fst (List.split (get_succRE d s3))) [s2; s1; s0] "Successors of q3 by R/E";
    asserteq (fst (List.split (get_predRE d s1))) [s3; s2; s0] "Predecessors of q1 by R/E";
    asserteq (fst (List.split (get_predRE d s3))) [] "Predecessors of q3 by R/E";
    asserteq (fst (List.hd (get_succRE d s1))) s2 "succ of q1 is q2";
    asserteq (snd (List.hd (get_succRE d s1))) (Formula.mk_atom s1 s2 Eq)
      "succ of q1 is q2 with formula Eq(q1,q2)";

    addE d s.(2) s.(3); print_string "add approximation to d";
    asserteq (get_predE d s3) [s2; s1] "Epredecessors of q3";
    print_fclosure d.epsilonE;
    asserteq (get_succE d s1) [s3; s2] "get_succE of q1";
    asserteq (is_succE d s2 s1) true "q1 is succE of q2";
    asserteq (is_succE d s1 s3) true "q3 is succE of q1";
    asserteq (is_succE d s3 s1) false "q1 is not succE of q3";
    removeE d s.(2) s.(3); (* infere la transition q1 -> q3 *)
    removeE d s.(1) s.(3)  (* on la retire : ok *)
  end
END

  (* the add function should avoid to add covered transitions *)
IFDEF TEST THEN
  print_string "Test mem\n\n"
END

IFDEF TEST THEN
let trans3 = (make ()).transitions
let trans4 = (make ()).transitions
let trans5 = (make ()).transitions
END



IFDEF TEST THEN
begin
  print_string "Transitions should not be added twice\n\n";
  asserteq (add_fast_trans t1 d.transitions) false "";
  print_string (trans_to_string d.transitions);
  asserteq (trans_to_string d.transitions) "g(q1,q0) -> q2\na -> q0\nf(q1,q2) -> q0\n" "";
  assertneq (add_fast_trans t6 trans3) false "";
  asserteq (add_fast_trans t6 trans3) false "";
  asserteq (trans_to_string trans3) "a -> q0\n" "";
  assertneq (add_fast_trans t1 trans3) false "";
  (* assertneq (mem t1 trans3) false ""; *)
end
END


 (* adding the 3 first transitions should produce the same table with
    add and add_fast *)

IFDEF TEST THEN
  print_string "Test add_fast and add_fast_trans\n\n"
END

IFDEF TEST THEN
begin
  let _= add_fast t1 trans4 in
  let _= add_fast t2 trans4 in
  let _= add_fast t6 trans4 in
  let _= add_fast_trans t1 trans5 in
  let _= add_fast_trans t2 trans5 in
  let _= add_fast_trans t6 trans5 in
    asserteq trans4 trans5 "";
end
END


IFDEF TEST THEN
  print_string "Test epsilon transitions and get_trans\n"
END

IFDEF TEST THEN
let d2= make ()
let s0 = make_state d2
let s1 = make_state d2
let s2 = make_state d2
let s3 = make_state d2
let s4 = make_state d2

let t0d2 = ("f", [s0], s1)
let t0bd2= ("a", [], s1)
let t1d2 = t0d2  (* t1d2= f(q0) -> q1 *)
let t1bd2= t0bd2 (* t1bd2= a -> q1 *)
let t2d2= (s1,s2,Formula.mk_atom s4 s3 Eq) (* q1 -> q2|| q4 -> q3 *)
let t3d2= (s2,s3,Formula.mk_true) (* q2 -x1> q3 *)
let t4d2= (s3,s2,Formula.mk_true) (* q3 -> q2 *)
END

IFDEF TEST THEN
begin
  Format.printf "d2:\n";
  add d2 ("f", [s0], s1); Format.printf "add f(q0) -> q1\n";
  add d2 ("a", [], s1);   Format.printf "add     a -> q1\n";
  addR d2 t2d2;            Format.printf "addR   q1 -> q2 | Eq(q4, q3)\n";
  addE d2 s2 s3;           Format.printf "addE   q2 -> q3\n";
  addE d2 s3 s2;           Format.printf "addE   q3 -> q2\n@?";

  (*Format.printf "Rlast = %i\n@?" d2.epsilonR#last;
  print_string ".\n";
  print_delta d2;
  print_string ".\n";
  *)
  (* Trying to join f with q2, we find: f(q0) -> q1 ( ->R q2) *)
  asserteq (get_trans s2 "f" d2) [t0d2] "we expect f(q0) -> q1";
  (* Trying to join f with q3, we find: f(q0) -> q1 ( ->R q2 = q3) *)
  asserteq (get_trans s3 "f" d2) [t0d2] "we expect f(q0) -> q1, again";
  asserteq (get_trans s2 "a" d2) [t0bd2] "we expect a -> q1";
  asserteq (get_trans s0 "f" d2) [] ""
end
END


IFDEF TEST THEN
let s3= make_state d
let d'= copy d
END

IFDEF TEST THEN
begin
  print_string ("copy and delta_equal\n\n");
  asserteq (delta_equal d d) true "";
  asserteq (delta_equal d d') true "";
  print_string "d3="; print_delta d';  print_string "\n";
  addR d' (s0, s2, Formula.mk_true); Format.printf "add to d3\n";
  asserteq (delta_equal d d') false "d and d3 should be no longer equivalent";
  print_string "propagate_epsilon\n\n";
  print_string ("d avant propagate epsilon\n" ^ (to_string d) ^ "\n");
  let new_trans =
    propagate_epsilon
      (List.map
         (function Trans t2 -> t2 | _ -> assert false)
	 (trans_to_list d.transitions)
      ) [] d
  in
    print_string ("New transitions with epsilon propagate=\n" ^ (symlist_to_string new_trans));
    asserteq (symlist_to_string new_trans)
      "f(q1,q2) -> q2
f(q1,q2) -> q1
f(q1,q2) -> q0
a -> q2
a -> q1
a -> q0
g(q1,q0) -> q2
g(q1,q0) -> q1\n"
      "Transitions to q0 are added to q1"
end
END


IFDEF TEST THEN
let dbig = make ()
let si = ref s1
let sj = ref s2
let ls = ref []

END

IFDEF TEST THEN
begin
  print_string "Testing epsilonR et epsilonE matrix automatic extension\n\n";

  (* Build a sequence of Common.max_states+4 states all related by epsilon
     transitions, to check extension of epsilon matrix *)

  si := make_state dbig;
  ls := !si::!ls;
  for i = 0 to Common.max_state_symb + 3 do
    sj := make_state dbig;
    print_string "Adding epsilonR transitions to dbig\n";
    addR dbig (!sj, !si, Formula.mk_true);
    print_string "Adding epsilonE transitions to dbig\n";
    addE dbig !sj !si;
    ls := !sj::!ls;
    si := !sj
  done;
  (* We check that after extension all epsilon transitions and states
     are still there! *)
  si:= List.hd !ls;
  ls:= List.tl !ls;
  asserteq (valid_state dbig !si) true "Last added state is still there";
  List.iter
    (fun sj ->
       asserteq (valid_state dbig sj) true ("State " ^ (State.to_string sj) ^
                                            " has desappeared");
(*       asserteq (is_predR !si sj dbig) Formula.mk_true
         ("Epsilon trans: " ^
	    (external_to_string (Epsilon(!si,sj,Formula.mk_true))) ^
	    " has desappeared from EpsilonR");
       asserteq (is_predE !si sj dbig) true
         ("Epsilon trans: " ^
	    (external_to_string (Epsilon(!si,sj,Formula.mk_true))) ^
	    " has desappeared from EpsilonE");
*)
       si := sj
    ) !ls
end
END


IFDEF TEST THEN
let d11 = make ()
let s1 = make_state d11
let s2 = make_state d11
let s3 = make_state d11
let _ = addR d11 (s1,s2,Formula.mk_true)
let set1 = Stateset.singleton s2
let set2 = get_all_pred set1 d11
let set3 = Stateset.add s1 set1
END

IFDEF TEST THEN
begin
  print_string "Test get_all_pred and state_set_replace";
  asserteq set2 set3 "get_all_pred on sets of states fails";
  asserteq set1 (Stateset.replace s1 s2 set3) "replace replaces when state found";
  asserteq set1 (Stateset.replace s1 s3 set1) "replace doesn't replace when state not found"
end
END


IFDEF TEST THEN
(* Test du papier *)
(* Test sur les transitions issues de l'approximation de la première étape de complétion*)
let delta = make () in
let q0 = make_state delta in (* q0 *)
let q1 = make_state delta in (* q1 *)
let q2 = make_state delta in (* q2 *)
let q3 = make_state delta in (* q3 *)
let q4 = make_state delta in (* q4 *)
let t1 = ("f", [q1], q0) in  (* f(q1) -> q0 *)
let t2 = ("zero", [], q1) in (* zero -> q1 *)
let t3 = ("s", [q1], q2) in  (* s(q1) -> q2 *)
let t4 = ("s", [q2], q3) in  (* s(q2) -> q3 *)
let t5 = ("f", [q3], q4) in  (* f(q3) -> q4 *)
let t9 = (q4, q0, Formula.mk_true) in
(* let tapprox1 = q2, q3, Formula.mk_true in
   let tapprox2 = q3, q2, Formula.mk_true in
*)
  begin
    add delta t1; Format.printf "add transitions";
    add delta t2; Format.printf "add transitions";
    add delta t3; Format.printf "add transitions";
    add delta t4; Format.printf "add transitions";
    add delta t5; Format.printf "add transitions";
    addR delta t9; Format.printf "add transitions";

    addE delta q2 q3; Format.printf "add approximation";
    addE delta q3 q2; Format.printf "add approximation";

    print_string ("\n"^(to_string delta));
    print_string "\n";
    print_string (fmatrix_to_string delta.epsilonR);
    print_string "\n";

    let phi1 =
      Formula.mk_and
        delta.epsilonR#fclosure.(3).(2)
        delta.epsilonR#fclosure.(2).(5)
    in
      print_string ("phi1: " ^ (Formula.to_string phi1) ^ "\n");
      let phi2 =
        Formula.mk_and
          delta.epsilonR#fclosure.(3).(6)
          delta.epsilonR#fclosure.(5).(3)
      in
	print_string (Formula.to_string phi2 ^ "\n");
        let phi3 = delta.epsilonR#fclosure.(3).(2) in
	  print_string (Formula.to_string phi3 ^ "\n");
          let phi4 = delta.epsilonR#fclosure.(2).(3) in
	    print_string (Formula.to_string phi4 ^ "\n");
            let phi5 =
              Formula.mk_and
                delta.epsilonR#fclosure.(5).(2)
                delta.epsilonR#fclosure.(5).(6)
            in
	      print_string (Formula.to_string phi5 ^ "\n")
  end
END





IFDEF TEST THEN
    (* Test réunion : 14/12/10 *)
    (* Test sur les transitions issues de l'approximation de la
       première étape de complétion *)

let delta_reunion = make ()


let qa = make_state delta_reunion  (*qa*)
let q0 = make_state delta_reunion  (*q0*)
let q1 = make_state delta_reunion  (*q1*)
let q2 = make_state delta_reunion  (*q2*)
let q3 = make_state delta_reunion  (*q3*)
let q4 = make_state delta_reunion  (*q4*)
let q5 = make_state delta_reunion  (*q5*)
let q6 = make_state delta_reunion  (*q6*)



let t1 = ("f", [qa], q0) in (*f(qa) -> q0 *)
let t2 = ("a", [], qa) in   (*a -> qa *)
let t3 = ("s", [qa], q3) in (*s(qa) -> q3 *)
let t4 = ("s", [q3], q1) in (*s(q3) -> q1 *)
let t6 = ("s", [q1], q4) in (*s(q1) -> q4 *)
let t7 = ("s", [q4], q5) in (*s(q4) -> q5 *)
let t5 = ("f", [q1], q2) in (*f(q1) -> q2 *)
let t8 = ("f", [q5], q6) in (*f(q5) -> q6 *)

  begin
    add delta_reunion t1;
    add delta_reunion t2;
    add delta_reunion t3;
    add delta_reunion t4;
    add delta_reunion t5;
    add delta_reunion t6;
    add delta_reunion t7;
    add delta_reunion t8;
    addR delta_reunion (q2, q0, Formula.mk_true);
    addR delta_reunion (q6, q2, Formula.mk_true);

    addE delta_reunion q3 q1;
    addE delta_reunion q3 q4;
    addE delta_reunion q4 q3;
    addE delta_reunion q1 q4;
    addE delta_reunion q4 q1;
    addE delta_reunion q4 q5;
    addE delta_reunion q5 q4;

    print_string ("\n" ^ (to_string delta_reunion) ^ "\n");
    print_string (fmatrix_to_string delta_reunion.epsilonR);
    print_string "\n";
end
END


IFDEF TEST THEN
begin
  print_string "Test all_rewritings\n";
  asserteq (all_rewritings [1;2;3] 4 5) [] "when replacement fails";
  asserteq (all_rewritings [1;2;3;4;3] 3 5) [[1; 2; 3; 4; 5]; [1; 2; 5; 4; 3]; [1; 2; 5; 4; 5]]
    (* [[1; 2; 5; 4; 5]; [1; 2; 5; 4; 3]; [1; 2; 3; 4; 5]]*) "when replacement succeeds";
  print_string "Test propagate_epsilon_up\n";
  let le1 = [(2,3,Formula.mk_true);(3,4,Formula.mk_true)] in
  let le2 = [(3,4,Formula.mk_true);(2,3,Formula.mk_true)] in
  let ls = [("f", [4;3;5], 0)] in
  let res1= (propagate_epsilon_up le1 ls) in
  let res2= (propagate_epsilon_up le2 ls) in
    print_string ("\nres1= "^(symlist_to_string res1));
    print_string ("\n\nres2= "^(symlist_to_string res2));
    asserteq (List2.equal res1 res2) true "result of propagate_epsilon_up does not depend on order of epsilon transitions";
    (* propagate epsilon up of f(q4,q3,q5) -> q0  with q2 -> q3 -> q4 results in the following trans*)
    asserteq res1 [("f",[2;2;5],0);("f",[2;3;5],0);("f",[3;2;5],0);("f",[3;3;5],0);("f",[4;2;5],0);
                   ("f",[4;3;5],0)] "propagate epsilon up of f(q4,q3,q5) -> q0  with q2 -> q3 -> q4 results in the following trans"
 end
END



